/*
 * Copyright (c) 2000-2004
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <assert.h>		// assert
#include "dsu.h"
#include "nonspec.h"
#include "hash.h"

/*****************************************************************************
 *                                                                           *
 *   Type Declarations                                                       *
 *                                                                           *
 *****************************************************************************/


typedef enum {dsu_raw, dsu_inited, dsu_query} dsu_state;

struct dsu_node_ {
  gen_e node_cv;
  char *unique_name;
};

DEFINE_NONPTR_LIST(dsu_node_list,dsu_node);

/*****************************************************************************
 *                                                                           *
 *   Global Variables                                                        *
 *                                                                           *
 *****************************************************************************/

// A map from indices to the clustered open constructor containing that index 
static hash_table constant_to_node_map = NULL;
static hash_table variable_to_node_map = NULL;
static hash_table constant_to_int_map = NULL;

static int const_count = 0;
static int var_count = 0;

static region dsuregion = NULL;
static dsu_state state;

int flag_dsu_print_constraints = 0;

static dsu_node_list all_nodes = NULL;

/*****************************************************************************
 *                                                                           *
 *   Static Utility Functions                                                *
 *                                                                           *
 *****************************************************************************/

static void my_call_setif_inclusion(gen_e e1, gen_e e2)
{
  assert(state == dsu_inited);
  if (flag_dsu_print_constraints) {
    expr_print(stdout,e1);
    printf(" <= " );
    expr_print(stdout,e2);
    printf("\n");
  }
  call_setif_inclusion(e1,e2);
}


/*****************************************************************************
 *                                                                           *
 *   Initialization/ Resetting                                               *
 *                                                                           *
 *****************************************************************************/

void dsu_init()
{
#ifdef DSU__DEBUG
  printf("dsu_initc...\n");
#endif

  assert(state == dsu_raw);
  
  dsuregion = newregion();

  constant_to_node_map = make_hash_table(dsuregion, 32, ptr_hash, ptr_eq);
  variable_to_node_map = make_hash_table(dsuregion, 32, ptr_hash, ptr_eq);
  constant_to_int_map  = make_hash_table(dsuregion, 32, ptr_hash, ptr_eq);

  all_nodes = new_dsu_node_list(dsuregion);
  state = dsu_inited;

  const_count = 0;
  var_count = 0;

#ifdef DSU__DEBUG
  printf("dsu_init done.\n");
#endif
}

void dsu_reset(void)
{
  deleteregion(dsuregion);

  dsuregion = NULL;

  all_nodes = NULL;
  state = dsu_raw;

  const_count = 0;
  var_count = 0;

}

/*****************************************************************************
 *                                                                           *
 *   Graph Construction API                                                  *
 *                                                                           *
 *****************************************************************************/

dsu_node make_const_dsu_node(const char *name)
{
  char unique_name[512];
  dsu_node result = ralloc(dsuregion,struct dsu_node_);
  assert(state == dsu_inited);

  snprintf(unique_name,512,"%s_%d",name, const_count);
  result->node_cv = setif_constant(unique_name);
  result->unique_name = rstrdup(dsuregion,unique_name);

  dsu_node_list_cons(result, all_nodes);
  hash_table_insert(constant_to_node_map, (hash_key) result->node_cv,
		    (hash_data)result);

  hash_table_insert(constant_to_int_map, (hash_key) result->node_cv,
		    (hash_data)const_count);

  const_count++;

  return result;
}

dsu_node make_var_dsu_node(const char *name)
{
  char unique_name[512];
  dsu_node result = ralloc(dsuregion,struct dsu_node_);
  assert(state == dsu_inited);

  snprintf(unique_name,512,"%s_%d",name, var_count);
  result->node_cv = setif_fresh(name);
  result->unique_name = rstrdup(dsuregion,unique_name);

  dsu_node_list_cons(result, all_nodes);
  hash_table_insert(variable_to_node_map, (hash_key) result->node_cv,
		    (hash_data)result);

  var_count++;

  return result;
}

void make_dsu_subtype_edge(dsu_node n1, dsu_node n2)
{
  assert(state == dsu_inited);
  my_call_setif_inclusion(n1->node_cv, n2->node_cv);

}

/*****************************************************************************
 *                                                                           *
 *   Query API                                                               *
 *                                                                           *
 *****************************************************************************/

void dsu_finished_adding()
{
  assert(state == dsu_inited);
  state = dsu_query;

  return;
}

bool dsu_check_reaches(dsu_node n1, dsu_node n2)
{  
  gen_e_list_scanner scan;
  gen_e next_lb;
  // compute the transitive lower bounds of n2
  gen_e_list tlb = setif_tlb(n2->node_cv);
  
  assert(state == dsu_query);
  
  // check whether n1's constant is a member of n2
  gen_e_list_scan(tlb,&scan);
  
  while (gen_e_list_next(&scan,&next_lb)) {
    if (expr_eq(n1->node_cv,next_lb)) return TRUE;
  }

  return FALSE;
}

void dsu_sweep_reachable_consts(dsu_node n, int * scratch_pad)
{  
  int i;

  gen_e_list_scanner scan;
  gen_e next_lb;

  // compute the transitive lower bounds of n
  gen_e_list tlb = setif_tlb(n->node_cv);
  
  assert(state == dsu_query);
  
  gen_e_list_scan(tlb,&scan);

  // now for all sets in n's tlb: it they're constants, tick them off in the scratchpad

  while (gen_e_list_next(&scan,&next_lb)) {
    if (hash_table_lookup(constant_to_int_map,
                        (hash_key)next_lb,
                        (hash_data *)&i)) {  
      scratch_pad[i] = 1;
    }
  }
}

