(* 
 *
 * Copyright (c) 2003-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



module H = Hashtbl

type   dtree = Node of string * int * dtree list
type   dforest = (string, dtree) H.t


let sumUp (tr:dtree) : dtree = 
  let rec su (t:dtree) : (int * dtree) = 
    match t with
        Node (pathComponent,count,[]) -> (count, t)
      | Node (pathComponent,count,l) -> 
          let childrenList : (int * dtree) list = List.map su l in
          let childrenCounts = List.map fst childrenList in
          let childrenTrees  = List.map snd childrenList in
          let sumOfChildren = List.fold_right 
                                (fun a accu -> (a + accu)) 
                                childrenCounts 0 in
            (sumOfChildren, 
             Node(pathComponent, sumOfChildren, childrenTrees))
  in 
  let (_, res) = su tr in 
    res

let prettyPrint (tr:dtree) : unit = 
let rec pp (level:int) (t:dtree) : unit = 
  let Node(pathComponent,count,l) = t in
  let format = 
    for i = 0 to level do Printf.printf "     "   done; in
    begin
      if l = [] 
      then Printf.printf "%s : %d\n" pathComponent count
      else Printf.printf "%s/ : %d\n" pathComponent count;
      List.iter (pp (level + 1)) l
    end
in pp 0 tr

let toTree (spath:string) (count:int) : dtree = 
  let lpath =  Str.split (Str.regexp "/") spath in
  let rec ins (sl:string list) (count:int) : dtree = 
    match sl with
        []  -> Node ("",count, [])
      | [h] -> Node (h,count, [])
      | h::tl -> Node(h,0,[ins tl count])
  in ins lpath count
       

let rec insertInTree (oldTree:dtree) (newItem:dtree) : dtree =
    let Node(op,oc,ol) = oldTree and Node(np,nc,nl) = newItem in 
      if op = np 
      then 
        if nl = [] 
        then Node (op, oc + nc, ol)
        else 
          if ol = [] 
          then Node (op, oc + nc, nl)
          else
            Node (op, oc, List.fold_right insertInForest nl ol )
      else
        oldTree
and
  insertInForest  (t:dtree) (f:dtree list) : dtree list = 
  let Node(rpathComponent,rcount,rl) = t in
  let (yes,no) = List.partition 
                   (fun x -> let Node(p,c,l) = x in  p = rpathComponent) f
  in 
    match yes with
        [] ->  f @ [t]
      | h::_ -> no @ [insertInTree h t]
          
class distributionTree (name: string) = object (self) 
  val storeName : string = name
  val store : dforest = (H.create 255)

  method addPathVal (newPath:string) (count:int) : unit = 
(*     let nada= Printf.printf "Adding <%s> <%d> \n" newPath count in *)
    let lpath =  Str.split (Str.regexp "/") newPath in
    let root = if lpath = [] then "" else List.hd lpath in
      try  
        let oldPath = H.find store root in
          begin
            H.remove store root;
            H.add store root  (insertInTree oldPath (toTree newPath count))
          end
      with Not_found -> 
        H.add store root  (toTree newPath count)
        
  method printAll = 
    Printf.printf "\n------- Distribution Tree : %s  ------- \n" storeName;
    H.iter (fun key v -> prettyPrint (sumUp v)) store;

end

(*
let t = Node ("types.h", 2, [])
let a = Node ("asm.h", 1, [])
let s = Node ("stdio.h", 3, [])
let l = Node ("linux",0,[t;a;s])
let y = Node ("sys",0,[a])
let u = Node ("include",0,[l;y])

let f = [t;a;s]

let t2 = toTree "/include/linux/nasm.h" 5
let t1 = toTree "/include/linux/asm.h" 5
let t3 = insertInTree t1 t2
let t4 = insertInTree u t3
*)
