(* 
 *
 * Copyright (c) 2003-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Commondefs

type varFact = 
{ 
   v_name : string;  
   v_change: change_t 
}

type globalVarFact = 
{ 
   g_name : string;  
   g_change: change_t 
}

type typedefFact = 
{ 
   t_name : string;  
   t_change: change_t
}

type functionFact = 
{ 
   f_name : string;  
   mutable f_change : change_t; 
   mutable f_returnValChange : change_t;
   mutable f_argumentsChange : (varFact list) 
}

type compFact = 
{ 
   c_name : string;
   c_isStruct : bool;
   mutable c_change : change_t; 
   mutable c_fieldsChange : (varFact list) 
}

type enumFact = 
{ 
   e_name : string;
   mutable e_change : change_t; 
   mutable e_valsChange : (varFact list); 
}

type fact =   
     FGlobalVar of globalVarFact 
   | FTypedef of typedefFact 
   | FFunction of functionFact
   | FComp of compFact
   | FEnum of enumFact

let getChangeOfFact (f : fact) : change_t = match f with
     FGlobalVar g -> g.g_change 
   | FTypedef t -> t.t_change 
   | FFunction f -> f.f_change 
   | FComp c -> c.c_change 
   | FEnum e -> e.e_change 
  

let filterFactList (fl : fact list) (ct : change_t) = 
  let rec internalFilter fl accu = 
    match fl with
        [] -> accu
      | h::t -> 
          begin
            let hc = getChangeOfFact h in
            match ct, hc with 

                Added, Added  
              | Deleted, Deleted 
              | TypeChanged (_,_), TypeChanged (_,_)
              | ValChanged, ValChanged
              | NoChange, NoChange
              | OtherChange _, OtherChange _ 
              | Renamed _, Renamed _ -> internalFilter t (h::accu) 

              | _, _ -> internalFilter t accu
          end
  in List.rev (internalFilter fl [])

let filterVarFactList (vl : varFact list) (ct : change_t) = 
  let rec internalFilter vl accu = 
    match vl with
        [] -> accu
      | h::t -> 
          begin
            match ct, h.v_change with 
                Added, Added
              | Deleted, Deleted
              | ValChanged, ValChanged
              | TypeChanged (_,_), TypeChanged (_,_)
              | Renamed _, Renamed _  -> internalFilter t (h::accu)
              | _,_ -> internalFilter t accu
          end
  in List.rev (internalFilter vl [])

let getNamesOnly (vl : varFact list) = List.map (fun x -> x.v_name) vl

let getOldType = function
    TypeChanged (tOld,tNew) -> tOld
  | _ -> ""

let getNewType = function
    TypeChanged (tOld,tNew) -> tNew
  | _ -> ""

let doTypeChangeList (vl : varFact list) (af : string) : unit = 
  let 
    d = filterVarFactList vl Deleted and
    a = filterVarFactList vl Added and
    c = filterVarFactList vl (TypeChanged ("","")) and
    r = filterVarFactList vl (Renamed ("","")) and
    v = filterVarFactList vl ValChanged 
  in
  let 
    nd = List.length d and
    na = List.length a and
    nr = List.length r and
    nc = List.length c and
    nv = List.length v 
  in
    begin
      if nd > 0 || na > 0 || nc > 0 || nr > 0 || nv > 0 then begin
        Printf.printf "%s :" af;
      end;
      if nd > 0 then begin
        Printf.printf "\n\t %d deleted: " nd;
        List.iter (fun x -> Printf.printf "\"%s\" " x) (getNamesOnly d);
      end;
      if na > 0 then begin
        Printf.printf "\n\t %d added: " na;
        List.iter (fun x -> Printf.printf "\"%s\" " x) (getNamesOnly a);
      end;
      if nr > 0 then begin
        Printf.printf "\n\t %d renamed: " nr;
        List.iter (fun x -> Printf.printf "\"%s\" " x) (getNamesOnly r);
      end;
      if nc > 0 then begin
        Printf.printf "\n\t %d changed type: " nc;
        List.iter (fun x -> Printf.printf "\n\t\t\"%s\" changed type from %s to %s"  x.v_name (getOldType x.v_change) (getNewType x.v_change))  c;
      end;
      if nv > 0 then begin
        Printf.printf "\n\t %d val changed: " nv;
        List.iter (fun x -> Printf.printf "\"%s\" " x) (getNamesOnly v);
      end;
      if nd > 0 || na > 0 || nc > 0 || nv > 0 then begin
        Printf.printf "\n";
      end
    end


let printFact (f : fact) = match f with
    FGlobalVar g ->
      begin
        match g.g_change with
            Added ->   Printf.printf "+ global var \"%s\"\n" g.g_name
          | Deleted -> Printf.printf "- global var \"%s\"\n" g.g_name
          | TypeChanged (tOld, tNew) -> Printf.printf "  global var \"%s\" changed type from %s to %s\n" g.g_name tOld tNew
          | ValChanged -> Printf.printf "  global var \"%s\" changed static initializer\n" g.g_name
          | Renamed (o,n) -> Printf.printf "  global var \"%s\" renamed from \"%s\" to \"%s\"\n" g.g_name o n
          | _ -> ()
      end

  | FTypedef t ->
      begin
        match t.t_change with
            Added ->   Printf.printf "+ typedef \"%s\"\n" t.t_name
          | Deleted -> Printf.printf "- typedef \"%s\"\n" t.t_name
          | TypeChanged (tOld, tNew) -> Printf.printf "  typedef \"%s\" changed base type from %s to %s\n" t.t_name tOld tNew
          | Renamed (o,n) -> Printf.printf "  typedef \"%s\" renamed from \"%s\" to \"%s\"\n" t.t_name o n
          | _ -> ()
      end

  | FFunction f ->
      begin
        match f.f_change with
            Added ->   Printf.printf "+ function \"%s\"\n" f.f_name
          | Deleted -> Printf.printf "- function \"%s\"\n" f.f_name
          | Renamed (o,n) -> Printf.printf "  function \"%s\" renamed from \"%s\" to \"%s\"\n" f.f_name o n
          | ValChanged -> Printf.printf "  function \"%s\" changed body\n" f.f_name 
          | (OtherChange _ |TypeChanged(_,_)) -> 
              begin
                (*Printf.printf "function %s: \n" f.f_name;*)
                begin
                  match f.f_returnValChange with
                      TypeChanged (tOld, tNew) -> Printf.printf "  function \"%s\": return type changed from %s to %s\n" f.f_name tOld tNew
                    | _ -> ()
                end;
                if List.length f.f_argumentsChange > 0 then begin
                  Printf.printf "  function \"%s\": " f.f_name;
                  doTypeChangeList f.f_argumentsChange "arguments"
                end
              end
          | _ -> ()
      end

  | FComp c ->
      begin
        let su = if c.c_isStruct then "struct" else "union " in
        match c.c_change with
            Added ->   Printf.printf "+ %s \"%s\"\n" su c.c_name
          | Deleted -> Printf.printf "- %s \"%s\"\n" su c.c_name
          | Renamed (o,n) -> Printf.printf "  %s \"%s\" renamed from \"%s\" to \"%s\"\n" su c.c_name o n
          | OtherChange s -> 
              begin
                Printf.printf "  %s \"%s:\" %s " su c.c_name s;
                doTypeChangeList c.c_fieldsChange "fields"
              end
          | _ -> ()
      end

  | FEnum e ->
      begin
        match e.e_change with
            Added ->   Printf.printf "+ enum \"%s\"\n" e.e_name
          | Deleted -> Printf.printf "- enum \"%s\"\n" e.e_name
          | Renamed (o,n) -> Printf.printf "  enum \"%s\" renamed from \"%s\" to \"%s\"\n" e.e_name o n
          | OtherChange s -> 
              begin
                Printf.printf "  enum \"%s\": %s " e.e_name s;
                doTypeChangeList e.e_valsChange "vals"
              end
          | _ -> ()
      end
