(* 
 *
 * Copyright (c) 2003-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* $Id: lmetrics.ml,v 1.4 2004/01/21 23:30:46 neamtiu Exp $ *)

open Cil
open Trace
open Statistics

module E = Errormsg
module H = Hashtbl

type hashOfVars = (string, bool) H.t
type hashOfAllModules = (string, hashOfVars) H.t

exception ImDone

let generalStatistics = new aggregateStatistics "Global Definitions"

let callsPerFunctionStatistics = new firstClassStatistics "Calls per Function"
let varReferencesStatistics    = new firstClassStatistics "References to Global Variables"
let commonCouplingsStatistics  = new firstClassStatistics "Common Couplings between Modules"

let allModules: hashOfAllModules ref = ref (H.create 127)

let rec analyzeExpression (containerFunction:global) (e : exp) : unit = 
  match e with
      Const c -> ()
    | Lval l -> analyzeLval containerFunction l
    | SizeOf t -> ()
    | SizeOfE e -> analyzeExpression containerFunction e
    | SizeOfStr s -> ()
    | AlignOf t -> ()
    | AlignOfE e -> analyzeExpression containerFunction e
    | UnOp (op,e,_) -> analyzeExpression containerFunction e
    | BinOp (op,e1,e2,_) -> analyzeExpression containerFunction e1;analyzeExpression containerFunction e2
    | CastE (t,e) -> analyzeExpression containerFunction e
    | AddrOf lval ->   analyzeLval containerFunction lval
    | StartOf lval ->  analyzeLval containerFunction lval

and analyzeLval (containerFunction:global) (l : lval) : unit = 
  match l with
      (Var v, o) -> 
        begin
          if (v.vglob) then
            begin
              (*Printf.printf "reference to global var \"%s\"\n" v.vname;*)
              varReferencesStatistics#incr v.vname;
            match containerFunction with
                GFun(f,l) ->  
                  begin 
                    callsPerFunctionStatistics#incr f.svar.vname;
                    Printf.printf 
                      "reference to global var \"%s\", used in \"%s\" and defined in \"%s\"\n" v.vname l.file v.vdecl.file;
                    if l.file <> v.vdecl.file 
                    then commonCouplingsStatistics#incrBin v.vdecl.file
                  end
              |  _ -> raise  (Invalid_argument "containerFunction not a GFun ?!")

            end
        end
    | (Mem e, o) -> analyzeExpression containerFunction e

and analyzeInstruction (containerFunction:global) (i: instr) : unit = 
    match i with
        Set (l,r,_) -> 
          begin
            analyzeLval containerFunction l; 
            analyzeExpression containerFunction r
          end
      | Call (res,fexpr,actuals,_) ->
          begin
            match containerFunction with
                GFun(f,l) ->  callsPerFunctionStatistics#incr f.svar.vname
              |  _ -> raise  (Invalid_argument "containerFunction not a GFun ?!")
;
            match fexpr with
                Lval(Var v, _) -> 
                  begin
                    (*Printf.printf "call to function \"%s\"\n" v.vname;*)
                    try 
                      List.iter (analyzeExpression containerFunction) actuals
                    with Invalid_argument _ -> ()
                  end
              | _ -> ()
          end
    | Asm _  -> ()

and analyzeStatement (containerFunction:global) (s : stmt) : unit =
  match s.skind with
      Instr il ->
        begin
          try 
            List.iter (analyzeInstruction containerFunction) il
          with Invalid_argument _ -> ()
        end
    | Return (eo,_) ->
        begin
          match eo with
            | Some e ->
                begin
                  analyzeExpression containerFunction e
                end
            | None -> ()
        end
    | If (e,bt,bf,_) ->
        begin
          analyzeExpression containerFunction e;
          analyzeBlock containerFunction bt;
          analyzeBlock containerFunction bf;
        end
    | Switch (e,b,sl,_) ->
        begin
          analyzeExpression containerFunction e;
          analyzeBlock containerFunction b;
          begin
            try
              List.iter (analyzeStatement containerFunction) sl
            with Invalid_argument _ -> ()
          end
        end
    | Loop (b,_,_,_) -> analyzeBlock containerFunction b
    | Block (b)  -> analyzeBlock containerFunction b
    | Goto (_,_) | Break _ | Continue _ | TryExcept (_, _, _, _) | TryFinally (_, _, _) -> ()

  and
    analyzeBlock (containerFunction:global) (b : block) : unit =
    List.iter (analyzeStatement containerFunction) b.bstmts




class lmetricsVisitor = object (self)
  inherit nopCilVisitor

  method private processGlobal (g:global) : unit = 
    match g with
      | GVar (v, i, l) -> generalStatistics#incr "Global variables"
      | GFun (f, l) -> generalStatistics#incr "Functions"; analyzeBlock g f.sbody
      | _ -> () 

  method vglob (g:global) : global list visitAction =  
    self#processGlobal g; DoChildren
      
end

let printStatistics () : unit = 
  begin
    generalStatistics#printAll ""; 

    callsPerFunctionStatistics#printDistribution false;   
    varReferencesStatistics#printDistribution false;
    commonCouplingsStatistics#printDistribution true;
  end

let doAnalysis () = 
  begin
    printStatistics (); 
    raise ImDone "Everything is fine"  (*might consider exiting at this point by using mergeGlobals *)
  end


let pre () =  
  begin
    (* Printf.printf "Lmetrics: pre\n" *)
  end


let post () =  
  begin
    doAnalysis ();
    (* Printf.printf "Lmetrics: post\n" *)
  end

let lmetricsEnabled = ref false

let feature : featureDescr = 
  { fd_name = "lmetrics";
    fd_enabled = lmetricsEnabled;
    fd_description = "compute some metrics to check Lehman SW evolution laws";
    fd_extraopt = [];
    fd_doit = 
                    (function (f: file) -> 
                       let lVisitor =  pre (); new lmetricsVisitor in
                       let res = visitCilFileSameGlobals lVisitor f
                       in post (); res
                    );
    fd_post_check = true;
  } 

