(* 
 *
 * Copyright (c) 2003-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



module H = Hashtbl

open Distree

(***************************   aggregateStatistics *********************************)

(* type (File1/File2/changed/added/deleted), headcount *)
type aggregateStoreType = (string, int) H.t

class aggregateStatistics (name: string) = object (self) 
  val storeName : string = name
  val store : aggregateStoreType = (H.create 7)
  val descr : (string, string) H.t = (H.create 7)
      
  method incrN key n =   
    try
      let v = H.find store key in
        begin
          H.remove store key;
          H.add store key (v + n) 
        end
    with Not_found -> 
      begin
        H.add store key n 
      end

  method incr key =  self#incrN key 1

  method set key v = 
    try  let nema = (H.remove store key) in
      begin
        H.add store key v 
      end
    with Not_found -> 
      begin
        H.add store key v
      end

  method get key = 
    try H.find store key
    with Not_found -> 0

  method reset (key,description) = 
     self#set key 0;
     H.add descr key description

  method printAll category = 
    Printf.printf "\n------- %s  ------- \n" storeName;

    let all_entries = Commonutil.hashtbl_keys store in
    let all_entries_sorted = List.sort Pervasives.compare all_entries in
      List.iter (fun key -> 
                   let value = H.find store key in
                     Printf.printf "%s %s : %6d\n" category key value) all_entries_sorted

  method sweepDescrVal : (string * int) list = 
     let sweep key value acc = 
        let desc = try H.find descr key with Not_found -> raise (Failure (storeName ^ key)) in
        (desc, value)::acc 
     in
     H.fold sweep store []
     
end

(***************************   firstClassStatistics *********************************)

(* objectName, headcount, location *)
type firstClassStoreType = (string, (int * string)) H.t

class firstClassStatistics (name: string) = object (self)
  val storeName : string = name
  val store : firstClassStoreType = (H.create 127)
  val dTree : distributionTree = new distributionTree name
      
  method add objectName headcount location = 
    begin
      dTree#addPathVal location headcount;
      try
        let v = H.find store objectName in
          begin
            H.remove store objectName;
            H.add store objectName (headcount,location)
          end
      with Not_found -> 
        begin
          H.add store objectName (headcount,location)
        end
    end

  method incrN objectName n =   
    try
      let headcount,location = H.find store objectName in
        begin
          H.remove store objectName;
          H.add store objectName (headcount + n,location)
        end
    with Not_found -> 
      begin
        H.add store objectName (n,"")
      end

  method incr objectName = self#incrN objectName 1

  method incrBin objectName  =   
    try
      let headcount,location = H.find store objectName in
        begin
          (* if already in, do nothing *)
        end
    with Not_found -> 
      begin
        H.add store objectName (1,"")
      end

  method printAll = 
    Printf.printf "\n------- %s  ------- \n" storeName;
    H.iter (fun objectName p -> let (headcount,location) = p in 
              Printf.printf "%s : %d (%s)\n" objectName headcount location) store;

  method printDistribution printIndividualItems = 
    let totalHeadcount = ref 0 in
    let overallMin = ref 100 in
    let overallMax = ref (-1) in
    let findMinMax objectName (headcount,location)  =
      begin
        if headcount < !overallMin
        then overallMin := headcount;

        if headcount > !overallMax
        then overallMax := headcount;
      end
    in
    let sumUp objectName (headcount,location)  =
      totalHeadcount := !totalHeadcount + headcount
    in
    let lada = H.iter sumUp  store in
    let nada = H.iter findMinMax  store in
      if !overallMin <> 100 &&  !overallMax <> (-1)
      then 
        let a =    Array.create (!overallMax - !overallMin + 1) 0 in
        let mada = H.iter (fun objectName (headcount,location) -> 
                             let oldval = a.(headcount - !overallMin) in 
                               a.(headcount - !overallMin) <- (oldval+1)) 
                     store in
          begin
            if !totalHeadcount > 0 then
              begin
                dTree#printAll;
                Printf.printf "\n------- %s (distribution) ------- \n" storeName;
                Printf.printf "[min %d, max %d, total %d] \n" !overallMin !overallMax !totalHeadcount;
                for i = 0 to Array.length a - 1 
                do 
                  Printf.printf "%d : %d " (i + !overallMin) a.(i) ;
                  if printIndividualItems
                  then 
                    begin
                      Printf.printf "(";
                      H.iter (fun objectName (headcount,location) -> 
                                if headcount = (i + !overallMin) 
                                then Printf.printf "%s," objectName) 
                        store;
                      Printf.printf ")"
                    end;
                  Printf.printf "\n"
                done
              end;
          end
end
