(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil
open Musolver
open Deltasolver


module H = Hashtbl



type effect = 
{ 
   effect_delta_in  : delta;   (* input effect  *)
   effect_delta_out : delta;   (* output effect *)
   effect_labels    : delta;   (* set of labels (function names) *)
   effect_mu        : mu      (* updatability  *)
}

(* Type of hash table to store function effects keyed on a unique effect id *)
type funEffectHash_t = (int, effect) H.t
(* Function name -> function type hash *)
type funTypeHash_t = (string, Cil.typ) H.t

(* replace "" with string_of_int s to print id's *)
let debug_string_of_int s = "";;

(* type of hash tables that map from statement ids to input/output effect pairs *)
type statementEffect = (int, (delta * delta)) H.t

(* type of hashtbl for goto targets: maps statement ids to deltas *) 
type gotoTargetHash = (int, delta) H.t

type constDeltaHash_t = (string, delta) H.t

(* constant nodes in constraint graph can be types/funs/gvars; 
these will appear on __DSU_update() annotations *)
type deltaConstKind = DckType | DckFun | DckGvar
type constNamesHash_t = (string, deltaConstKind) H.t
let int_of_dck  dck = match dck with DckGvar -> 2 | DckFun -> 1 | DckType -> 0;;

(* delta -> (longjmp target, crt_fun_mu) hash *)
type longjmpHash_t = (delta, (Cil.varinfo * mu)) H.t

(* delta id -> capability set hash *)
type deltaSolutionHash_t = (int, CapabilitySet.t) H.t

(* Results that are passed out of the analysis *)
type analysis_result = 
   {
      ar_delta_constraints       : DeltaConstraintSet.t;  
      ar_mu_constraints          : MuConstraintSet.t;
      ar_delta                   : delta;
      ar_mu                      : mu;
      ar_stmtid_to_delta_id      : (int,(int*int)) H.t;
      ar_funEffectHash           : funEffectHash_t;
      ar_speculativeUpdateMuHash : (int,int) H.t;
      ar_mu_solution             : (int,bool) H.t;
      ar_delta_solution          : deltaSolutionHash_t;
      ar_constNamesHash          : constNamesHash_t
   }





(* Misc. functions 
   TODO: sort these out
*)

let noFunction = Cil.emptyFunction "**No FUNCTION**"

(* transforms a hash of constraints (e.g. as used internally by the monad) into a set of constraints *)
let deltaConstraintHashToSet (constraintHash : ((bool * int * int), delta_constraint) H.t) init : DeltaConstraintSet.t = 
  let deltaConstraintHashToSet_fold key deltaConstraint acc = 
    makeSureThereSNoConstOnRight deltaConstraint;
    DeltaConstraintSet.add deltaConstraint acc 
  in
  H.fold deltaConstraintHashToSet_fold constraintHash init

(* transforms a hash of constraints (e.g. a s used internally by the monad) into a set of constraints *)
let muConstraintHashToSet (constraintHash : ((mu * mu), unit) H.t) : MuConstraintSet.t = 
  let muConstraintHashToSet_fold muConstraint _ acc = MuConstraintSet.add muConstraint acc in
  H.fold muConstraintHashToSet_fold constraintHash MuConstraintSet.empty



let makeDeltaDebugStrings (d1 : delta) (d2 : delta) : string*string = 
  let d1_id_str = string_of_int d1.delta_id in
  let d2_id_str = string_of_int d2.delta_id in
  let d1str = match d1.delta_kind with DKVar _ -> d1_id_str | DKConst s -> (s ^ "(" ^ d1_id_str ^ ")") in
  let d2str = match d2.delta_kind with DKVar _ -> d2_id_str | DKConst s -> (s ^ "(" ^ d2_id_str ^ ")") in
    d1str,d2str


(*let debugPrint str = Printf.printf str*)
let debugPrint (str:string) : unit = ignore (str)













(*============================================================================*)
(* Monad for the analysis *)

module AnalysisMonad = 
struct
  (* monad input / output types *)
  type mi    =
      { 
        ami_delta_constraints         : ((bool * int *int), delta_constraint) H.t;   (* delta constraint hash *)
        ami_mu_constraints            : ((mu * mu), unit) H.t;      (* updatability constraint hash *)
        ami_statement_effects         : statementEffect;            (* mapping from statement id's to effects *)
        ami_fwd_goto_targets          : gotoTargetHash;             (* forward goto targets (sid to bool hash) *)
        ami_mu                        : mu ref;                     (* current updatability *)
        mutable ami_delta             : delta;                      (* current delta *)
        ami_mu_id_counter             : int ref;                    (* the next unique id for mu variables *)
        ami_fun_effect_id_counter     : int ref;                    (* the next unique id for function effects *)
        ami_funEffectHash             : funEffectHash_t;            (* hash from effect id to function effect *)
        ami_solutionHash              : deltaSolutionHash_t;        (* hash from delta id to capability set *)
        ami_constDeltaHash            : constDeltaHash_t;           (* hash from type names to const deltas.
                                                                       Used so every const delta with the same name
                                                                       has the same delta_id *)
	ami_constNamesHash            : constNamesHash_t;           (* hash from type/fun/gvar name to its kind*) 
        ami_voidptrfun_effect_id      : int;                        (* The id of the effect to associate with funtion 
                                                                       ptrs cast to void *)                    
        mutable ami_current_fd        : Cil.fundec;                 (* fundec of the current function being analysed *)
        ami_stmtid_to_delta_id        : (int,(int*int)) H.t;        (* Hash from stmt id to delta_in_id * delta_out_id *)
        ami_stmt_delta_in_id          : int ref;                    (* Delta id current at start of current statement *)
        ami_mallocFunHash             : Mallocpragma.mallocHash_t;  (* Hash representing set of user specified malloc functions *)
        ami_speculativeUpdateMuHash   : (int,int) H.t;              (* Hash to associate mu variables with the statement id's of 
                                                                       speculative update points *)
        ami_suspect_hash              : (string,(delta*delta)) H.t;    (* maps an function name to a constraint 
                                                                    induced by the suspect alias phase *)

        ami_intern_var                : (string, unit) H.t;         (* internal variables *)
        ami_longjmp_hash              : longjmpHash_t;              (* delta_id -> longjmp target hash *)

        (* -- Debuging information -- *)
        mutable ami_loc               : Cil.location;               (* current location in the file *)
      }
        
  type 'a mo = 'a
        
  (* Type of the monad *)
  type 'a m = mi -> 'a mo
    


  (* basic monadic operations *)
  let return (x:'a) = fun (ami:mi) -> x (* {(\* amo_delta = ami.ami_delta;  *\)amo_result = x} *)

  let (>>=) (m : 'a m) (f : 'a -> 'b m) : 'b m = 
   fun (ami : mi) -> 
      let amo = m ami in
      f amo ami
  

     


  (* -- Operations specific to this monad -- *)  

  (* The result type of running this state monad *)
  type monad_run_result = 
      {
        delta_constraints       : DeltaConstraintSet.t;  
        mu_constraints          : MuConstraintSet.t;
        delta                   : delta;
        mu                      : mu;
        stmtid_to_delta_id      : (int,(int*int)) H.t;
        funEffectHash           : funEffectHash_t;
        speculativeUpdateMuHash : (int,int) H.t;
	constNamesHash          : constNamesHash_t;
      }


  (* Standard monad glue *)
  let run 
        (mallocFunHash : Mallocpragma.mallocHash_t) 
        (effectHash : funEffectHash_t)
        (solutionHash : deltaSolutionHash_t)
        (suspectResult : Suspectalias.SAM.mo_result) 
        (internVarHash : (string, unit) H.t) 
        (m : 'a m)         = 
    let suspect_hash = suspectResult.Suspectalias.SAM.sar_fname_to_LPS_hash in
    let init_constrs = suspectResult.Suspectalias.SAM.sar_LPC_constraints in
    let ami =   { ami_delta_constraints = H.create 65535;
                  ami_mu_constraints = H.create 4099;
                  ami_statement_effects = H.create 133103;
                  ami_fwd_goto_targets = H.create 4099;
                  ami_mu = ref (MuConst true);
                  ami_delta = {delta_loc = Cil.locUnknown; delta_id = -1; delta_kind = DKVar None};
                  ami_mu_id_counter = ref 0;
                  ami_fun_effect_id_counter = ref 1;
                  ami_funEffectHash = effectHash; 
                  ami_solutionHash = solutionHash; 
                  ami_constDeltaHash = H.create 4099;
                  ami_constNamesHash = H.create 4099;
                  ami_voidptrfun_effect_id = 0;
                  ami_loc = Cil.locUnknown;
                  ami_current_fd = noFunction; 
                  ami_stmtid_to_delta_id = H.create 4099;
                  ami_mallocFunHash = mallocFunHash;
                  ami_stmt_delta_in_id = ref (-1); 
                  ami_speculativeUpdateMuHash = H.create 4099; 
                  ami_suspect_hash = suspect_hash;
                  ami_longjmp_hash = H.create 7;
                  ami_intern_var = internVarHash; } in
    let amo = m ami 
    in
      { delta_constraints = deltaConstraintHashToSet ami.ami_delta_constraints init_constrs;
        mu_constraints = muConstraintHashToSet ami.ami_mu_constraints; 
(*         delta = amo.amo_delta; *)
        delta = ami.ami_delta;
        mu = !(ami.ami_mu);
        stmtid_to_delta_id = ami.ami_stmtid_to_delta_id;
        funEffectHash = ami.ami_funEffectHash; 
        speculativeUpdateMuHash = ami.ami_speculativeUpdateMuHash;
	constNamesHash = ami.ami_constNamesHash
      }


  (* Set the current mu state to updatable (true) *)
  let amSetUpdatable = fun ami -> ami.ami_mu := MuConst true; return () ami

  (* Set the current mu state to non-updatable (false) *)
  let amSetNonUpdatable = fun ami -> ami.ami_mu := MuConst false; return () ami

  let amGetDelta = fun ami -> return ami.ami_delta ami

  let amSetDelta (d : delta) = fun ami -> ami.ami_delta <- d; return () ami

  let amGetUpdateConstraint fname =
    let find_opt h k =
      try Some (H.find h k) with Not_found -> None in
    fun ami -> return (find_opt ami.ami_suspect_hash fname) ami

  (* Generate a fresh variable and use it as the default delta *)
  let amFreshDeltaVar : delta m = 
    fun ami -> 
       let d = Deltasolver.freshDelta (DKVar None) ami.ami_loc in 
       return (d) ami

  let amMakeDeltaConst (const_name : string) (const_kind : deltaConstKind): delta m =
    fun ami ->
      let addDck () = Commonutil.hashtbl_add_once ami.ami_constNamesHash const_name const_kind in
      (* Reuse const deltas we've seen before, else generate a fresh one *)
      let delta = 
        try
          H.find ami.ami_constDeltaHash const_name
        with
            Not_found -> 
              let d = Deltasolver.freshDelta (DKConst const_name) ami.ami_loc in
                H.add ami.ami_constDeltaHash const_name d;
		addDck ();
                ignore(Updatedebug.log [Updatedebug.BTAlways] "amMakeDeltaConst '%s' : %d %d\n" const_name d.delta_id (int_of_dck const_kind));
                d
      in 
        return delta ami


  let amFreshMuVar : mu m = 
    fun ami -> 
      incr ami.ami_mu_id_counter;
      return (MuVar !(ami.ami_mu_id_counter)) ami


  let amGetNextFunEffectID : int m = 
    fun ami -> 
      incr ami.ami_fun_effect_id_counter; 
      return !(ami.ami_fun_effect_id_counter) ami

  let amAddFunEffectEntry key effect : unit m =
    fun ami ->
      H.add ami.ami_funEffectHash key effect;
      return () ami

  let amGetFunEffectEntry key : effect m = 
    fun ami ->
      return (H.find ami.ami_funEffectHash key) ami

  let amGetMu = fun ami -> return !(ami.ami_mu) ami
  let amSetMu mu = fun ami -> ami.ami_mu := mu; return () ami


  (* Add the mu constraint mu1 < mu2 *)
  let amAddMC (mu1:mu) (mu2:mu) = 
    fun ami ->
      begin
        if not (H.mem ami.ami_mu_constraints (mu1, mu2)) 
        then H.add ami.ami_mu_constraints (mu1, mu2) ();
        (*printConstraint (mu1,mu2);*)
      end;
      return () ami


  (* Add a constraint between mu1 and the current mu *)
  let amAddMCLessThanCurrent (mu1:mu) = 
    amGetMu >>= fun current_mu -> 
      amAddMC mu1 current_mu 

  let amAddMCGreaterThanCurrent (mu1:mu) = 
    amGetMu >>= fun current_mu -> 
      amAddMC current_mu mu1



  (* Add the constraint d1 \subset d2 *)
  let amAddDCSubset (d1 : delta) (d2 : delta) : unit m = 
    fun ami -> 
      begin
        if not (H.mem ami.ami_delta_constraints (false, d1.delta_id, d2.delta_id))
        then begin H.add ami.ami_delta_constraints (false, d1.delta_id, d2.delta_id) (DCSubset(d1,d2));
             let d1str,d2str = makeDeltaDebugStrings d1 d2 in 
               ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Adding delta constraint %s <= %s (%d <= %d)\n" d1str d2str d1.delta_id d2.delta_id)
        end
      end;
      return () ami


  (* Add a conditional constraint mu => d1 < d2 *)
  let amAddDCCond (mu : mu) (d1 : delta) (d2 : delta) : unit m = 
    fun ami ->
      match mu with 
          MuVar mu_id ->
            if not (H.mem ami.ami_delta_constraints (true, d1.delta_id, d2.delta_id))
            then begin H.add ami.ami_delta_constraints (true, d1.delta_id, d2.delta_id) (DCCond(mu_id,(d1,d2)));
               let d1str,d2str = makeDeltaDebugStrings d1 d2 in ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Adding cond delta constraint %s <= %s (%d <= %d)\n" d1str d2str d1.delta_id d2.delta_id) end;

            return () ami
              
        | MuConst true ->
            amAddDCSubset d1 d2 ami
              
        | MuConst false ->
            return () ami
              

  (* Add a constraint between d and the current delta: d \subseteq ami_delta *)
  let amAddDCSubsetOfCurrent (d : delta) : unit m = 
    amGetDelta >>= fun current_delta ->
    amAddDCSubset d current_delta 


  (* Add a constraint between d and the current delta: ami_delta \subseteq d *)
  let amAddDCSupersetOfCurrent (d : delta) : unit m = 
    amGetDelta >>= fun current_delta ->
    amAddDCSubset current_delta  d


  (* get effect of statement *)
  let amGetStatementEffect (sid:int) = 
    fun ami -> 
      return (H.find ami.ami_statement_effects sid) ami
        

  let amSetStatementEffect (sid:int) (d:delta) = 
    fun ami -> H.add ami.ami_statement_effects sid (d, ami.ami_delta);
      return ()

  (* save crt input delta just in case we might be the target of a goto *)
  let amAddFwdLabel (sid:int) = 
    fun ami -> 
      H.add ami.ami_fwd_goto_targets sid ami.ami_delta;
      return () ami

  (* remove a forward label marker *)
  let amRemoveFwdLabel (sid:int) = 
    fun ami -> 
      H.remove ami.ami_fwd_goto_targets sid;
      return () ami


  (* get the delta associated with a forward label;
     if not found (i.e. forward goto), return current delta;
     this will introduce a trivial constraint d <= d but we don't bother *)

  let amGetFwdDelta (sid:int) : delta m = 
    fun ami -> 
      try
        return (H.find ami.ami_fwd_goto_targets sid) ami
      with 
          Not_found -> amGetDelta ami

  (* For backward gotos we need to generate a constraint 
     target_delta <= our_delta
     Forward goto's need not be taken into account since this constraint
     comes naturally from the flow *)

  let amDoGoto (targetSid:int) : unit m =
    try
      amGetFwdDelta targetSid >>= fun current_delta ->
        amAddDCSubsetOfCurrent current_delta
    with
        Not_found -> return ()

  let amConstrainVoidPtrFunEffect : unit m =
    fun ami ->
      (*
        let m = 
      (* For all single constants (named types) k add the constraint  k \subseteq d_top.
        Associate the effect N;d_top;d_top with the predefined key for void* fun effects in the monad. *)
        amGetFunEffectEntry ami.ami_voidptrfun_effect_id
        >>= fun effect -> return (Commonutil.hashtbl_values (ref ami.ami_constDeltaHash))
      (* Doesn't matter whether we choose the in or out delta capability as they have the same delta id *)
        >>= fun (dl: delta list) -> AMOps.iter (fun (c:delta) -> amAddDCSubset c effect.effect_delta_in) dl
        in 
        m ami
      *)
      return () ami

  (* Create the effect that we use on functions cast to void* *)
  let amCreateVoidPtrFunEffect : unit m =
    fun ami ->
      let m =              
        amFreshDeltaVar       
        >>= fun d_top  -> amAddFunEffectEntry ami.ami_voidptrfun_effect_id
          {
            effect_delta_in  = d_top; 
            effect_delta_out = d_top;
            effect_labels = d_top;
            effect_mu = MuConst false
          }
      in
        m ami


  (* Return the effect id used on functions that are cast to void* *)
  let amGetVoidPtrFunEffectId : int m = 
    fun ami -> return ami.ami_voidptrfun_effect_id ami


  (* Get statistical information *)
  let amGetNumMuConstraints : int m =
    fun ami -> return (Commonutil.hashtbl_size ami.ami_mu_constraints) ami
      
  let amGetNumDeltaConstraints : int m =
    fun ami -> return (Commonutil.hashtbl_size ami.ami_delta_constraints) ami

  (* Get the hash that store the malloc function names *)
  let amGetMallocFunHash : Mallocpragma.mallocHash_t m = 
    fun ami -> return (ami.ami_mallocFunHash) ami


  (* Associate the statement id sid of a speculative update point with the current mu *)
  let amStoreSpeculativeUpdateMu (sid:int)= 
    fun ami ->
      (* Assume current mu is a variable and extract the mu_id *)
      match !(ami.ami_mu) with 
          MuVar mu_id ->
            (*ignore(Updatedebug.log [Updatedebug.BTAlways] "Adding speculative update to mu hash: %d |-> %d\n" sid mu_id);*)
            H.add ami.ami_speculativeUpdateMuHash sid mu_id;
            return () ami

        | _ -> Updatedebug.abort (Updatedebug.bug "amStoreSpeculativeUpdateMu: current mu not a var at statement %d\n" sid)
  


  (* ---- members that make this module a subtype of SyntaxMonad ---- *)

  (* Get/Set for the current location *)
  let getLoc = fun ami -> return ami.ami_loc ami
  let setLoc (loc : Cil.location) = fun ami -> ami.ami_loc <- loc; return () ami

  (* Get/Set the current function being analysed *)
  let unsetCurrentFun = fun ami -> ami.ami_current_fd <- noFunction; return () ami
  let setCurrentFun (current_fd : Cil.fundec) = fun ami -> ami.ami_current_fd <- current_fd; return () ami
  let getCurrentFun = fun ami -> return ami.ami_current_fd ami

  (* Register the start and end of statements so that we can asssociate statement id's to 
     deltas -- needed to print statement effects *)
  let amStmtStart sid = 
    fun ami -> ami.ami_stmt_delta_in_id := ami.ami_delta.delta_id; return () ami

  let amStmtEnd sid = 
    fun ami ->
      H.add ami.ami_stmtid_to_delta_id sid (!(ami.ami_stmt_delta_in_id), ami.ami_delta.delta_id);
      return () ami

  (* return the hash containing all the intern names *)       
  let amGetInternVarHash = fun ami -> return (ami.ami_intern_var) ami      

  (* add (crt_delta, target_jmp_env) pair to the longjmp hash *)
  let amAddLongjmp vi = fun ami -> 
     H.add ami.ami_longjmp_hash ami.ami_delta (vi, !(ami.ami_mu));
     return () ami

  let amGetLongjmpHash = fun ami -> return ami.ami_longjmp_hash ami     

end


(* Create a set of monadic operations for this monad *)
module AMOps = Monad.MonadOps(AnalysisMonad)


(* ---- End of analysis monad ---- *)
(*============================================================================*)




















(* ---- A specific fresh effect generator for the update analysis ---- *)

(* This module contains all of the machinery that deals with actually putting 
   effects into and out of the type grammar
*)
module AnalysisEffectManipulator =  
struct

  module M = AnalysisMonad


  (* ---- Some helper functions ---- *)

  (* Bind our monad operations locally *)
  let (>>=) = AnalysisMonad.(>>=) 
  let return = AnalysisMonad.return

  let attrIsEffect a = match a with
      Attr(aname, _) when aname = Updateconfig.effectAttrName -> true
    | _ -> false

  let getEffects = List.filter attrIsEffect
  let containsEffects = List.exists attrIsEffect

  (* Generate some fresh constraint variables and put them into the given attribute list 
     If the attr list already has effects, do nothing *)
  let genFreshFunConstraints al : (Cil.attribute list) M.m = 
    if (containsEffects al)
    then 
      (* don't insert the effects again *)
      return al
    else
      begin
   (* generate effect *)
                       AnalysisMonad.amFreshMuVar 
        >>= fun mu  -> AnalysisMonad.amFreshDeltaVar
        >>= fun d   -> AnalysisMonad.amFreshDeltaVar
        >>= fun l   -> AnalysisMonad.amFreshDeltaVar
        >>= fun d'  -> AnalysisMonad.amGetNextFunEffectID
        >>= fun key -> AnalysisMonad.amAddFunEffectEntry key {
           effect_delta_in = d; 
           effect_delta_out = d';
           effect_labels = l;
           effect_mu = mu }
        >>= fun _   -> 
          return (Cil.addAttribute (Attr(Updateconfig.effectAttrName,[AInt key])) al)
      end


  (* Extract the effect from an attribte list if it exists *)
  let extractFunConstraints al : effect M.m = 
    let getKey () =  
      match Cil.filterAttributes Updateconfig.effectAttrName al with 
          [Attr(name,[AInt key])] when name = Updateconfig.effectAttrName ->
            return key
              
        | [] -> 
            AnalysisMonad.getLoc >>=
              fun loc ->  Updatedebug.abort (Updatedebug.bug 
                          "(%a) extractFunConstraints: error extracting function constraints (no effects in this TFun)" 
                          d_loc loc)
        | _ -> 
            AnalysisMonad.getLoc >>=
              fun loc ->  Updatedebug.abort (Updatedebug.bug 
                          "(%a) extractFunConstraints: error extracting function constraints (more than one effect in this TFun)" 
                          d_loc loc)
    in   
      getKey ()
      >>= fun key -> AnalysisMonad.amGetFunEffectEntry key



  (* check whether the type contains effects *)
  let typeContainsEffects (typ : Cil.typ) : bool =
    let compHash = H.create 4099 in
    let rec typeContainsEffects_inner (typ_inner : Cil.typ) : bool =
      match typ_inner with 
          TVoid _ | TInt _  | TFloat _ | TEnum _ | TBuiltin_va_list _ -> false
            
        | TPtr (t, _) -> typeContainsEffects_inner t
        | TArray(t, _, _) -> typeContainsEffects_inner t
            
        | TFun(tret, targs, isva, al) -> containsEffects al
            
        | TNamed(ti, _) -> typeContainsEffects_inner ti.ttype
            
        | TComp(ci,al) ->  
            if H.mem compHash ci.cname 
            then 
              false
            else 
              begin
                H.add compHash ci.cname true; 
                let field_has_effects (fi : Cil.fieldinfo) : bool = typeContainsEffects_inner fi.ftype in
                  List.exists field_has_effects ci.cfields
              end
    in
      typeContainsEffects_inner typ


  (*----------------------------------------------------------------------------*)
  (* ---- Unused utility functions ---- *)

  (* add effect attr to current type ; die if type already has effects *)
  let addEffectToType (t:Cil.typ) (attr:Cil.attribute) : Cil.typ = 
    (* add recursively behind pointers *)
    let addEffectToType' (t':Cil.typ) (attr:Cil.attribute) : Cil.typ = 
      let ta = Cil.typeAttrs t' in
        if containsEffects ta
        then
          t'
            (*Updatedebug.abort (Updatedebug.bug "addEffectToType : tried to add 2nd effect to type %a\n" d_plaintype t')*)
        else
          begin
            Cil.typeAddAttributes [attr] t'
          end
    in
      match t with
          TPtr(t',a) -> TPtr(addEffectToType' t' attr,a)
        | _ -> addEffectToType' t attr


  (* ---- End of unused utility funcitons ---- *)
  (*----------------------------------------------------------------------------*)



  (* ---- Required functions so we are an instance of Effectinserter.EffectCreator ---- *)
  
  let insertEffectInOuterConstructor (typ:Cil.typ) : Cil.typ M.m =
    match typ with 
        TFun(tret, targs, isva, al) -> 
          genFreshFunConstraints al >>= fun al' -> return (TFun(tret, targs, isva, al'))

      | _ -> return typ
  
end




(* Create a fresh effect inserter for the update analysis *)
module AnalysisGenFreshEffects = Effectinserter.EffectInserter(AnalysisEffectManipulator)(AnalysisMonad)

 


(* ---- End of effect insertion ---- *)
(*============================================================================*)










(* Locally bind the analysis monad basic operations *)
let (>>=) = AnalysisMonad.(>>=)
let return = AnalysisMonad.return

(* shortcur to analysis monad module *)
module AM = AnalysisMonad







         
(* When we reach a statement that was the target of previous gotos, then we should 
   add the constraints generated by these gotos *)
let rec purgeGotoConstraints (sid:int) : unit AM.m = 
   try
                    AM.amGetFwdDelta sid
      >>= fun d  -> AM.amAddDCSupersetOfCurrent d
      >>= fun _  -> AM.amRemoveFwdLabel sid
      >>= fun _  -> purgeGotoConstraints sid
   with 
      Not_found -> return ()




(* given a type, do a user-specified action on all fun effects in this type *)
let doActionOnFunEffect (typ : Cil.typ) (action : effect -> unit AM.m ) : unit AM.m =
   let compHash = H.create 4099 in
   let rec doActionOnFunEffect' (typ' : Cil.typ) : unit AM.m = 
      match typ' with 
         TVoid _
       | TInt _
       | TFloat _ -> return ()
       | TPtr (t,attr) -> doActionOnFunEffect' t
            
       | TArray(t,e_opt,attr) -> doActionOnFunEffect' t
            
       | TFun(tret,targs,isva,al) ->
            let args_map (s,t,a) = doActionOnFunEffect' t
            in 
                 begin match targs with 
                         Some l -> AMOps.iter args_map l 
                       | None   -> return ()
                      end
            >>= fun ()             -> doActionOnFunEffect' tret
            >>= fun ()             -> AnalysisEffectManipulator.extractFunConstraints al
            >>= fun eff            -> action eff

       | TNamed _ -> return ()
       
       | TComp(ci,al) ->  
           if H.mem compHash ci.cname then return ()
             else
                   begin
              H.add compHash ci.cname true; 
              let field_map (fi : Cil.fieldinfo) : unit AM.m = 
                                    doActionOnFunEffect' fi.ftype
                   in
              AMOps.iter field_map ci.cfields 
              end

       | TEnum _
       | TBuiltin_va_list _ -> return ()
   in 
   doActionOnFunEffect' typ;;

(* Add constraints to force all effects on function types to be N;Top;Top (the void fun ptr effect in the monad) *)
let pacifyFunTypeEffects (typ : Cil.typ) : unit AM.m =
   let action = 
                fun eff            -> AM.amAddMC eff.effect_mu (MuConst false)
            >>= fun ()             -> AM.getLoc
            >>= fun loc            -> return (ignore (Updatedebug.log [Updatedebug.BTAnalysis] "(%a) Pacifying function\n" d_loc loc ))

(*
            >>= fun eff            -> AM.amGetVoidPtrFunEffectId
            >>= fun voidfun_eff_id -> AM.amGetFunEffectEntry voidfun_eff_id
            
            (* Add equality conctraint Top = delta_out *)
            >>= fun voidfun_eff    -> AM.amAddDCSubset eff.effect_delta_out voidfun_eff.effect_delta_out   
            >>= fun ()             -> AM.amAddDCSubset voidfun_eff.effect_delta_out eff.effect_delta_out
            
            (* Add equality conctraint Top = delta_in *)
            >>= fun ()             -> AM.amAddDCSubset eff.effect_delta_in voidfun_eff.effect_delta_in
            >>= fun ()             -> AM.amAddDCSubset voidfun_eff.effect_delta_in eff.effect_delta_in
            
            >>= fun ()             -> AM.amAddMC eff.effect_mu voidfun_eff.effect_mu
*)
   in
      doActionOnFunEffect typ action;;


(*============================================================================*)
(* Arguments to extern functions *)


(* All arguments in exp_arg_list which have function type are forced to have their output
   effect a superset of the external function output effect. This ensures proper effect propagation
   by assuming once we pass a function pointer 'f' to an external function, the external function
   will call 'f'
*)
let constrainArgumentsToExternalFunctions externFunEffect exp_arg_list = 

   let action = fun argumentEffect  -> AM.amAddDCSubset externFunEffect.effect_delta_out argumentEffect.effect_delta_out in
   let argument_types = List.map Updatecilutil.properTypeOf exp_arg_list in
   let constrain_one_argument t = doActionOnFunEffect t action in
   AMOps.iter constrain_one_argument argument_types


(* set res to true is an extern var was found *)
class externVarMarker internVarHash (res : bool ref) : cilVisitor = object(self)
  inherit nopCilVisitor

 (* we need to test both Extern and internVarHash to avoid
    - local vars
    - extern decls (definitions from the original program) in patch mode
 *)    
  method vvrbl v = 
     (if v.vstorage = Extern && not (H.mem internVarHash v.vname)
      then
         res := true;
     );
     DoChildren

end

(* given an expresion, return true if it contains an external var *)
let isExternalFunCall (funExpr : Cil.exp) (internVarHash : (string, unit) H.t) = 
   let res = ref false in
   let visitor = new externVarMarker internVarHash res in
   ignore(Cil.visitCilExpr visitor funExpr); (* read-only visitor, so we don't bother with the result *)   
   !res

(* End arguments to extern functions *)
(*============================================================================*)

(*         turn a T[] into a T              *)
let discardArrayType t = match t with
   TArray(ta,_,_) -> ta
 | _ -> t 


module AnalysisSubtypeActions = 
struct
  module M = AnalysisMonad
  module MOps = Monad.MonadOps(AnalysisMonad)

  (* Called when SubtypeCmp performs a standard subtype step *)
  let subtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    match t1,t2 with 
      | TFun(t1',args_opt1,isVA1,attr1), TFun(t2',args_opt2,isVA2,attr2) ->  
          AnalysisEffectManipulator.extractFunConstraints attr1 >>= fun effect1 -> 
          AnalysisEffectManipulator.extractFunConstraints attr2 >>= fun effect2 -> 
          M.amAddDCSubset effect2.effect_delta_out effect1.effect_delta_out >>= fun _ ->

          (* Iulian : why was it the case the t_1 <: t_2 => mu_2 <= mu_1 ?
             shouldn't it be mu_1 <= mu_2
             
             I mean, if t_2 is inferred to be N, t_1 should be N and 
             if t_2 is unconstrained, then t_1 is unconstrained  
             
          *)               
          (*  OLD:           >>= fun _ ->  AM.amAddMC effect2.effect_mu effect1.effect_mu *)
          M.amAddMC effect1.effect_mu effect2.effect_mu
   
          (* Function types:   

             t1 ------ {S1} ----> t1' <:  t2 ------ {S2} -----> t2' 

             if S1 \subseteq S2  
          *)     
          >>= fun _ ->  AM.amAddDCSubset effect1.effect_labels effect2.effect_labels


      | _ -> return ()



  (* Called when a physical subtype step is performed, i.e. a pointer
     to a record is a subtype of a pointer to the records first field
   *)
  let physicalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    return ()
    
  (* Called when one of the types is external *)
  let externalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    pacifyFunTypeEffects t1 >>= fun () -> 
    pacifyFunTypeEffects t2 >>= fun () ->
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: One type is external %s and %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  (* Called when trying to determine a subtype between comp pointer type
     and another type where the types neither lead to a valid subtype or to a valid 
     physical subtype *)
  let ptrIncompatible (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m =
    pacifyFunTypeEffects t1 >>= fun () -> 
    pacifyFunTypeEffects t2 >>= fun () ->
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: Incompatible pointers %s and  %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  
  (* Called when C would not accept the subtype *)
  let invalidCSubtype (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m =
    pacifyFunTypeEffects t1 >>= fun () -> 
    pacifyFunTypeEffects t2 >>= fun () ->
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: invalid C subtype %s and %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  let toVoidPtr t1 t2 = pacifyFunTypeEffects t1
  let fromVoidPtr t1 t2 = pacifyFunTypeEffects t2

end


module AnalysisSubtypeConstraints = Subtype.SubtypeCmp(AnalysisSubtypeActions)(AnalysisMonad)

let subTypeConstraints t1 t2 = AnalysisSubtypeConstraints.subtypeTraverse t1 t2


(* For introducing constraints on update points, called from analysis.ml *)

let suspConstr =
  AM.getCurrentFun >>= fun fd ->
  AM.amGetUpdateConstraint fd.svar.vname >>= fun constr ->
    begin match constr with
      Some (lps,locsG) -> 
   ignore (Updatedebug.log [Updatedebug.BTSuspect] 
        "introducing update constraint: func(%s) ===> %d <= curr, %d <= curr\n"
        fd.svar.vname lps.delta_id locsG.delta_id);
   AM.amAddDCSubsetOfCurrent lps  >>= fun ()   -> (* L <= current_d *)
   AM.amAddDCSubsetOfCurrent locsG    (* locs(G) <= current_d *)
    | None -> 
   ignore (Updatedebug.log [Updatedebug.BTSuspect]
        "missing update constraint: %s\n" fd.svar.vname);
   return ();
    end

(*============================================================================*)
(* ---- Traversal of syntax to gather constraints ---- *)


let rec analyzeLval ((lhost,offset) : Cil.lval) : unit AM.m = 
   match lhost with 
      Mem (e) -> analyzeExp e
    | _       -> return ()



and analyzeExp (exp : Cil.exp) : unit AM.m = 
   (*amGetNumDeltaConstraints >>= fun num_mu ->*)
   let id = return () in
(*    let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "analyzeExp %a:%s\n " d_plainexp exp (Updatedebug.ppType (Updatecilutil.properTypeOf exp))) in *)
   match exp with
        Const _ -> id
        
      | Lval lval -> analyzeLval lval
         
      | SizeOf typ -> id   
                           
      | SizeOfE e -> analyzeExp e
         
      | SizeOfStr str -> id

      | AlignOf typ ->
         AM.getLoc >>=
         fun loc -> ignore(Updatedebug.log [Updatedebug.BTAlways] "WARNING : analyzeExp : (%a) can't deal with AlignOf(%a)!"
                                      d_loc loc d_plaintype typ);
                    id
      
      | AlignOfE exp ->
         AM.getLoc >>= 
         fun loc -> ignore(Updatedebug.log [Updatedebug.BTAlways] "WARNING : analyzeExp : (%a) can't deal with AlignOfE(%a)!"
                                       d_loc loc d_plainexp exp); 
                    id             
                    
      | UnOp (unop, e, typ) -> analyzeExp e

      (* FIXME: C doesn't say which order these get evaluated in!  We need to be more conservative *)
      | BinOp (binop, e1, e2, typ) -> analyzeExp e1 >>= fun _ -> analyzeExp e2
         
      | CastE (typ, e) -> 
                          return (Updatecilutil.properTypeOf e)
           >>= fun e_t -> subTypeConstraints e_t typ
           >>= fun ()  -> analyzeExp e

      | AddrOf lval 
      | StartOf lval -> analyzeLval lval


and analyzeCall (instr : Cil.instr) : unit AM.m = 

   AM.amGetMallocFunHash >>= fun mallocFunHash ->
   match instr with 

    (* we ignore the arguments to these C-magic-calls *)
    |  Call(_, Lval(Var vi, NoOffset), _, _) when  Commoncilutil.isBuiltin vi.vname -> 
        return ()

    (* When con *)
    | Call(Some(l), (Lval(Var vi, NoOffset) as f), [e], _) 
      when  (Updateconfig.isConFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname)  ->
         let t = Namedtypes.getAbstractTypeByConAbsName vi.vname in 
                   let type_name = Updatecilutil.getAbstractTypeName t  in 
                analyzeExp e
            >>= fun _ -> subTypeConstraints (Updatecilutil.properTypeOf e) (Updatecilutil.properTypeOfLval l)
            >>= fun _ -> AM.amMakeDeltaConst type_name DckType >>= AM.amAddDCSubsetOfCurrent

    (* when abs *)
    | Call(None, (Lval(Var vi, NoOffset) as f), [e; Lval(l)], _) 
      when  (Updateconfig.isAbsFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname) ->
         let t = Namedtypes.getAbstractTypeByConAbsName vi.vname in
                       analyzeExp e
         >>= fun () -> analyzeLval l
         (* push the effects of the thing being abs onto the type of the abstract type *)
         >>= fun () -> subTypeConstraints (Updatecilutil.properTypeOf e) t
    
    
    (* A speculative update function *)
    | Call(_, Lval(Var vi, NoOffset), _, _) when  Updateconfig.isSpeculativeUpdateFunc vi.vname -> 
         (* This is should have matched in analyzeStmt *)
                       AM.getLoc
         >>= fun loc -> Updatedebug.abort (Updatedebug.bug  "(%a) analyzeCall: hit a speculative update point;it should have been hit in analyzeStmt not here.\n" d_loc loc)

    
    (* A real update function *)
    | Call(_, Lval(Var vi, NoOffset), _, _) when  Updateconfig.isRealUpdateFunc vi.vname -> 
                       AM.amFreshDeltaVar                         (* d fresh *)
         >>= fun d  -> AM.amAddDCSubsetOfCurrent d                (* d <= current_d *)
         >>= fun () -> AM.amSetDelta d                            (* current_d := d *)
         >>= fun () -> AM.amAddMCLessThanCurrent (MuConst true)   (* U < current_mu  *)
         (* Add constraints to do with suspect aliases *)
         >>= fun () -> suspConstr 

    (* Ignore without any guilt mallocs *)
    | Call(_, Lval(Var vi, NoOffset), _, _) when  Mallocpragma.isMallocFunc mallocFunHash vi.vname -> 
         return ()

    (* Handle constraints for setjmp/longjmp *)        
    | Call(_, Lval(Var vi, NoOffset), _, _) when  Setjmp.isSetjmpFunction vi.vname -> 
                        AM.getLoc
         >>= fun loc -> let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "(%a) setjmp-style function '%s' called\n" d_loc loc vi.vname) in

         return ()

    | Call(_, Lval(Var vi, NoOffset), e :: _, _) when  Setjmp.isLongjumpFunction vi.vname -> 
                        AM.getLoc
         >>= fun loc -> ignore (Updatedebug.log [Updatedebug.BTAlways] "(%a) longjmp-style function '%s' called\n" d_loc loc vi.vname);
                        AM.amAddLongjmp (Commoncilutil.getBaseVarinfo e)

   
    (* standard function call *)
    | Call (lval_result_opt, exp_func, exp_arg_list, loc) ->
         let (exp_func_ret_t, 
              exp_func_args_t, 
              isva)                 = Updatecilutil.getTypesInFunExp exp_func in
         let numFixedArgs           = List.length exp_func_args_t in 
         let numActualArgs          = List.length exp_arg_list in
         
         (* Make two different lists for fixed and variable arguments *)
         let ( fixedArgs,
               varArgs )   = Commonutil.list_split numFixedArgs exp_arg_list in 

         (* Create a list of pairs of types (t1,t2) for the fixed arguments that we need to create
            subtype constraints between *)
         let fixed_args_src_target_types = List.combine (List.map Updatecilutil.properTypeOf fixedArgs) exp_func_args_t in
    
              
              (* Do standard function stuff *)
         (* FIXME: is this the right evaluation order for C? *)
                      analyzeExp exp_func 
         >>= fun _ -> AMOps.iter analyzeExp exp_arg_list 
         >>= fun _ -> 
                begin match lval_result_opt with 
                    None      -> return ()
                  | Some lval -> 
                       (* Analyse the lval and collect constraints on the assignment of the result to the lval *)
                                         analyzeLval lval
                       >>= fun ()     -> return (Updatecilutil.properTypeOfLval lval)
                       >>= fun lval_t -> subTypeConstraints exp_func_ret_t lval_t
                     end
         
         >>= fun _ -> AMOps.iter (fun (t1,t2) -> subTypeConstraints t1 t2) fixed_args_src_target_types 
         >>= fun _ -> AM.amFreshDeltaVar
         >>= fun d -> AM.amAddDCSubsetOfCurrent d
         >>= fun _ -> return (Updatecilutil.properTypeOf exp_func)
         >>= fun exp_func_t -> 
                    begin match (Cil.unrollType exp_func_t) with 
                       TFun(_,_,_,al) -> AnalysisEffectManipulator.extractFunConstraints al
            
                     | _              -> Updatedebug.abort (Updatedebug.bug
                                    "analyzeCall: expected function type, but got:\n %a\n" d_plaintype exp_func_t)
                    end
       
         (* Add the conditional constrain  effect_mu = U => d < delta_out *)
         >>= fun effect -> AM.amAddDCCond effect.effect_mu d effect.effect_delta_out

         (* Constrain function addresses passed to external functions *)
         >>= fun () -> AM.amGetInternVarHash 
         >>= fun internVarHash -> 
                     begin match isExternalFunCall exp_func internVarHash with
                          true -> constrainArgumentsToExternalFunctions effect exp_arg_list
                        | false -> return ()
                     end

         (* Function types : for this code

            bar() 
            {
               ...
               D  foo();  D'
               ...
            }

          where foo has type t1 ------- { S1 } -------> t1'

          Then add then constraint : S1 \subseteq D    

          *)        

         >>= fun ()           -> AM.amAddDCSubsetOfCurrent effect.effect_labels 
         
         (* If the callee can do an update then we're unconstrained 
            If we're N, the callee must be N too 
            hence the constraint mu_callee <= mu_caller     *)

         >>= fun ()     -> AM.amAddMCLessThanCurrent effect.effect_mu

         >>= fun ()     -> AM.amSetDelta d
       
    | _ -> Updatedebug.abort (Updatedebug.bug "analyzeCall : no way can be here")

and analyzeInstr (instr : Cil.instr) : unit AM.m = 

                AM.setLoc (Cil.get_instrLoc instr) 
   (* add all gvars used in this instr to the current capability *)
   >>= fun _ -> let gvarsUsed = Commoncilutil.getGlobalVarsUsedInInstr instr in
                let addOneGvar_iter gvarName = 
                                 AM.amMakeDeltaConst gvarName DckGvar
                   >>= fun ss -> AM.amAddDCSubsetOfCurrent ss
                in 
                AMOps.iter addOneGvar_iter gvarsUsed


   >>= fun _ -> 
      match instr with 
            Set (lval,exp,loc) -> 
                                    analyzeLval lval 
                  >>= fun ()     -> analyzeExp exp 
                  >>= fun ()     -> return (Updatecilutil.properTypeOfLval lval)
                  >>= fun lval_t -> return (Updatecilutil.properTypeOf exp)
                  >>= fun e_t    -> subTypeConstraints e_t lval_t
         
          | Call _ -> analyzeCall instr
         
          | Asm _ ->
                              AM.getLoc 
               >>= fun loc -> ignore(Updatedebug.log [Updatedebug.BTAnalysis] 
                                        "(%a) WARNING! analyzeInstr: can't deal with Asm at instr level, but you used it!\n\n" 
                                        d_loc loc);
                               return ()


and analyzeStmt (stmt : Cil.stmt) : unit AM.m = 
   let id = return () in 
   AM.setLoc (Cil.get_stmtLoc stmt.skind) >>= fun _ ->
   (* Register the start of the statment, so that the current delta ID can be associated with the stmt id *)
   AM.amStmtStart stmt.sid >>= fun _ ->
   AM.amAddFwdLabel stmt.sid >>= fun _ ->
   begin match stmt.skind with 
    
    (* A speculative update function *)
    | Instr [ Call(_, Lval(Var vi, NoOffset), _, loc) ] when  Updateconfig.isSpeculativeUpdateFunc vi.vname -> 
                       AM.amFreshDeltaVar                         (* d fresh *)
         >>= fun d  -> AM.amAddDCSubsetOfCurrent d                (* d <= current_d *)
         >>= fun () -> AM.amSetDelta d                            (* current_d := d *)
         >>= fun () -> AM.amStoreSpeculativeUpdateMu stmt.sid     (* Associate a speculative update point with the current mu *)
         (* We don't constrain mu to be U as this is a speculative udpdate *)
    >>= fun () -> suspConstr

   
    | Instr  il ->     AMOps.iter analyzeInstr il

       
       (* FIX? Seems as if you would have to constraint the output
           effect of the function with the current capability.  On
           some thought, though, it seems as if the capability at the
           conclusion of the block will be more restrictive, so it's
           not actually needed here. *)
   
    | Return (None,loc) -> id
    
    | Return (Some e, loc) ->
                          analyzeExp e  
         >>= fun ()    -> AM.getCurrentFun 
         >>= fun fd    -> return (Updatecilutil.getReturnType fd)
         >>= fun ret_t -> return (Updatecilutil.properTypeOf e)
         >>= fun e_t   -> subTypeConstraints e_t ret_t
       
    | Goto (target, loc) ->
         AM.amDoGoto !target.sid
       
    | Break loc -> 
         Updatedebug.abort (Updatedebug.bug
                           "(%a) analyzeStmt: CIL should have eliminated break!\n" 
                           d_loc loc)
    
    | Continue loc -> 
         Updatedebug.abort (Updatedebug.bug
                           "(%a) analyzeStmt: CIL should have eliminated continue!\n" 
                           d_loc loc)      
    
    | If (eTest,blockT,blockF,loc) ->
                        analyzeExp eTest 
              >>= fun _   -> AM.amGetDelta 
              >>= fun d_e -> analyzeBlock blockT
              >>= fun _   -> AM.amGetDelta
              >>= fun d_T -> AM.amSetDelta d_e
              >>= fun _   -> analyzeBlock blockF
              >>= fun _   -> AM.amGetDelta
              >>= fun d_F -> AM.amFreshDeltaVar
              >>=            AM.amSetDelta
              >>= fun _   -> AM.amAddDCSupersetOfCurrent d_T
              >>= fun _   -> AM.amAddDCSupersetOfCurrent d_F
      
    | Switch (exp, block,stmt_list,loc) ->
        Updatedebug.abort (Updatedebug.bug 
                          "(%a) analyzeStmt: CIL should have eliminated switch!\n" 
                                    d_loc loc)
       
    | Loop (block,loc,s1,s2) -> analyzeBlock block
  
    | Block block -> analyzeBlock block
       
    | TryFinally  (block1,block2,loc) ->
       Updatedebug.abort (Updatedebug.bug 
                         "(%a) analyzeStmt: M$ feature -- as if we support that!\n" 
                         d_loc loc)
  
    | TryExcept (block_try, inst_list, block_except, loc) ->
       Updatedebug.abort (Updatedebug.bug 
                         "(%a) analyzeStmt: M$ feature -- as if we support that!\n"
                                   d_loc loc)
   end
   (* Associate the current delta with the output capability of this statement *)
   >>= fun _ -> AM.amStmtEnd stmt.sid




and analyzeBlock (b : Cil.block) : unit AM.m =
   AMOps.iter analyzeStmt b.bstmts 
      
  



(* Collect constraints for a function *)
let analyzeFunction (f : Cil.fundec) : unit AM.m =

  let _ = ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Analyzing function %s ...\n" f.svar.vname) in

       AM.setCurrentFun f
   >>= fun _ ->
                match f.svar.vtype with 
                   TFun(_,_,_,al) ->
                                        AnalysisEffectManipulator.extractFunConstraints al
                           >>= fun effect -> AM.amSetDelta effect.effect_delta_in
                           >>= fun _      -> AM.amSetMu effect.effect_mu

                           (* Function types, i.e. fun foo : t1 ---- {foo} -----> t2*)   
                           >>= fun _      -> AM.amMakeDeltaConst f.svar.vname DckFun
                           >>= fun funNameSingletonSet  -> AM.amAddDCSubset funNameSingletonSet effect.effect_labels

                           >>= fun _      -> analyzeBlock f.sbody
                           (* add equality constraint to force output delta on function to be the same as
                                        the delta we ended up with *)
                           (*>>= fun _      -> AM.amAddDCSupersetOfCurrent effect.effect_delta_out*)
                           >>= fun _      -> AM.amGetDelta 
                           >>= fun d      -> ignore(Updatedebug.log [Updatedebug.BTAnalysis] "analyzeFunction(%a): %d\n" d_loc d.delta_loc d.delta_id);
                                             AM.amAddDCSubsetOfCurrent effect.effect_delta_out  
                      
                 | _ -> Updatedebug.abort (Updatedebug.bug
                        "analyzeFunction: expected function type, but got:\n %a\n" d_plaintype f.svar.vtype)

   >>= fun _ -> AM.unsetCurrentFun 


(* The following three functions take care of analyzing static initializers *)

let rec analysis_foldLeftCompound (ct: Cil.typ) (initl: (Cil.offset * Cil.init) list) : unit AM.m = 
  let bail s = Updatedebug.abort (Updatedebug.bug "analysis_foldLeftCompound %s" s) in
  let off_l, init_l = List.split initl in
  match unrollType ct with
    TArray(bt, _, _) ->
       AMOps.iter (analyzeInit bt) init_l

  | TComp (comp, _) as tc -> 
       let getTypesForOffset = function
          Field(f, NoOffset) -> (f.ftype, (Commoncilutil.getFieldinfoByName tc f.fname None).ftype)
        | _ -> bail "analysis_foldLeftCompound: malformed initializer"
       in
       let doOneField ((o, i) : (Cil.offset * Cil.init)) : unit AM.m = 
         let field_t,comp_field_t = (getTypesForOffset o) in 
                           subTypeConstraints field_t comp_field_t
            >>= fun () ->  analyzeInit field_t i
       in
       AMOps.iter doOneField initl 

  | _ -> bail "Type of Compound is not array or struct or union"


and analyzeInit (var_t : Cil.typ) (ii : Cil.init) : unit AM.m = 
   begin match ii with
      SingleInit e -> let exp_t = Updatecilutil.properTypeOf e in
                         analyzeExp e
         >>= fun ()   -> subTypeConstraints exp_t var_t
  
    | CompoundInit (t,ol) -> 
                         analysis_foldLeftCompound t ol
         >>= fun ()   -> subTypeConstraints t var_t
   end

let analyzeInitinfo (ii : Cil.initinfo) (var_t : Cil.typ ) : unit AM.m = 
   begin match ii.init with 
      Some i -> 
                       analyzeInit var_t i
         >>= fun () -> return ()
    
    | None   ->        return ()
   end


(* let crtGlobal = ref 0;; *)

let analyzeGlobal (g : Cil.global) : unit AM.m = 
   (* AMOps.lift print_string (return ("At global" ^ (incr crtGlobal; string_of_int (!crtGlobal)) ^ "\n")) >>= fun () -> *)
   (match g with 
        GFun(fd,loc) -> 
           if Namedtypes.isConAbsFuncName fd.svar.vname then
              return ()
           else 
              analyzeFunction fd

       (* analysis-wise, a global init 
          var = exp 
          has to be regarded as a Set(var,exp) 
          We simulate this by forcing a subtype relationship:
               exp_t <: var_t
       *)

      | GVar(vi, ii, loc) -> 
                                AM.setLoc loc 
              >>= fun ()     -> analyzeInitinfo ii vi.vtype 

      | _            -> return ())

let analyzeFile (file : Cil.file) : unit AM.m =
   AMOps.iter analyzeGlobal file.globals


(* ---- End of traversal of syntax to gather constraints ---- *)
(*============================================================================*)

(* add interprocedural 'goto' constraints due to setjmp/longjmp *)
let purgeLongjmpConstraints (setjmpHash : Setjmp.setjmpHash_t)  : unit AM.m = 

   let purgeLonjmp_iter (longjmpDelta, (env_vi,mu)) =  
   (* in a function with updateability 'mu' , 
      at statement 'delta', we had a 
      longjmp(env,val) 

      now look for statements 
      D;     setjmp(env) ; D'
      and add 
      D' <= delta 
   *)
      try
         let setjmpSid = H.find setjmpHash env_vi in
            begin
               try
                                       AM.amGetFwdDelta setjmpSid
                >>= fun setjmpDelta -> ignore(Updatedebug.log [Updatedebug.BTAnalysis] "purgeLonjmp_iter : adding longjmp -> setjmp constraint %d <= %d \n " setjmpDelta.delta_id longjmpDelta.delta_id);
                                       AM.amAddDCSubset setjmpDelta longjmpDelta
                     
               with Not_found -> 
                  Updatedebug.abort (Updatedebug.bug "purgeLongjmpConstraints: can't find sid=%d in fwdDelta \n" setjmpSid)         
            end
      with Not_found -> ignore(Updatedebug.log [Updatedebug.BTAlways] " WARNING : purgeLonjmp_iter : can't find destination for '%s' in setjmpHash\n " env_vi.vname);
                        return ()
(*          Updatedebug.abort (Updatedebug.bug "purgeLongjmpConstraints: can't find destination for '%s' in setjmpHash\n" *)
(*                            env_vi.vname)          *)

   in
                              AM.amGetLongjmpHash
   >>= fun longjmpHash     -> let longjmp_hash_contents = Commonutil.hashtbl_key_val_pairs longjmpHash in
                              AMOps.iter purgeLonjmp_iter longjmp_hash_contents;;


(*============================================================================*)



(* Filter statement effects based on the user-specifed effect printing policy *)
let filterPrintedEffects (delta_set : CapabilitySet.t) : CapabilitySet.t = 

   (* this tells us whether we're dealing with a type or function name*)
   let isTypeName (name : string) : bool = 
      try
         ignore (Namedtypes.NamedTypeInfo.isUpdatable name);
         true
      with 
         Not_found -> false
   in
   let notIsTypeName = fun x -> not (isTypeName x) in
   match Updateargs.args.Updateargs.arg_effectPrintingPolicy with
      Updateargs.EPFull      -> delta_set
    | Updateargs.EPTypes     -> CapabilitySet.filter isTypeName delta_set
    | Updateargs.EPFunctions -> CapabilitySet.filter notIsTypeName delta_set
    | Updateargs.EPNone      -> CapabilitySet.empty;;
   




(*============================================================================*)
(* ---- Printing analysis effects ---- *)


(* install a hook into cil to print capability sets with each statement *)
let installPrintEffectsHook (solution : (int,CapabilitySet.t) H.t) (stmtid_to_delta_id : (int,(int*int)) H.t)  =  
   let hook sid = 
      (* print types/function labels only if they're specified in the current effectPrintingPolicy *)   
      if Updateargs.args.Updateargs.arg_effectPrintingPolicy <> Updateargs.EPNone 
      then 
         begin  
            try 
               let (d_in_id, d_out_id) = H.find stmtid_to_delta_id sid in
               let d_out_set = H.find solution d_out_id in
               let d_in_set  = H.find solution d_in_id in 
                  "/* D(" ^ (debug_string_of_int d_in_id) ^ "): " ^ (printDelta (filterPrintedEffects d_in_set)) ^ "   *** D'(" ^ (debug_string_of_int d_out_id) ^ "): " ^ (if d_out_id = d_in_id then "NO CHANGE" else (printDelta (filterPrintedEffects d_out_set))) ^ " */"
            with 
                  Not_found -> "/* NOEFFECT */" (*Updatedebug.abort (Updatedebug.bug "PrintEffectsHook: could not find effect for statement %d!\n" sid)*)
         end
      else
         ""
   in 
   Cil.setStmtPrintEffectHook hook


(* ---- insert function effects as comments ---- *)
let annotateWithFunEffects (funEffectHash : funEffectHash_t) 
                           (funTypeHash : funTypeHash_t) 
                           (solution : (int,CapabilitySet.t) H.t) 
                           (mu_solution : (int,bool) H.t) 
                           (fname_to_LPS_hash : (string,delta*delta) H.t)
                           (file:Cil.file) 
                           : unit = 
                           
   let string_of_mu (mu:mu) : string = 
      let ustring_of_bool b = if b then "U" else "N" in
      match mu with 
        MuConst b -> ustring_of_bool b
      | MuVar id  -> 
           try
              (debug_string_of_int id) ^ ":" ^ (ustring_of_bool (H.find mu_solution id))
           with 
              Not_found -> (debug_string_of_int id) ^ ":" ^ (ustring_of_bool true) (*Updatedebug.abort (Updatedebug.bug "annotateWithFunEffects:  function has mu var %d that is not in the solution!\n" id)*)
   in
   
   let annotate_fold (acc:global list) (g:global) = 
      match g with 
         (* only annotate user functions -- Note the definition for update is not added yet, so we don't have to exclude it *)
         GFun(fd,loc) when 
            (not (Namedtypes.isConAbsFuncName fd.svar.vname) )  ->
 
            begin match AnalysisEffectManipulator.getEffects (Cil.typeAttrs fd.svar.vtype) with
               [Attr(attrName,[AInt key])] ->
                  begin try
                     let () = H.add funTypeHash fd.svar.vname fd.svar.vtype in (* save fun type now, for the next step in runAnalysis is gonna remove the attrs from the fun type *)
                     if (Updateargs.args.Updateargs.arg_effectPrintingPolicy <> Updateargs.EPNone)
                     then begin
                        let effect = H.find funEffectHash key in
                        let d_in_id  = effect.effect_delta_in.delta_id in  
                        let d_out_id = effect.effect_delta_out.delta_id in  
                        let d_in_set = H.find solution d_in_id in 
                        let d_out_set = H.find solution d_out_id in
                        let labels_id = effect.effect_labels.delta_id in
                        let labels_set = H.find solution labels_id in
                           (* print types/function labels only if they're specified in the current effectPrintingPolicy *)   
                        let do_susp = Updateargs.args.Updateargs.arg_dosuspect in
                        let types_string = 
                           let (lps_id,locsG_id,lps_set,locsG_set) = 
                              if do_susp then
                                 let (l,locsG) = H.find fname_to_LPS_hash fd.svar.vname in
                                    (l.delta_id,locsG.delta_id,
                                     H.find solution l.delta_id,
                                     H.find solution locsG.delta_id)
                              else (0,0,CapabilitySet.empty,CapabilitySet.empty) 
                           in
                              match Updateargs.args.Updateargs.arg_effectPrintingPolicy with

                                    Updateargs.EPFull 
                                 | Updateargs.EPTypes ->  "  D(" ^ (debug_string_of_int d_in_id) ^ "): { " ^ (printDelta (filterPrintedEffects d_in_set)) ^ " }" ^
                                      "   *** D'(" ^ (debug_string_of_int d_out_id) ^ "): " ^ 
                                          (if d_out_id = d_in_id then "NO CHANGE" 
                                           else ("{ " ^ (printDelta (filterPrintedEffects d_out_set)) ^ " }") ) ^
                                          (if do_susp then
                                              " L(" ^ (debug_string_of_int lps_id) ^ "): { " ^ (printDelta (filterPrintedEffects lps_set)) ^ " }  G("^ (debug_string_of_int locsG_id) ^ "): { " ^ (printDelta (filterPrintedEffects locsG_set)) ^ " }"
                                           else "")
                                 | _ -> ""
                                      
                        and funlabels_string = match Updateargs.args.Updateargs.arg_effectPrintingPolicy with

                              Updateargs.EPFull 
                           | Updateargs.EPFunctions -> " Labels(" ^ (debug_string_of_int labels_id) ^ "): { " ^ (printDelta (filterPrintedEffects labels_set)) ^ " }"

                           | _ -> ""       
                        in
                        let fun_header_string =  "/* Mu:" ^ (string_of_mu effect.effect_mu) ^ 
                           types_string ^
                           funlabels_string ^
                           " */"  
                        in
                           acc @ [GText( fun_header_string );  
                                  g]
                     end
                     else
                        acc @ [g]
                  with 
                        Not_found ->
                           printSolution solution;
                           Updatedebug.abort (Updatedebug.bug "annotateWithFunEffects: Couldn't find something in hash")
                  end
                     
               | _ -> Updatedebug.abort (Updatedebug.bug "annotateWithFunEffects: too many effects or couldn't find effect on function")
            end
         | _ -> acc @ [g]
   in
      file.globals <- List.fold_left annotate_fold [] file.globals

(*============================================================================*)









(*============================================================================*)
(* ---- Conditional constrain solving ---- *)


(* Given a solution to the mu constraints resolve each conditional delta constraint into a 
   regular subsetset constraint *)
let resolveConditionalConstraints (mu_solution : (int,bool) H.t) (cset : DeltaConstraintSet.t) : DeltaConstraintSet.t =
   let resolve dc cset = 
      match dc with 
        DCCond (muid,(d1,d2)) -> 
           begin try
              if H.find mu_solution muid then begin
       (match (d2.delta_kind) with
       DKConst _ -> let d1str,d2str = makeDeltaDebugStrings d1 d2 in 
       ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Adding delta constraint %s < %s\n" d1str d2str)
      | _ -> ());
 
                 DeltaConstraintSet.add (DCSubset(d1,d2)) cset
               end
              else 
                 cset
           with
              Not_found -> 
                 (* FIXME: Not sure about this? *)
                 (* If we didn't find it in the solution then assume mu to be true *)
                 (*DeltaConstraintSet.add (DCSubset(d1,d2)) cset*) cset
           end

      | DCSubset(d1,d2) -> 
           DeltaConstraintSet.add dc cset
   in
   DeltaConstraintSet.fold resolve cset DeltaConstraintSet.empty

(*============================================================================*)







(*============================================================================*)
(************************* Debug printers **********************************)
let printSpeculativeUpdateMuHash (speculativeUpdateMuHash : (int,int) H.t) = 
   let printEntry stmt_id mu_id =
      ignore(Updatedebug.log [Updatedebug.BTAnalysis] "%d  |->  %d\n" stmt_id mu_id)
   in
   ignore(Updatedebug.log [Updatedebug.BTAnalysis] " Speculative update mu hash START *********************\n");
   H.iter printEntry speculativeUpdateMuHash;
   ignore(Updatedebug.log [Updatedebug.BTAnalysis] " Speculative update mu hash END ***********************\n")

(*============================================================================*)







(*============================================================================*)
(* ---- Do analysis ---- *)

let printStatistics () : unit AM.m = 
   AM.amGetNumMuConstraints >>= fun num_mu ->
   AM.amGetNumDeltaConstraints >>= fun num_delta ->
   ignore(Updatedebug.log [Updatedebug.BTStatistics] "# Constraints : mu = %d, delta = %d\n" num_mu num_delta);
   return ()



(* return the result of running the monad and a solution to the mu constraints *)
let runAnalysis (mallocFunHash : Mallocpragma.mallocHash_t)  
                (effectHash : funEffectHash_t) 
                (funTypeHash : funTypeHash_t) 
                (solutionHash : deltaSolutionHash_t) 
                (setjmpHash : Setjmp.setjmpHash_t) 
                (file : Cil.file) 
                (suspect_result : Suspectalias.SAM.mo_result)
                (internVarHash : (string, unit) H.t)
                : analysis_result  = 
  let monad_result = 
     Stats.time "Stats : Run analysis monad - doAN"
      AM.run mallocFunHash effectHash solutionHash suspect_result internVarHash
          (                           (AnalysisGenFreshEffects.genFreshEffectsGlobal file)
            >>= fun _              -> AM.amCreateVoidPtrFunEffect
            >>= fun _              -> (analyzeFile file) 
            >>= fun _              -> AM.amConstrainVoidPtrFunEffect 
            >>= fun _              -> purgeLongjmpConstraints setjmpHash 
            >>= fun _              -> printStatistics ())
   in
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu constraints start ***********************************\n" in
   let _ = printMuConstraintSet monad_result.AM.mu_constraints in
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu constraints end *************************************\n" in
   let mu_solution = Stats.time "Stats : solveMu - doAN" solveMu monad_result.AM.mu_constraints in
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu solution start ***********************************\n" in
   let () = printMuSolution mu_solution in
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu solution end *************************************\n" in
   let dcs_without_cond = Stats.time "Stats : resolveConditionalConstraints - doAN" resolveConditionalConstraints mu_solution monad_result.AM.delta_constraints in

   let () = Stats.time "Stats : solveDelta" solveDelta dcs_without_cond solutionHash in
   let _ = Updatedebug.log [Updatedebug.BTDeltasolver] "Delta solution start ***********************************\n" in
   let () = Deltasolver.printSolution solutionHash in
   let _ = Updatedebug.log [Updatedebug.BTDeltasolver] "Delta solution end ***********************************\n" in
   let () = installPrintEffectsHook solutionHash monad_result.AM.stmtid_to_delta_id in
   let () = Stats.time "Stats : annotateWithFunEffects - doDSU" annotateWithFunEffects monad_result.AM.funEffectHash 
                                                                                       funTypeHash
                                                                                       solutionHash 
                                                                                       mu_solution 
                                                                                       suspect_result.Suspectalias.SAM.sar_fname_to_LPS_hash file in 

   Stats.time "Stats : removeEffectAttrVisitor - doDSU" Cil.visitCilFile (new Updatecilutil.removeEffectAttrVisitor(Updateconfig.effectAttrName)) file;
   printSpeculativeUpdateMuHash monad_result.AM.speculativeUpdateMuHash;
   {
      ar_delta_constraints = monad_result.AM.delta_constraints;       
      ar_mu_constraints = monad_result.AM.mu_constraints;
      ar_delta = monad_result.AM.delta;                 
      ar_mu = monad_result.AM.mu;                     
      ar_stmtid_to_delta_id = monad_result.AM.stmtid_to_delta_id;     
      ar_funEffectHash = monad_result.AM.funEffectHash;          
      ar_speculativeUpdateMuHash = monad_result.AM.speculativeUpdateMuHash;
      ar_mu_solution = mu_solution;           
      ar_delta_solution = solutionHash;
      ar_constNamesHash = monad_result.AM.constNamesHash
   }

(*============================================================================*)


