(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* 
   This module defines a generic fresh effects inserter GenFreshEffects that is 
   parametised on monadic glue and an Effect inserter.  

  The purpose of this module is to provide the generic part of fresh effect
  insertion into all types in a program.  Fresh effect insertion is tied 
  in with the monad used in a particular analysis as we have to keep track of
  effect variable id's.  

  This module is parametized as currently we need to insert effects into 
  function arrows for the update analysis, and function arrows and references 
  for suspect alias analysis.  The functor takes two arguments:
    
     EI - an effect inserter that uses specific properties of the monad to 
          insert fresh effects
     M  - the monad to use in gluing together the generic traversal of the 
          syntax

  The complicated type constraint says that the the effect inserter 
  supplied must use the same monad type as the monad supplied.  This
  ensures that the effect inserter and this module can glue things together
  in a compatible (the same) way.

*)



open Cil

module H = Hashtbl



(*let debugPrint str = Printf.printf str*)
let debugPrint (str:string) : unit = ignore (str)




module type EffectCreator =
sig
  (* The monad that this effect inserter works over *)
  module M : Monad.SyntaxMonad

  (* This function acts non-inductively, inserting an effect in the 
     outer constructor if it wants to *)
  val insertEffectInOuterConstructor : Cil.typ -> Cil.typ M.m
end
  



module EffectInserter = 
  functor (EC : EffectCreator) -> 
    functor (M : Monad.SyntaxMonad with type mi = EC.M.mi and type 'a mo = 'a EC.M.mo) -> 
struct

  (* Create the monad operations module for the monad we are operating over *)
  module MOps = Monad.MonadOps(M)
  
  (* locally bind monad operations *)
  let (>>=) = M.(>>=)
  let return = M.return


  (* ---- Code to traverse the syntax ---- *)


  (* Insert the given effect key into all function arrows in type, or if None, generate
     fresh effects 

     NOTE: This only calls EI.insertEffectInOuterConstructor for array, pointer and 
           functions at the moment as this is all we need.  Would be more general to 
           call it for all of them.
  *)
  let insertEffectsInType (typ : Cil.typ) : Cil.typ M.m =
    let compHash = H.create 4099 in
    let rec insertEffectsInType' (typ' : Cil.typ) : Cil.typ M.m = 
       (* Iulian : this produces big-time output, in the order of GBs; uncomment if you want it *)
(*       MOps.lift  *)
(*         ((\* fun () -> ignore (Updatedebug.log [Updatedebug.BTAnalysis]  *\) *)
(* (\*            "insertEffectsInType: type is %a\n" d_plaintype typ)) (return ()) >>=  *\)fun () -> *)
          
        match typ' with 
            TVoid _
          | TInt _ 
          | TFloat _ -> EC.insertEffectInOuterConstructor typ'
          | TPtr (t,attr) -> 
              insertEffectsInType' t
              >>= fun t' -> EC.insertEffectInOuterConstructor (TPtr(t', attr))  
            
          | TArray(t,e_opt,attr) -> 
              insertEffectsInType' t
              >>= fun t' -> EC.insertEffectInOuterConstructor (TArray(t', e_opt, attr)) 
            
          | TFun(tret,targs,isva,al) ->
              MOps.lift debugPrint (return "Geneffects TFun\n") >>= fun _ ->
              let args_map (s,t,a) =
                insertEffectsInType' t
                >>= fun t' -> return (s, t', a)
              in 
                begin match targs with 
                    Some l -> MOps.map args_map l >>= fun args -> return (Some args)
                  | None   -> return None
                end
                >>= fun targs' -> insertEffectsInType' tret
                >>= fun tret'  -> 
                  (* call the generic effect inserter *)
                  EC.insertEffectInOuterConstructor (TFun(tret', targs', isva, al))
                            
          | TNamed _ -> return typ'
       
          | TComp(ci,al) ->  
              if H.mem compHash ci.cname then 
                return typ'
              else
                begin
                  H.add compHash ci.cname true; 
                  let field_map (fi : Cil.fieldinfo) : unit M.m = 
                    insertEffectsInType' fi.ftype
                    >>= fun ftype' -> 
                      begin
                        fi.ftype <- ftype';
                        return ()
                      end
                  in
                    MOps.iter field_map ci.cfields 
                    >>= fun _ -> return typ' 
                end
                        
          | TEnum _
          | TBuiltin_va_list _ -> return typ'
    in 
      insertEffectsInType' typ


        
  (* Traverse the type an insert fresh effect variables at function types *)
  let insertFreshEffectsInType (typ : Cil.typ) : Cil.typ M.m =
    insertEffectsInType typ



  let rec genFreshEffectsStmt (stmt : Cil.stmt) : Cil.stmt M.m = 
    begin match stmt.skind with 
        Instr(il) -> MOps.map genFreshEffectsInstr il >>= fun il' -> return (Instr il')
          
          | Return(eopt, loc) -> 
              begin match eopt with 
                  None    -> return stmt.skind
                | Some e -> genFreshEffectsExp e >>= fun e' -> return (Return((Some e'), loc))
              end 
                
          | If(e_test, b_t, b_f, loc) -> 
              genFreshEffectsExp e_test
              >>= fun e_test'   -> MOps.map genFreshEffectsStmt b_t.bstmts
              >>= fun b_t_stmts -> MOps.map genFreshEffectsStmt b_f.bstmts
              >>= fun b_f_stmts -> 
                b_t.bstmts <- b_t_stmts;
                b_f.bstmts <- b_f_stmts;
                return (If(e_test', b_t, b_f, loc))
                  
                | Loop (b_body, loc, start, finish) -> 
                    MOps.map genFreshEffectsStmt b_body.bstmts
                    >>= fun b_body' ->
                      b_body.bstmts <- b_body';
                      return (Loop (b_body,loc,start,finish))
    
                | Block b -> 
                                  MOps.map genFreshEffectsStmt b.bstmts
                    >>= fun b' -> 
                      b.bstmts <- b';
                      return (Block b)
    
                | Goto _ 
                | Break _ 
                | Continue _ 
                | Switch _ 
                | TryFinally _ 
                | TryExcept _ -> return stmt.skind
    end
    >>= fun stmtKind -> 
      stmt.skind <- stmtKind;
      return stmt



  and genFreshEffectsInstr (instr : Cil.instr) : instr M.m = 
    match instr with 
        Set(l, e, loc) ->
                        genFreshEffectsLval l 
          >>= fun l' -> genFreshEffectsExp e
          >>= fun e' -> return (Set(l', e', loc))
         
            | Call(lopt, e_f, e_args, loc) -> 
                                   genFreshEffectsExp e_f
                >>= fun e_f'    -> MOps.map genFreshEffectsExp e_args
                >>= fun e_args' -> 
                  begin match lopt with 
                      None   -> return None
                    | Some l -> genFreshEffectsLval l >>= fun l' -> return (Some l')
                  end
                    
                >>= fun lopt'   -> return (Call(lopt', e_f', e_args', loc))
         
            | Asm _ -> return instr




  and genFreshEffectsExp (exp:Cil.exp) : Cil.exp M.m =
    match exp with 
      | Const _ -> return exp
      | UnOp (uop, e, t) -> 
                        genFreshEffectsExp e 
          >>= fun e' -> insertFreshEffectsInType t 
          >>= fun t' -> return (UnOp (uop, e', t'))

    | BinOp (bop, e1, e2, t) -> 
                       genFreshEffectsExp e1
        >>= fun e1' -> genFreshEffectsExp e2
        >>= fun e2' -> insertFreshEffectsInType t
        >>= fun t'  -> return (BinOp (bop, e1', e2', t'))
         
    | Lval l -> 
                      genFreshEffectsLval l
        >>= fun l' -> return (Lval l')
         
    | SizeOf t -> 
        (* No effects needed in type in size of as it can never interact with anything else *) 
        return (SizeOf t) 

    | SizeOfE e -> 
        genFreshEffectsExp e >>= fun e' -> return (SizeOfE e')
    
    | SizeOfStr _ 
    | AlignOf _ 
    | AlignOfE _ -> return exp
    
    | CastE (t, e) -> 
                       genFreshEffectsExp e
        >>= fun e'  -> insertFreshEffectsInType t
        >>= fun t'  -> return (CastE (t',e'))
         
    | AddrOf lval  -> genFreshEffectsLval lval >>= fun lval' -> return (AddrOf lval')
    | StartOf lval -> genFreshEffectsLval lval >>= fun lval' -> return (StartOf lval') 



  and genFreshEffectsLval ((lhost,loffset) : Cil.lval) : Cil.lval M.m = 
    let rec walkOffset offset = 
      match offset with 
          NoOffset -> return NoOffset
            
        | Field (fi,offset') -> 
                                walkOffset offset' 
            >>= fun offset'' -> return ( Field (fi, offset'') )
             
        | Index (e,offset') ->
                                walkOffset offset'
            >>= fun offset'' -> genFreshEffectsExp e
            >>= fun e'       -> return ( Index (e', offset') )
    in
      begin match lhost with 
          Var _ -> return lhost
        | Mem e ->
                          genFreshEffectsExp e
            >>= fun e' -> return ( Mem e' )
      end
      >>= fun lhost'   -> walkOffset loffset
      >>= fun loffset' -> return (lhost', loffset')



(* The following 4 functions take care of 
   recursively inserting effects in static initializers 
*)
let rec ei_foldLeftCompound (ct: Cil.typ) (initl: (Cil.offset * Cil.init) list) : (Cil.offset * Cil.init) list M.m = 
  let bail s = Updatedebug.abort (Updatedebug.bug "ei_foldLeftCompound %s" s) in
  match unrollType ct with
    TArray(bt, _, _) -> 
       MOps.map o_genFreshEffectsInit initl 

  | TComp (comp, _) -> 
       MOps.map o_genFreshEffectsInit initl 


  | _ -> bail "Type of Compound is not array or struct or union"



and o_genFreshEffectsInit ((off, ii) : Cil.offset * Cil.init) : (Cil.offset * Cil.init) M.m = 
   begin match ii with
      SingleInit e -> 
                         genFreshEffectsExp e
         >>= fun e'   -> return ( off, SingleInit e' )
  
    | CompoundInit (t,ol) -> 
                         ei_foldLeftCompound t ol
         >>= fun ol'  -> insertFreshEffectsInType t
         >>= fun t'   -> return ( off, CompoundInit (t', ol') )
   end

let genFreshEffectsInit (ii : Cil.init) : Cil.init M.m = 
   begin match ii with
      SingleInit e -> 
                         genFreshEffectsExp e
         >>= fun e'   -> return ( SingleInit e' )
  
    | CompoundInit (t,ol) -> 
                         ei_foldLeftCompound t ol
         >>= fun ol'  -> insertFreshEffectsInType t
         >>= fun t'   -> return ( CompoundInit (t', ol') )
   end

let genFreshEffectsInitinfo (ii : Cil.initinfo) : Cil.initinfo M.m =  
   begin match ii.init with 
      Some i -> 
                       genFreshEffectsInit i
         >>= fun i' -> return ( {init = Some i'} )
    
    | None   ->  
                       return ( {init = None} )
   end

  (* Traverse the globals and insert fresh effects on function types *)
  (* We have to replace the types in-place, because the bodies of the functions have references 
     to global functions and variables, and we want function/variable types to be decorated with effects *)

let genFreshEffectsGlobal (file : Cil.file) : unit M.m = 
   let genFreshEffectsVarinfo (vi : varinfo) : unit M.m =
                   insertFreshEffectsInType vi.vtype
      >>= fun t -> return(vi.vtype <- t)
      >>= fun _ -> return ()
   in 
      
   let genFreshEffectsGlobal_map (g : Cil.global) :  Cil.global M.m =    

(*       let _ = ignore(Updatedebug.log [Updatedebug.BTAnalysis] "generateFunFreshEffects for global '%s'\n" (Commoncilutil.getGlobalName g)) in *)

      M.setLoc (Cil.get_globalLoc g) >>= fun _ ->

      match g with
          GType(ti,loc) -> 
                           insertFreshEffectsInType ti.ttype 
            >>= fun t ->   (ti.ttype <- t; 
                           return g)
   
        | GVar(vi,ii,loc) -> 
                           insertFreshEffectsInType vi.vtype 
            >>= fun t   -> genFreshEffectsInitinfo ii
            >>= fun ii' -> (vi.vtype <- t; 
                           return (GVar(vi, ii', loc)))
   
        | GVarDecl(vi,loc) -> 
                           insertFreshEffectsInType vi.vtype 
            >>= fun t   -> (vi.vtype <- t; 
                           return g)
   
        | GCompTag(ci,loc) -> 
             let field_iter (fi : Cil.fieldinfo) : unit M.m = 
                insertFreshEffectsInType fi.ftype
                >>= fun ftype' -> (fi.ftype <- ftype'; return ())  
             in
                           MOps.iter field_iter ci.cfields 
            >>= fun () ->  return g
           
        | GFun(fd,loc) -> 
                           M.setCurrentFun fd 
              >>= fun _ ->
                           let old_t = fd.svar.vtype in
                              begin match old_t with 
                                    TFun(_,_,_,_) -> insertFreshEffectsInType old_t >>= fun t   -> 
                                       Cil.setFunctionType fd t; (* This changes the formals as well *)
                                                        return ()
                                       (* Now add effects to the types of the local variables *)
                                       >>= fun ()    -> MOps.iter genFreshEffectsVarinfo fd.slocals

                                       (* Gen fresh effects in the body *)
                                       >>= fun ()     -> MOps.map genFreshEffectsStmt fd.sbody.bstmts
                                       >>= fun bstmts' ->  fd.sbody.bstmts <- bstmts'; 
                                                           return ()
                               
                                  | _ -> Updatedebug.abort (Updatedebug.bug
                                                               "generateFunFreshEffects_fold(2): expected function type, but got:\n %a\n" 
                                                               d_plaintype fd.svar.vtype)
                              end
              >>= fun () -> M.unsetCurrentFun 
              >>= fun _  -> return g
      
        | _ -> return g
      
    in

                          MOps.map genFreshEffectsGlobal_map file.globals
      >>= fun globals' -> return (file.globals <- globals')
        
  
end







