(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil

module H = Hashtbl
(*  
    In order to deal with changes in types of global variables, 
    we need to convert their types to named ones; type transformers will
    buy us 'free' conversions between global variables in different versions.

    Procedure: all definitions
    
         t v [= init];
    
    will be converted into:
 
        typedef t DSU_gvar_typeof_v;
        DSU_gvar_typeof_v v [= init];

    Note that t can be a named or base type. In either case, we need to introduce 
    a new named type for it, in order to cope with gvar type changes like this: 

    [version 1] typedef ... T; T g;
    [version 2] int g;

*)

(* Given a GVar/GVarDecl, return a new typeinfo w/the appropriate name and type
   Callers are free to construct a TNamed/GType out of it
*)
let makeTypeForGlobalVariable (g : Cil.global) : Cil.typeinfo = match g with
      GVar(vi, _, loc) | GVarDecl (vi, loc)  -> 
         (* based on the GVar's name/type, make up a new named type*)
         let uniqTypeName = Updateconfig.makeGlobalVarTypeName vi.vname in
         let ti = { tname = uniqTypeName;
                    (* we have to drop consts beacuse these might be used in a CON *)
                    ttype = Cil.typeRemoveAttributes ["const"] vi.vtype; 
                    treferenced = true;
                  } 
         in
         ti

   | _ -> Updatedebug.abort (Updatedebug.bug "makeTypeForGlobalVariable : argument should be a GVar/GVarDecl, but it is %a instead !\n" d_shortglobal g);;
         

(* Given a file, add new named types for all global variables and change all global variables'types
   to be instances of this named type
*)
let genGlobalVarUniqTypes (file : Cil.file) (vd : Versiondata.version_data) : unit =

   (* two-purpose hash; 
      first it holds the names of all GVar we have to create types for;
      as we go along and create types, remove the varname from the hash, so that
      subsequent GVarDecl/GVar's  won't try to create a new type again  
   *)
   let gvarNames = 
      let res : (string, unit) H.t = H.create 4095 in
      let sweepGvarNames_iter g = 
         match g with
               GVar(vi, ii, loc) when not (Cil.isFunctionType vi.vtype) -> Commonutil.hashtbl_add_once res vi.vname ()
            |  GVarDecl (vi, loc) when not (Cil.isFunctionType vi.vtype) -> 
                  (* if we're in patch mode, don't forget the external decls *)
                     if Updateargs.args.Updateargs.arg_currentMode = Updateargs.PatchMode &&
                        Versiondata.isExistingVar vd vi.vname
                     then
                        Commonutil.hashtbl_add_once res vi.vname ()
            | _ -> ()
      in

         (* Find out all GVar names *)   
         List.iter sweepGvarNames_iter file.globals;
         res
   in
   let genGlobalVarUniqType_fold acc g = 
      match g with
         GVar(vi, _, loc) | GVarDecl (vi, loc) when H.mem gvarNames vi.vname  -> 
            let newTi = makeTypeForGlobalVariable g in
            let uniqType = TNamed(newTi, []) in

            (* replace the type of the GVar in-place*)   
            vi.vtype <- uniqType;
            ignore(Updatedebug.log [Updatedebug.BTGvartypes] "genGlobalVarUniqType_fold : adding named type : type %s = %s\n" newTi.tname (Updatedebug.ppType vi.vtype));

            (* remove it from the hash so that subsequent GVarDecl/GVar 's 
               won't try to create a new type again  *)
            H.remove gvarNames vi.vname;

            (* insert the named type before the var *)   
            acc @ [ GType(newTi, loc) ]
         
        | _ -> acc 
   in

   let gvarTypesList = List.fold_left genGlobalVarUniqType_fold [] file.globals in

   (* paste global var types that we generated in gvartypes.ml *)
   let splitglobals = Updatecilutil.splitGlobalsBeforeVars file.globals in
   file.globals <- (fst splitglobals) @ 
                   (gvarTypesList) @ 
                   (snd splitglobals);;






