(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Function indirection *)

open Cil

module H = Hashtbl

(* 
   This is the general idea: we have 'updatable' and 'non-updatable' functions.

   'Updatable' functions are those whose address is never taken, hence
   they can be safely indirected. Function calls to 'updatable' functions 
   are replaced like this:

   f(x) => *__indirect_f(x)

   where 'f' is a updatable function, and '__indirect_f' is a global variable 
   initialized to &f
   

   'Non-Updatable' functions are those whose address is taken; we can't indirect
   them because we might pass them to a caller that has no clue about 
   indirection and won't dereference the function pointer passed in.

   All references to 'non-updatable' functions are replaced like this:

   f => f'
   
   where 'f' is a non-updatable function, and f' is a fresh global function 
   that's non-updatable and its only purpose is to call f, e.g.

   f : t-> t_1 ->... -> t_n

   will add
   
   f': t-> t_1 ->... -> t_n
   {
   f(args)
   }
   
   to the globals. 

   The idea is to render non-updatable functions updatable by introducing a non-updatable
   f' whose sole purpose is to call f.
   
*)


(* module that stores information about functions *)
(*====================*)
module FunctionInfo =
(*====================*)
struct
  
  exception Non_static_function of string
     
 
  type funKind = 
      FKextern 
    | FKbuiltin
    | FKwrapper
                (* indirectionVar;   forward decl;    wrapper;               "descriptor" *)
    | FKstatic of  Cil.global *      Cil.global *     Cil.global option *    Cil.varinfo

  type funInfo = 
      {
        fun_info_updatable    : bool;        (* is the function updatable ? *)
        fun_info_kind         : funKind;     (* if it's static then
                                                the global contains the indirect var *)
      };;
  
  (* Hash table storing a map from the function name to the corresponding
     funInfo *)
  type funInfoHash_t = (string, funInfo) H.t
  let funInfoHash : funInfoHash_t = H.create 4099

  (* by default, static (e.g. defined) functions are updatable *)
  let defaultUpdatable = true

  (* Global type : Updateconfig.functionVersionListStructName
     all indirection vars are this type *)   
  let fVersionList_ci = 
     (* host struct *)
     let ci = Commoncilutil.makeCompInfo true Updateconfig.functionVersionListStructName in
     (* populate filed list *)
     let field_list = List.map 
                         (fun (fieldName, fieldType) -> Commoncilutil.makeField ci fieldName fieldType) 
                         Updateconfig.functionVersionListFields 
     in
     ci.cfields <- field_list;
     ci

  let fVersionList_type () = TComp(fVersionList_ci,[])
  let fVersionList_global () = GCompTag(fVersionList_ci, Cil.locUnknown)

  let addFVersionList_args = [
     ("listHead", TPtr(fVersionList_type (),[]));
     ("functionName", Cil.charPtrType);
     ("functionVersion", Cil.uintType);
     ("patchVersion", Cil.uintType);]

  let lock_args = [
     ("listHead", TPtr(fVersionList_type (),[]));
     ("functionVersion", Cil.uintType)]

  let unlock_args = [
     ("listHead", TPtr(fVersionList_type (),[]));
     ("functionVersion", Cil.uintType)]

  let makeVi funName args =
     let newFunction = Cil.emptyFunction funName in 
     let createNewFormals_map (vname,vtype) = Cil.makeFormalVar newFunction ~where:"^" vname vtype in
     let newFormals = List.map createNewFormals_map args in
     (* fix function type *)
     let () = Cil.setFormals newFunction newFormals in
     (* declare it as extern *)
     newFunction.svar.vstorage <- Extern; 
     newFunction.svar      


  let addFVersionList_vi = makeVi  Updateconfig.addFVersionListFunctionName addFVersionList_args
  let lock_vi            = makeVi  Updateconfig.lockFunctionName            lock_args
  let unlock_vi          = makeVi  Updateconfig.unlockFunctionName          unlock_args 


  let getAddFVersionList_vi () = addFVersionList_vi
  let getLock_vi () = lock_vi
  let getUnlock_vi () = unlock_vi

  (* create indirection variable; it has name based on function name and type (functionType  ref )

     used internally by this module *)
  (*------------------------------------------------*)
  let makeIndirectionVar (v : Cil.varinfo) (version : int) : Cil.global = 
  (*------------------------------------------------*)
    let functionName = v.vname in
    let functionType = v.vtype in
    let indirectionVarName = Updateconfig.makeIndirectionVarName functionName version in
    let indirectionVarTyp = fVersionList_type () in
    let indirectionVar = Cil.makeGlobalVar indirectionVarName indirectionVarTyp in
    let funAddr = Cil.mkAddrOf (var v) in
    let init_items = [
           (*       NAME                                            INITIALIZER    *)
           ((fst Updateconfig.fv_itemBaseFunName),          (SingleInit (Const(CStr functionName))));
           ((fst Updateconfig.fv_itemFirst),                (Cil.makeZeroInit (snd Updateconfig.fv_itemFirst)));
           ((fst Updateconfig.fv_itemActiveFunction),       (SingleInit funAddr));
           ((fst Updateconfig.fv_itemOldVersionTotalCount), (Cil.makeZeroInit (snd Updateconfig.fv_itemOldVersionTotalCount)));
           ((fst Updateconfig.fv_itemLength),               (Cil.makeZeroInit (snd Updateconfig.fv_itemLength)));] in

    let compound_init_list = List.map 
                           (fun (name, single_init) -> let field_info = Commoncilutil.getFieldinfoByName (fVersionList_type ()) name None in
                                                       (Field(field_info,NoOffset), single_init))    
                           init_items 
    in   
      GVar(indirectionVar, {init = Some (CompoundInit(indirectionVarTyp, compound_init_list))}, Cil.locUnknown)

  (* We look in the vd for existing indirection variables and return those if they exist 
     If we don't find them in the hash this is the first version of this function and
      we define new global indirection variables 

     create forward declaration for function; *)
  (*------------------------------------------------------------------------------------------------*)
   let getIndirectionVars (vd : Versiondata.version_data) (v : Cil.varinfo) : (Cil.global * Cil.global) = 
  (*------------------------------------------------------------------------------------------------*)
      let indirectionVarGlobal, forwardDeclGlobal = 

      if Versiondata.isExistingFunction vd v.vname
      then
         let oldFtype = Versiondata.getFunctionType vd v.vname
         and newFtype = v.vtype in

         (* Test whether the function has changed signature. 
            If it has, make it appear as a new function,
            so that old clients don't try to call it
         *)
         if true (* Commoncilutil.isEqualFuntypeQuick oldFtype newFtype *) (* ruled out because we now do subtyping/hold the patch application until it's safe to do so *)
         then
            (* OK, we can use the old indirection var since function signature hasn't changed *)

            (* getIndirectionVarNameForFunction can't throw an exception because we've tested membership above*)
            let oldIndirectionVarName = Versiondata.getFunctionIndirectionVarName vd v.vname in
            begin
               (* get the existing indirection variable and add it as an extern *)
               let indirection_vi = 
                  try 
                     Versiondata.getExistingVar vd oldIndirectionVarName
                  with 
                     Not_found -> Updatedebug.abort 
                        (Updatedebug.bug "getIndirectionVars for function '%s' : indirection var '%s' not found in vd !\n" v.vname oldIndirectionVarName)
               in
               indirection_vi.vstorage <- Cil.Extern;
               GVarDecl(indirection_vi, Cil.locUnknown), GVarDecl(indirection_vi, Cil.locUnknown)
            end
         else
            (* function has changed signature; make up a brand new indirection var *)
            let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "Function '%s' has changed signatures between versions\n" v.vname) in 
            (* first set up function version *)
            let currentVDversion = Versiondata.getCurrentVersion vd in

            (* and finally generate a new indirection var *)
            (makeIndirectionVar v (currentVDversion + 1)), (GVarDecl (v, Cil.locUnknown))
      else
         (* this is version 0; generate a fresh indirection var *)
         (makeIndirectionVar v 0, (GVarDecl (v, Cil.locUnknown)))

      in
      (indirectionVarGlobal, forwardDeclGlobal)  

  (* Add function info to hash; private *)
  (*-------------------------------------------------------------------------------------------------------------*)
  let addFunction (vd : Versiondata.version_data option) (functionName : string) (functionKind : funKind) : unit = 
  (*-------------------------------------------------------------------------------------------------------------*)

    let doAddFunction (functionName' : string) (functionKind' : funKind) : unit = 
    begin
      let fi = (* if it's a static, we generate a global for the indirection variable *)
        match functionKind' with

            FKstatic (_, _, _, v) -> 
              let indirectionVars = match vd with 
                                      Some vd -> getIndirectionVars vd v 
                                    | None -> Updatedebug.abort (Updatedebug.bug "(%a) : doAddFunction (%s): should have passed us a vd !\n" 
                                                                d_loc !(Cil.currentLoc)
                                                                functionName')

              in
              let indirectionVar = fst indirectionVars in
              let forwardDeclVar = snd indirectionVars in
                
              { fun_info_updatable = defaultUpdatable;
                fun_info_kind = FKstatic (indirectionVar, forwardDeclVar, None, v) }

          | _ -> (* externs and builtins are non-updatable *)
              { fun_info_updatable = false;
                fun_info_kind = functionKind' }

      in
        H.add funInfoHash functionName' fi 
    end
    in
      try 
        begin
          (* replace actions are not legit *)
          let fi = H.find funInfoHash functionName in 
            Updatedebug.abort (Updatedebug.bug "addFunction (%a): Tried to replace fun info for %s !\n" 
                                 d_loc !(Cil.currentLoc)
                                 functionName)
        end

      with Not_found ->
	((* Updatedebug.log [Updatedebug.BTAlways] "(%a) adding function %s \n" d_loc !(Cil.currentLoc) functionName; *)
        doAddFunction functionName functionKind)


    let addBuiltinFunction (functionName : string) : unit = 
      addFunction None functionName FKbuiltin

    (* TODO : should we allow redeclaration of externs ? *) 
    let addExternFunction (functionName : string) : unit = 
    begin 
      if not (H.mem funInfoHash functionName)
      then addFunction None functionName FKextern
    end

    let addWrapperFunction (functionName : string) : unit = 
      addFunction None functionName FKwrapper

    let addStaticFunction (vd : Versiondata.version_data option)  (functionName : string) (g : Cil.global) (v : varinfo): unit = 
      begin
        match g with
            GFun (fd, loc) -> 
               addFunction vd functionName (FKstatic (g, g, None, fd.svar))

          | GVarDecl (vi,loc) when Updateargs.args.Updateargs.arg_currentMode = Updateargs.PatchMode ->
               if Funversioning.shouldRename vi.vname
               then addFunction vd functionName (FKstatic (g, g, None, vi))
               else addExternFunction vi.vname

          | _ -> Updatedebug.abort (Updatedebug.bug "doAddFunction (%a): Illegal global for FKstatic %s. Must be GFun or at most GVarDecl in patch mode, but it was %a instead!\n" 
                                      d_loc !(Cil.currentLoc)
                                      functionName
                                      d_global g)
      end

  (* return indirection var associated with this name; may raise Not_found / Non_static_function *)
  (*---------------------------------------------------------*)
  let getIndirectionVar (functionName : string) : Cil.varinfo =
  (*---------------------------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with
        FKstatic (_, _, Some wr, _) -> 
          begin
            match wr with
                GFun(fd, loc) -> fd.svar
              | _ -> Updatedebug.abort (Updatedebug.bug "getIndirectionVar (%s):  junk in global field : %a\n" 
                                          functionName
                                          d_global wr)
          end

      | FKstatic (gi, _, None, _) -> 
          begin
            match gi with
              | GVar(vi, ii, loc) -> vi
              | GVarDecl(vi, loc) -> vi
              | _ -> Updatedebug.abort (Updatedebug.bug "getIndirectionVar (%s):  junk in global field : %a\n" 
                                          functionName
                                          d_global gi)
          end
      | _ -> raise (Non_static_function ("getIndirectionVar " ^ functionName))


  (*  *)
  (*----------------------------------------------------------------------*)
  let buildIndirectionCall (gi : Cil.varinfo) (function_type : Cil.typ) : Cil.exp = 
  (*----------------------------------------------------------------------*)

  (* this does f(x) ->  (vi.vtype)(__indirect_f.activeVersion)(x) *)
           
  (* first get the .activeFunction *)
  let field_info = Commoncilutil.getFieldinfoByName (fVersionList_type ()) ((fst Updateconfig.fv_itemActiveFunction)) None in
     (* build the new lval *)   
  let new_lval = (Var gi), (Field(field_info, NoOffset)) in 
     (* indirectionVar.activeFunction has type void *; we need to force it to a function type *)
  let new_fexpr = Cil.mkCast (Lval new_lval) (TPtr(function_type,[])) in
  new_fexpr 


  (* create a wrapper for this function; the wrapper function has exactly the same type 
     as 'v', and its solely purpose is to call 'fd' and pass back the eventual return value *)
  (*----------------------------------------------------------------------*)
  let makeWrapperFunction (v : Cil.varinfo) (version : int) : Cil.global = 
  (*----------------------------------------------------------------------*)
    let newFunctionName = Updateconfig.makeFunctionWrapperName v.vname version in
    let newFunction  = Cil.emptyFunction newFunctionName in 
    let createNewFormals_fold (s,t,a) acc = (Cil.makeFormalVar newFunction ~where:"^" s t) :: acc in
    let newFormals = List.fold_right createNewFormals_fold (Updatecilutil.argsFromType v.vtype) [] in
    let _ = newFunction.svar.vtype <- v.vtype in
    let _ = newFunction.sformals <- newFormals in
    let newBody = 
      let retval = 
        begin
          match Cil.unrollType v.vtype with
              TFun(TVoid attr, args, isva, a) -> None
            | TFun(rt, args, isva, a) -> 
                let tmp = makeLocalVar newFunction "tmp" rt in
                  Some (var tmp)
            | _ -> Updatedebug.abort (Updatedebug.bug "makeWrapperFunction : function %s doesn't have function type ?\n" 
                                        v.vname)
        end
      in
      let functionName = v.vname in
      let indirectedOriginalFunExpr =
        try
          let gi = getIndirectionVar functionName in
            (* do the f(x) -> ( *__indirect_f)(x) *)
          buildIndirectionCall gi v.vtype
        with 
          Not_found -> Updatedebug.abort (Updatedebug.bug "makeWrapperFunction : couldn't find indirection var for %s in hash!\n" functionName)
      in
      let callToOriginalFunctionStmt =
        let argsAsExpr = List.map (fun v -> Lval(var v)) newFormals in
        let callInstr = Call(retval, indirectedOriginalFunExpr, argsAsExpr, Cil.locUnknown) in
          Commoncilutil.mkStmt (Instr [callInstr])
      in
      let eventualReturnStmt = 
        match retval with
            None -> []
          | Some lv -> [Commoncilutil.mkStmt (Return (Some (Lval lv), Cil.locUnknown))]
      in
      let newFunctionStmtList = callToOriginalFunctionStmt
                                ::eventualReturnStmt 
      in
        Cil.mkBlock newFunctionStmtList
    in
      newFunction.sbody <- newBody;
      addWrapperFunction newFunctionName;
      GFun(newFunction, Cil.locUnknown)



  (* true if the function is extern; may raise Not_found *)
  (*------------------------------------------*)
  let isExtern (functionName : string) : bool =
  (*------------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with
        FKstatic _ -> false
      | _ -> true (* externs and builtins are 'externs' *)

  (* true if the function is static; may raise Not_found *)
  (*-----------------------------------------*)
  let isStatic (functionName : string) : bool =
  (*-----------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with
        FKstatic _ -> true
      | _ -> false

  (* true if the function is static; may raise Not_found *)
  (*-----------------------------------------*)
  let isWrapper (functionName : string) : bool =
  (*-----------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with
        FKwrapper -> true
      | _ -> false

  (* return indirection var associated with this name; may raise Not_found / Non_static_function *)
  (*---------------------------------------------------------*)
  let getIndirectionGlobal (functionName : string) : Cil.global =
  (*---------------------------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with

        FKstatic (gi, _, _, _) -> gi
      | _ -> raise (Non_static_function ("getIndirectionGlobal " ^ functionName))


  (* return indirection var associated with this name; may raise Not_found / Non_static_function *)
  (*---------------------------------------------------------*)
  let getIndirectionVar (functionName : string) : Cil.varinfo =
  (*---------------------------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with

        FKstatic (GVar(vi,_,_), _, _, _) -> vi
      | FKstatic (GVarDecl(vi,_), _, _, _) -> vi
      | _ -> raise (Non_static_function ("getIndirectionVar " ^ functionName))

  (* return forward declaration var associated with this name; may raise Not_found / Non_static_function *)
  (*---------------------------------------------------------*)
  let getForwardDeclVar (functionName : string) : Cil.global =
  (*---------------------------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with

        FKstatic (_, gf, _, _) -> gf
      | _ -> raise (Non_static_function ("getForwardDeclVar " ^ functionName))

  (* return wrapper associated with this name; may raise Not_found / Non_static_function *)
  (*---------------------------------------------------------*)
  let getWrapper (functionName : string) : Cil.global option =
  (*---------------------------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with

        FKstatic (_, _, wr, _) -> wr
      | _ -> raise (Non_static_function ("getWrapper " ^ functionName))

  (* return wrapper var associated with this name; may raise Not_found / Non_static_function *)
  (*---------------------------------------------------------*)
  let getWrapperVar (functionName : string) : Cil.varinfo option =
  (*---------------------------------------------------------*)
    match (H.find funInfoHash functionName).fun_info_kind with

        FKstatic (_, _, None, _) -> None
      | FKstatic (_, _, Some (GFun(fd,_)), _) -> Some fd.svar
      | _ -> raise (Non_static_function ("getWrapperVar " ^ functionName))


  (* Set the updatability of a function *)
  (*------------------------------------------------------------*)
  let setUpdatability (functionName : string) (u : bool) : unit = 
  (*------------------------------------------------------------*)
    try 
      let fi = H.find funInfoHash functionName in
        H.replace funInfoHash functionName { fi with fun_info_updatable  = u }
    with 
        Not_found -> 
          Updatedebug.abort (Updatedebug.bug "setUpdatability (%a): Can't find %s in funInfoHash !\n" 
                               d_loc !(Cil.currentLoc)
                               functionName)  

  (* not needed (yet) *)
  (*----------------------------------------------*)
  let setUpdatable (functionName : string) : unit = 
  (*----------------------------------------------*)
    setUpdatability functionName true

  (*-----------------------------------------------------------------*)
  let setNonUpdatable (functionName : string) (version : int) : unit = 
  (*-----------------------------------------------------------------*)
    ignore (Updatedebug.log [Updatedebug.BTIndirection] "setNonUpdatable %s \n" functionName);
    begin
      let fi = H.find funInfoHash functionName in
        match fi.fun_info_kind with
            FKstatic (gi, gf, wr, v) ->
              if fi.fun_info_updatable
              then (* turn global into a function *)
                H.replace funInfoHash 
                  functionName 
                  { fi with fun_info_kind = FKstatic (gi, gf, Some (makeWrapperFunction v version), v) }

          | _ -> () (* extern or builtin; these were non-updatable anyway *)      
    end;
    setUpdatability functionName false

      
  (* true if the type is updatable; may raise Not_found *)
  (*---------------------------------------------*)
  let isUpdatable (functionName : string) : bool =
  (*---------------------------------------------*)
    (H.find funInfoHash functionName).fun_info_updatable


end

(*********************************   end of FunctionInfo module **********************************)


(* 
   This visitor sets non-updatable all the functions it encounters.
   N.B. : it is meant to run on expressions (called from detectNonUpdatable), 
   not on files !
*)
class detectNonUpdatableFunctionsInExpressionVisitor (vd : Versiondata.version_data) : cilVisitor = object(self)
  inherit nopCilVisitor

  method vlval lv = 
    begin
      match lv with
          Var vi, NoOffset when (Cil.isFunctionType vi.vtype && vi.vglob) ->
            let functionName = vi.vname in
              begin
                if FunctionInfo.isStatic functionName then
                  ignore (Updatedebug.log [Updatedebug.BTIndirection] "shooting down function '%s' because it appears in expression %a\n" functionName d_plainlval lv);
                FunctionInfo.setNonUpdatable functionName ((Versiondata.getCurrentVersion vd) + 1)
              end

        | _ -> ()
    end;
    DoChildren
end


(* call detectNonUpdatableFunction for every expression 
   in which we might take the address of a function;
   unfortunately we have to do this instead of using vexpr 
   because vexpr would be called on fun_expr in Call()
*)
class detectNonUpdatableFunctionsVisitor (detectNonUpdatableFunction : Cil.exp -> unit) : cilVisitor = object(self)
  inherit nopCilVisitor

  method vstmt s = 
    begin
      match s.skind with

          If (e, _, _, _)
        | Return(Some e, _) 
        | Switch (e, _, _, _) -> detectNonUpdatableFunction e

        | _ -> ()
    end;
    DoChildren

  method vinst i = 
    begin
      match i with

          Set (l, r, loc) -> 
            detectNonUpdatableFunction r

        (* fair function call in fun_expr ; let it thru *)
        | Call (result, Lval(Var vi, NoOffset), actuals, loc) -> 
            List.iter detectNonUpdatableFunction actuals

        | Call (result, fexpr, actuals, loc) -> 
            detectNonUpdatableFunction fexpr; List.iter detectNonUpdatableFunction actuals

        | _ -> ()
    end;
    DoChildren

 (* we have to visit the expressions in initializations, e.g.
    var v = &foo; or
    struct bar s = {&foo,...} *)     
  method vinit i = 
    begin
      match i with
          SingleInit exp -> detectNonUpdatableFunction exp
        | _ -> ()
    end;
    DoChildren
end


(* Add global indirection variables by folding over globals  *)
(*------------------------------------------------------------------------------------------------------*)
let addGlobalIndirectionVariables (vd : Versiondata.version_data) (gl : Cil.global list) : Cil.global list =
(*------------------------------------------------------------------------------------------------------*)

  (* Add indirection whenever we encounter a GFun *)
  let genIndirectionForGlobal_fold (g : Cil.global) (acc, nrUpdatable, nrTotal) =
    let id = (g::acc, nrUpdatable, nrTotal) in
    match g with
        GFun(fd, _) -> 
          let functionName = fd.svar.vname in
          (* Do nothing if this is a con/abs or 'update' function *)
          if not (FunctionInfo.isStatic functionName) then id
          else
            begin
              try  
                let nrUpdatable' = 
                  if FunctionInfo.isUpdatable functionName 
                  then nrUpdatable + 1
                  else nrUpdatable 
                in
                let nrTotal' = nrTotal + 1 in
                let gi = match FunctionInfo.getIndirectionGlobal functionName with
                   GVarDecl _ -> 
                      (* If genIdirectionGlobal has given us a GVarDecl, then this is an extern declaration to 
                         a function already in the program.  We do not generate the indirection variables as they
                         already exist. *)
                      []
                      
                 | GVar _ as gi'-> 
                      (* if genIdirectionGlobal has given us a GVar, then it generated it fresh and we should add
                          it as an indirection variable*)
                      [gi']
                      
                 | _ -> Updatedebug.abort (Updatedebug.bug "genIndirectionForGlobal_fold : indirection var for %s is not a GVar/GVarDecl ??\n" 
                                     functionName)
                in
                let gf = FunctionInfo.getForwardDeclVar functionName in
                let wr = match FunctionInfo.getWrapper functionName with 
                    None -> []
                  | Some thing -> [thing]
                in
                  ((gf::gi)  @ wr @ [g] @ acc,  nrUpdatable', nrTotal') 
              with Not_found ->
                Updatedebug.abort (Updatedebug.bug "genIndirectionForGlobal_fold : couldn't find indirection var for %s in hash!\n" 
                                     functionName)
            end

      | GVarDecl(vi, loc) -> 
          (* if this is a forward function declaration, 
             we have to forward declare the indirection variable too *)
          begin
            try
              let indirectionVarForwardDeclGlobal = 
                let indirectionVar = FunctionInfo.getIndirectionVar vi.vname in
                  GVarDecl(indirectionVar, Cil.locUnknown)
              in
              let wrapperFunctionForwardDeclGlobal = 
                match FunctionInfo.getWrapperVar vi.vname with
                    None -> []
                  | Some wrapperFunVar -> [GVarDecl(wrapperFunVar, Cil.locUnknown)]
              in
                (indirectionVarForwardDeclGlobal :: g :: (wrapperFunctionForwardDeclGlobal @ acc), 
                 nrUpdatable, 
                 nrTotal)

            with Not_found | FunctionInfo.Non_static_function _ -> (* these exceptions don't mean error, as we probe the hash for any global avariable *)
              id
          end
      | _ -> id


  in  
  let (newGlobals, numUpd, total) = List.fold_right genIndirectionForGlobal_fold gl ([],0,0) 
  in
    ignore(Updatedebug.log [Updatedebug.BTStatistics]
             "# Functions : Address taken and need wrapping = %d, Total = %d \n"
             (total - numUpd) total);
    newGlobals


(* This visitor replaces all calls to functions by the rule:

   f(x) -> ( *__indirect_f)(x)

*)

class replaceFunCallsWithIndirectCallsVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  method vinst i =
    match i with
        Call(result, Lval(Var vi, NoOffset), actuals, loc) ->
            let functionName = vi.vname in 
              begin
                try
                  begin
                    if FunctionInfo.isStatic functionName 
                    then
                      begin
                        try
                          (* do the f(x) -> ( *__indirect_f)(x) *)
                          let gi = FunctionInfo.getIndirectionVar functionName in
                          let new_fexpr = FunctionInfo.buildIndirectionCall gi vi.vtype in 
                          ChangeTo [Call(result, new_fexpr, actuals, loc)]
                            
                        with Not_found -> Updatedebug.abort (Updatedebug.bug "replaceFunCallsWithIndirectCallsVisitor : couldn't find indirection var for %s in hash!\n" 
                                                           functionName)
                      end
                    else
                      DoChildren
                  end
                with Not_found -> 
                  Updatedebug.abort (Updatedebug.bug "replaceFunCallsWithIndirectCallsVisitor : couldn't find whether %s is static!\n" 
                                       functionName)  
              end
      | _ -> DoChildren

end


(* functions that take varargs; troublemakers, see below *)
let varArgFunctionNames = H.create 127

(* This visitor replaces all variables representing non-updatable functions by the rule:

   f -> f'

   Vararg functions are exempted from this treatment since vararg+wrappers = trouble. 
   TODO: investigate why !
*)

class replaceNonUpdatableFunctionsWithWrappersVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  method vvrbl v =
    begin
      if v.vglob (* cut down searches for non-globals *)
      then
        begin
          let functionName = v.vname in 
            try
              let wr = FunctionInfo.getWrapperVar functionName in
                begin
                  if not (FunctionInfo.isUpdatable functionName) && (not (H.mem varArgFunctionNames functionName))
                  then 
                    begin
                      match wr with
                          None -> Updatedebug.abort (Updatedebug.bug "replaceNonUpdatableFunctionsWithWrappersVisitor : %s is non-updatable but doesn't have a wrapper!\n" 
                                                     functionName)
                        | Some vw -> ChangeTo vw
                    end
                  else DoChildren
                end
            with Not_found | FunctionInfo.Non_static_function _ -> (* these exceptions don't mean error, as we probe the hash for any global avariable *)
              DoChildren
        end
      else 
        DoChildren
    end
end


(* This will generate stmts for inserting lock/unlock calls *)
let makeLockUnlockStmt (v : Cil.varinfo) (lockUnlockVi : Cil.varinfo) (functionVersion : int) (loc : Cil.location) =
   (* the indirection var  (fun list head) is passed as first argument to lock/unlock *)
   let indirectionVar = FunctionInfo.getIndirectionVar v.vname in

   (* make an instruction containing the [un]lock() call *)
   let addrOfIndirectionVar_exp = Cil.mkAddrOf (var indirectionVar) in
   let functionVersion_exp = Cil.integer functionVersion in
   let callToLockUnlock = Call(None, Lval(var lockUnlockVi), [addrOfIndirectionVar_exp; functionVersion_exp], loc) in

   (* pack [un]lock () into a statement *)
   Commoncilutil.mkStmt (Instr [callToLockUnlock])


class insertUnlockVisitor (v : Cil.varinfo) (unlockVi : Cil.varinfo) (functionVersion : int) : cilVisitor = object(self)
  inherit nopCilVisitor

  method vstmt s = match s.skind with 

       Return (exp_opt,loc) ->
          let loc = !Cil.currentLoc in (* this is valid since we're in a visitor *)
               
          let unlockStmt = makeLockUnlockStmt v unlockVi functionVersion loc in

          (* pack the unlock() and the original Return into a block *)
          let newSBlock = Commoncilutil.mkStmt (Block (mkBlock ([unlockStmt; s]))) in

          ChangeTo newSBlock

     | _ -> DoChildren


end


let populateFunctionHash (vd : Versiondata.version_data) (f : file) : unit = 
   let isva v = match v.vtype with
         TFun(_,_,res,_) -> res
      | _ -> false in

   let addIfIsFunction (g : Cil.global) : unit = 
    match g with

        GVarDecl(vi, loc) when Cil.isFunctionType (Cil.unrollType vi.vtype) -> 
         let () = 
            if isva vi
            then (H.add varArgFunctionNames vi.vname ();
                  ignore (Updatedebug.log [Updatedebug.BTIndirection] "varArgFunctionName: %s \n" vi.vname))
         in
         begin  match Updateargs.args.Updateargs.arg_currentMode with
              
              Updateargs.NormalMode ->

                 begin
                    (* this might be an external function declaration; if so , add it as extern *)
                    if vi.vstorage = Extern 
                    then FunctionInfo.addExternFunction vi.vname
                 end

           | Updateargs.PatchMode ->
                begin
                   (* this is a function that hasn't changed; pretend it's static *)
                   if vi.vstorage = Extern 
                   then 
                      begin
                         if Versiondata.isExistingFunction vd vi.vname
                         then FunctionInfo.addStaticFunction (Some vd) vi.vname g vi
                         else FunctionInfo.addExternFunction vi.vname
                      end
                end

           | _ -> ()
         end
      (* add as static, unless it's a con/abs/update *)
      | GFun(fd, loc) ->
         (* if the function is a vararg write it down and proceed with caution when 
            replacing it with funwrappers *)
         let functionName = fd.svar.vname in
         let () = 
            if isva fd.svar
            then (H.add varArgFunctionNames fd.svar.vname ();
                  ignore (Updatedebug.log [Updatedebug.BTIndirection] "varArgFunctionName: %s \n" fd.svar.vname))
         in
         if (Namedtypes.isConAbsFuncName functionName) 
            then FunctionInfo.addBuiltinFunction functionName 
         else if (Updateconfig.isUpdateFunc functionName)
            then FunctionInfo.addExternFunction functionName 
         else if (Updateconfig.isEmptyTT functionName) 
            then FunctionInfo.addBuiltinFunction functionName 
         else if (Updateconfig.state_xform_function_name = functionName) 
            then FunctionInfo.addBuiltinFunction functionName 
         else
             FunctionInfo.addStaticFunction (Some vd) functionName g fd.svar

      (* not a function; don't do anything *)
      | _ -> ()
  in
    begin

      (* we have to leave alone builtin functions that are baked into GCC 
         Cil has a hash of those : Cil.msvcBuiltins/Cil.gccBuiltins   *)
      H.iter 
        (fun x y -> FunctionInfo.addBuiltinFunction x)
        (if !Cil.msvcMode then Cil.msvcBuiltins else Cil.gccBuiltins);

      (* iterate thru all globals and fill the hash with fun info *)
      Cil.iterGlobals f addIfIsFunction;
    end

let insertIndirections (vd : Versiondata.version_data) (f : file) : unit =
  begin
    (* populate function info hash with function names and kinds *)
    populateFunctionHash vd f;
    
    (* read-only visitor that marks non-updatable functions *)

    (* Iulian : keep this out of the visitor to avoid creating new visitors for each
       expression we visit; that'd be very inefficient
    *)
    let shooter = new detectNonUpdatableFunctionsInExpressionVisitor vd in

    (* detectNonUpdatableFunction is called for every expression 
       in which we might take the address of a function *)
    let detectNonUpdatableFunction (e : Cil.exp) : unit = 
    ignore(Cil.visitCilExpr shooter e) in

    Cil.visitCilFileSameGlobals (new detectNonUpdatableFunctionsVisitor detectNonUpdatableFunction) f;

    (* side-effecting visitor that will replace non-updatable functions with their wrappers 
       so the 'non-updatable' functions are rendered updatable *)
    Cil.visitCilFile (new replaceNonUpdatableFunctionsWithWrappersVisitor) f;

    (* side-effecting visitor that will replace function calls with indirect calls *)
    Cil.visitCilFile (new replaceFunCallsWithIndirectCallsVisitor) f;

    (* insert indirection variables/function wrappers *)
    f.globals <- addGlobalIndirectionVariables vd f.globals;

    (* Function GC *)

    (* add extern decls for lock/unlock functions *)   
    let lockVi = FunctionInfo.getLock_vi () 
    and unlockVi = FunctionInfo.getUnlock_vi () in
    let gvar_lockVi = GVarDecl(lockVi, Cil.locUnknown)  
    and gvar_unlockVi = GVarDecl(unlockVi, Cil.locUnknown) in

    f.globals <- gvar_lockVi :: gvar_unlockVi :: f.globals;

    (* and finally add FVersionList structure to the globals *)
    f.globals <-  (FunctionInfo.fVersionList_global ()) :: f.globals;

  end

let shouldConsider funName = 
   try
      Funversioning.shouldRename funName && not (FunctionInfo.isWrapper funName)
   with Not_found ->
      Updatedebug.abort (Updatedebug.bug "Indirection.shouldConsider : FunctionInfo.isWrapper returned Not found '%s' ??\n" funName)

(* Infrastructure for garbage collecting functions *)
let addFunctionGCinfrastructure (vd : Versiondata.version_data) (f : file) : unit =
   let addLockUnlockToFunction_iter (lockVi : Cil.varinfo) (unlockVi : Cil.varinfo) (g : Cil.global) : unit = 
         match g with
               GFun(fd, loc) when (not ((Ttpragma.isTT fd.svar.vname) || (Ttwrapinfo.isTTWrapper fd.svar.vname))) && 
                                  shouldConsider fd.svar.vname ->
                  (* we need to do two things
                     1. add the lock stmt at the beginning of the body
                     2. add the unlocks before each return
                  *)                 

                  let functionVersion = Versiondata.getFunctionVersion vd fd.svar.vname in

                  let insertUnlockVisitor = new insertUnlockVisitor fd.svar unlockVi functionVersion in
                  let bodyWithUnlockInserted = Cil.visitCilBlock insertUnlockVisitor fd.sbody in
                  let lockStmt = makeLockUnlockStmt fd.svar lockVi functionVersion loc in


                  (* insert the lock at the beginning of original function body *)
                  bodyWithUnlockInserted.bstmts <- (lockStmt :: bodyWithUnlockInserted.bstmts);
                  fd.sbody <- bodyWithUnlockInserted;

            | _ -> ()    
   in

   let lockVi = FunctionInfo.getLock_vi () 
   and unlockVi = FunctionInfo.getUnlock_vi () in


   (* insert lock/unlock calls *)   
   Cil.iterGlobals f (addLockUnlockToFunction_iter lockVi unlockVi)



