(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)




(* basic requirements to be a monad *)
module type Monad =
sig
  (* monad input / output types *)
  type mi
  type 'a mo
  type 'a m = mi -> 'a mo

  (* basic monadic operations *)
  val (>>=)  : 'a m -> ('a -> 'b m) -> 'b m
  val return : 'a -> 'a m
end




(* Functor that given a monad, generates the monadic operations over it *)
module MonadOps(M:Monad) =
struct
  (* Make the monad bind operation a local infix operator *)
  let (>>=) = M.(>>=)


  (* lift *)
  let lift (f: 'a -> 'b) (m: 'a M.m) : 'b M.m = 
    m >>= fun (x:'a) -> M.return (f x)  


  (* lifted fold left *)
  let rec foldl
      (mop : 'a -> 'b -> 'a M.m)  (* monadic operation *) 
      (acc : 'a)                  (* accumulator *)
      (l : 'b list)               (* list to fold over *)
      : 'a M.m
  =
    match l with
      | []     -> M.return acc
      | x::xs  -> mop acc x >>= fun res -> foldl mop res xs
          

  (* lifted fold right *)
  let rec foldr 
      (mop  : 'b -> 'a -> 'a M.m) 
      (init : 'a) 
      (l : 'b list)
      : 'a M.m
  = 
    match l with 
      | []     -> M.return init
      | x::xs  -> foldr mop init xs >>= fun res -> mop x res
        

  
  let rec map 
      (mop : 'a -> 'b M.m)
      (list : 'a list)
      : 'b list M.m
    = 
    match list with
        [] -> M.return []
      | (x::xs) -> 
          mop x      >>= function y ->
          map mop xs >>= function ys ->
          M.return (y::ys)
  
  let rec iter 
      (mop : 'a -> unit M.m)
      (list : 'a list)
      : unit M.m
    = 
    match list with
        [] -> M.return ()
      | (x::xs) -> 
          mop x      >>= function () ->
          iter mop xs 

  (* lifted iteration 
  let iter 
      (mop : 'a -> unit M.m) 
      (l : 'a list) 
      : unit M.m 
  = 
    foldl (fun acc elm -> mop elm >>= fun () -> M.return ()) () l >>= fun res -> 
    M.return res
  *)


  (* Optionally apply monad function to option type *)
  let apply_opt (f:'a -> 'b M.m) (o:'a option) : ('b option) M.m = 
    match o with 
        None -> M.return None
      | Some a -> f a >>= fun b -> M.return (Some b) 


end





(* 
   We now define some specific monads that we use in the coding of Proteus 
*)

(* SyntaxMonad is used for monads involved with traversing the CIL syntax tree *)
module type SyntaxMonad = 
sig

  (* this is everything Monad is, plus some *)
  include Monad

  (* Set/unset/get the current function (for debugging) *)
  val setCurrentFun : Cil.fundec -> unit m
  val unsetCurrentFun : unit m
  val getCurrentFun : Cil.fundec m

  (* Get/set the current location (for debugging) *)
  val setLoc : Cil.location -> unit m
  val getLoc : Cil.location m

end




(* the identidy monad *)

module IDMonad =
struct
  (* monad input / output types *)
  type mi = unit
  type 'a mo = 'a
  type 'a m = mi -> 'a mo

  (* basic monadic operations *)
  let return (x: 'a) : 'a m = fun () -> x
  let (>>=) (m: 'a m) (f: 'a -> 'b m) : 'b m = fun mi -> f (m mi) mi
end

