(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Solver for Mu-constraints *)

module H = Hashtbl

(* U = MuConst true    N = MuConst false *)
type mu = MuConst of bool  |  MuVar of int    (* updatability variable *)

(* SConst is MuConst 
   Svar ref None is a 'unsolved' variable
   Svar ref Some (T|F) is a 'solved' variable
*)
type solver_mu = SConst of bool | SVar of int * (bool option ref)

type mu_constraint = mu * mu

(* Ordered type for constraints on mu's *)
module MuConstraint_OT (*: Set.OrderedType*) = 
struct
   type t = mu_constraint
   let compare x y = Pervasives.compare x y
end

module MuConstraintSet = Set.Make(MuConstraint_OT)

(* signals an inconsistent set of constraints *)
exception MuSolverFailure of string

(************** NOTE *************************
Mu contraints are of form ( mu_1, mu_2 ) where mu_1 and mu_2
can be variables or constants. Constants can only be elements of
the two-point lattice

    U (true)
    |
    |
    N (false) 

We propagate constants according to the constraints. In the end, 
all unconstrained variables are given value U 
(because we want the most liberal solution, i.e. as many U functions as possible)

Here's how we handle constraints

let solve_one_constraint (mu_1, mu2) = match mu_1, mu_2 with

   (const c_1, const c_2)  -> if c_1 <> c_2 then fail "Inconsistent system" else nop

 | (const c_1, var v_1)    -> nop       (* because we want to leave v_1 unconstrained *)

 | (var v_1, const c_1)    -> v_1 = c_1 (* we have to; if c_1 is N, then v_1 must be N; else v_1 is U which is good *)
 
 | (var v_1, var v_2)      -> let v_1' = unfold v_1 and v_2'= unfold v_2 in 
                              solve_one_constraint v_1' v_2'
  
                              (* unfold enters here because variables can be solved already, i.e.
                                 they can be bound to a constant or they can still be free
                               *)

*********************************************)


let string_of_mu (mu:mu) : string = 
   match mu with 
     MuConst b -> string_of_bool b
   | MuVar id  -> string_of_int id;;


let string_of_solver_mu = function
     SConst c -> "SConst(" ^ (string_of_bool c) ^ ")"

   | SVar (id, r) -> 
        let string_r = match !r with 
              None -> "None"
           | Some c -> "Some(" ^ (string_of_bool c) ^ ")"
        in
        "SVar(" ^ (string_r) ^ ")";;

let mu_to_solver_mu (h : (int, solver_mu) H.t) (mu : mu) : solver_mu = 
   match mu with
      MuConst b -> SConst b
    | MuVar i -> 
        begin 
        try
           H.find h i 
        with Not_found -> 
           let res = SVar (i, (ref None)) in
           H.add h i res;
           res
        end

(* Given a constraint x <= y, resolve it and return true if any of the mu's have changed *)
let resolve ((x,y) : solver_mu * solver_mu) (acc:bool) : bool =
(*    let () = Printf.printf "Resolve %s  %s\n" (string_of_solver_mu x) (string_of_solver_mu y) in *)
   (* assigns constant c to variable v *)
   let assign v c = 
      begin match !v with
         Some c' -> 
            if c <> c'
            then raise (MuSolverFailure "Inconsistent")
            else acc
       | None -> 
          v := Some c;
          true (* we've solved one *)
      end
   in
   match x,y with
      SConst cx, SConst cy -> 
         if cx <> cy 
         then raise (MuSolverFailure "Inconsistent")
         else acc
   
    | SVar (_,v), SConst c -> assign v c
    | SConst c, SVar (_,v) -> acc (* assign v c    see note at the beginning of file *)
     
    | SVar (_,v1), SVar (_,v2) -> 
       begin match !v1, !v2 with
          None, None -> acc
        | Some c, None -> acc      (*assign v2 c   see note at the beginning of file *)
        | None, Some c -> assign v1 c
        | Some c1, Some c2 -> assign v1 c2 
       end
   

 
(* Given a constraint set, return a solution*)
let solveMu (constraint_set : MuConstraintSet.t) : (int,bool) H.t = 
   (* maps var ids to solver_mus *) 
   let solver_mus_hash : (int, solver_mu) H.t = H.create 4099 in

   (* maps var ids to bools; this is the output *)
   let res : (int, bool) H.t = H.create 4099 in
   
   (* this will be our list of constraints *)
   let solver_workset : (solver_mu * solver_mu) list = 
      let mu_constraint_to_solver_mu_constraint ((x,y) : mu * mu) (acc : (solver_mu * solver_mu) list) : (solver_mu * solver_mu) list = 
         let x' = mu_to_solver_mu solver_mus_hash x 
         and y' = mu_to_solver_mu solver_mus_hash y in
         (x',y')::acc
      in
      MuConstraintSet.fold mu_constraint_to_solver_mu_constraint constraint_set []
   in
   (* iterate the set of constraints until nothing changes (or we get an exception) *)
   let _ = 
      let rec solve (l : (solver_mu * solver_mu) list) : bool = 
         if (List.fold_right resolve l false)
         then solve l
         else false
      in
      solve solver_workset
   in
   let add_var_to_res id solver_mu =
      let add id c = 
         if not (H.mem res id)
         then H.add res id c
      in
      match solver_mu with
         SConst c -> add id c
       | SVar (id',_) when id <> id' -> raise (Failure "solveMu: bad hash")
       | SVar (id, r) -> 
          begin match !r with
             None   -> add id true (* Undetermined variables are set to true=U *)  (* raise (MuSolverFailure Undetermined) *)
           | Some c -> add id c
          end            
   in    
   H.iter add_var_to_res solver_mus_hash;
   res
  

let printMuSolution (mu_solution : (int,bool) H.t) : unit =
   H.iter (fun x y -> ignore(Updatedebug.log [Updatedebug.BTMusolver] "%d : %b\n" x y)) mu_solution


let printConstraint (mu1,mu2) = 
   ignore(Updatedebug.log [Updatedebug.BTMusolver] "MuConstraint: %s < %s\n" (string_of_mu mu1) (string_of_mu mu2))



let printMuConstraintSet (mcset : MuConstraintSet.t) : unit =
   MuConstraintSet.iter printConstraint mcset



(* Standalone testing *)
(* let set = MuConstraintSet.empty;; *)
(* let f = MuConst false;; *)
(* let t = MuConst true;; *)
(* let alpha = MuVar 1;; *)
(* let beta  = MuVar 2;; *)
(* let gamma = MuVar 3;; *)
(* let delta = MuVar 4;; *)

(* let set = MuConstraintSet.add (alpha,beta) set;; *)
(* let set = MuConstraintSet.add (beta,gamma) set;; *)
(* let set = MuConstraintSet.add (gamma,delta) set;; *)
(* let set = MuConstraintSet.add (delta,alpha) set;; *)
(* (\* let set = MuConstraintSet.add (delta,t) set;; *\) *)

(* let set = MuConstraintSet.add (delta,f) set;; *)
(* let res = solveMu set in *)
(*    H.iter (fun x y -> Printf.printf "%d : %b\n" x y) res; *)
(*    raise (Failure "Standalone Musolver testing: we won't run CIL");; *)
