(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Tracking information about named types in the program (structs/enums/unions ...) *)

open Cil

module H = Hashtbl


(* Used to select con or abs in makeConAbs *)
type con_or_abs = MAKE_CON | MAKE_ABS;;

type conAbsInfo = 
   {
      ca_info_con          : Cil.fundec;     (* concretisation function *)
      ca_info_abs          : Cil.fundec;     (* abstraction function *)
   };;


(* module that stores information about named types : this is stored imperatively *)
(*====================*)
module NamedTypeInfo =
(*====================*)
struct
      
   (* Hash table storing a map from the namedtype name to the corresponding
      con/abs function definitions *)
   type conAbsHash_t = (string, conAbsInfo) H.t
   let conAbsHash : conAbsHash_t = H.create 4099
   
   (* Map from con/abs function name to the abstract type *)
   type abstractTypesHash_t = (string, Cil.typ) H.t
   let abstractTypesHash : abstractTypesHash_t = H.create 4099

   (* Map from abstract type name to the abstract type *)
   type absNameToTypeHash_t = (string, Cil.typ) H.t
   let absNameToTypeHash : absNameToTypeHash_t = H.create 4099
   
   (* Map from the con/abs c function name to the fundec for that function *)
   type conAbsFundecHash_t = (string, fundec) H.t
   let conAbsFundecHash : conAbsFundecHash_t = H.create 4099
   
   (* Types that are defined externally *) 
   let externalTypes : (string, bool) H.t = (H.create 4099);;

   (* Map to determine updatability of a named type *)
   type updatabilityHash_t =  (string, bool) H.t 
   let updatabilityHash : updatabilityHash_t = H.create 4099

   (* Map from the namedtype name to the representation type of the named type *)
   type representationHash_t = (string, Cil.typ) H.t
   let representationHash : representationHash_t = H.create 4099 

   exception TypeAlreadyExists


   (* ------------------------------------------------------------------------- *)
   (* Access operations                                                         *)
   (* ------------------------------------------------------------------------- *)

   (* associate the canonical type name of some named type with the con/abs that 
      build and destruct objects and the abstract type itself *)
   (*--------------------------------------------------------------------------------------------- *)
   let setConAbs (typeName:string) (ca_info : conAbsInfo) (abs_typ : Cil.typ) (abs_rep_typ : Cil.typ) = 
   (*--------------------------------------------------------------------------------------------- *)
      if H.mem conAbsHash typeName then raise TypeAlreadyExists 
      else
      (* sanity check *)
      let _ = match abs_typ with
         TNamed _ 
       | TComp _ -> ()
       | _ -> Updatedebug.abort (Updatedebug.bug "setConAbs : tried to add junk type %a as abs_typ !" d_plaintype abs_typ)
      in
      begin 
         H.add conAbsHash typeName ca_info;
         
         H.add conAbsFundecHash ca_info.ca_info_abs.svar.vname ca_info.ca_info_abs;
         H.add conAbsFundecHash ca_info.ca_info_con.svar.vname ca_info.ca_info_con;
         
         H.add abstractTypesHash ca_info.ca_info_abs.svar.vname abs_typ;
         H.add abstractTypesHash ca_info.ca_info_con.svar.vname abs_typ;

         H.add absNameToTypeHash typeName abs_typ;

         H.add representationHash typeName abs_rep_typ;

         (*  If the this type name has not had its updatability stratus set, then 
             we set it to the default, which is treu (updatable) *)
         if not (H.mem updatabilityHash typeName) 
         then H.add updatabilityHash typeName true 
      end
   
   
   (* given the canonical type name for a named type, deliver the con/abd functions for it *)
   (*-----------------------------------------------------*)
   let getConAbs (typeName:string) : conAbsInfo = 
   (*-----------------------------------------------------*)
      H.find conAbsHash typeName


   (* Set the status of a named type to be updatable -- the default *)
   (*----------------------------------------*)
   let setTypeUpdatable (typeName : string) = 
   (*----------------------------------------*)
      H.replace updatabilityHash typeName true
   
   
   (* Set the status of a named type to be nonupdatable *)
   (*-------------------------------------------*)
   let setTypeNonUpdatable (typeName : string) = 
   (*-------------------------------------------*)
      (let res = try H.find updatabilityHash typeName with Not_found -> true in
      H.remove conAbsHash typeName;
      H.replace updatabilityHash typeName false;
(*       ignore (Updatedebug.log [Updatedebug.BTAlways] "*************setTypeNonUpdatable: %s\n" typeName); *)
      res)
      
  
   (*-----------------------------------*)
   let isUpdatable (typeName : string) =
   (*-----------------------------------*)
      H.find updatabilityHash typeName



   (* Declare a type as external, which automagically makes it non-updatable *)
   (*----------------------------------*)
   let setTypeExternal (name: string) =
   (*----------------------------------*)
      H.remove conAbsHash name;
      H.replace externalTypes name true;
(*       ignore (Updatedebug.log [Updatedebug.BTAlways] "*************setTypeExternal: %s\n" name); *)
      setTypeNonUpdatable name
  

   (* true if the type name is external *)
   (*----------------------------------*)
   let isExternalType (name : string) = 
   (*----------------------------------*)
      H.mem externalTypes name


   (* fold f over the conAbs hash table *)
   (*---------------------------------------------------------------*)
   let foldConAbs (f : string -> conAbsInfo -> 'a -> 'a) (init:'a) =
   (*---------------------------------------------------------------*)
      H.fold f conAbsHash init


   (* get the representation type of a named type given its name *)
   (*---------------------------------------------*)
   let getRepresentationType (name:string) : Cil.typ = 
   (*---------------------------------------------*)
      H.find representationHash name

   (* Get the abstract type associated with a given con/abs function name *)
   (*---------------------------------------------*)
   let getAbstractType (conAbsName:string) : Cil.typ = 
   (*---------------------------------------------*)
      H.find abstractTypesHash conAbsName

   (* Get the abstract type given the abstract type name returned by Commonutil.getAbstractTypeName *)
   (*-------------------------------------------*)
   let getAbsTypeFromName (tname:string) : Cil.typ = 
   (*-------------------------------------------*)
      H.find absNameToTypeHash tname

end




(* See corresponding entry in NamedTypeInfo *)
let getAbsTypeFromName (tname:string) : Cil.typ = NamedTypeInfo.getAbsTypeFromName tname


(* Determine if t is abstract *)
(*--------------------*)
let isAbstractType t = 
(*--------------------*)
  begin
    let res =      
        match t with
            TNamed (ti,_) -> true 
          | TComp(ci, _) -> true
          | TBuiltin_va_list _ -> true
          | _ -> false
    in
      if Updateconfig.debug_AbstractType 
      then ignore (Updatedebug.log [Updatedebug.BTNamedtypes] "isAbstractType (%a) : %b\n" d_plaintype t res);
      res
  end


(* true if the type is updatable; false if not updatable or not in the hash (generated by us) 
   N.B. this shouldn't be called before ptraliastracking has identified all non-updatable types !
   and conabsinsertion has flushed all updatable types *)
(*-------------*)
let isUpdatable t =
(*-------------*)
   if isAbstractType t then
      let type_name =  Updatecilutil.getAbstractTypeName t in  
      try
         NamedTypeInfo.isUpdatable type_name
      with 
         Not_found -> false
   else 
      Updatedebug.abort (Updatedebug.bug "Tried to see if non-abstract type %a was updatable!\n" d_plaintype t)  


(* check to see if a name returned by getAbstractTypeName is updatable *)
(*-------------------------------*)
let isUpdatableByAbsName t_abs_name =
(*-------------------------------*)
   NamedTypeInfo.isUpdatable t_abs_name

   
(*-----------------------------------------------------*)
let setAbsTypeNameExternal (abs_t_name:string) loc reason : unit = 
(*-----------------------------------------------------*)
   if Forcedpragma.isForcedUpdatable abs_t_name 
   then
      (* Error. We got a 'set extern' request for a forced updatable type; they're incompatible *)
      Updatedebug.abort (Updatedebug.error "%a : setTypeExternal : Trying to set forced-updatable type '%s' extern (%s)!\n" d_loc loc abs_t_name reason)
   else
      begin
         if (NamedTypeInfo.setTypeExternal abs_t_name) then
            ignore (Updatedebug.log [Updatedebug.BTAlways] 
                       "(%a) type external (%s):  %s\n"  d_loc loc reason abs_t_name);
         ()
      end

(* Add a type to the external types list *)
(*---------------------*)
let setTypeExternal t loc reason = 
(*---------------------*)
   setAbsTypeNameExternal (Updatecilutil.getAbstractTypeName t) loc reason

(* test to see if a type has been set external *)
(*--------------------*)
let isExternalType t =  
(*--------------------*)
   if isAbstractType t 
   then 
      NamedTypeInfo.isExternalType (Updatecilutil.getAbstractTypeName t)
   else
      false

let isExternalTypeName = NamedTypeInfo.isExternalType

(* fold f over the conAbs hash table *)
(*---------------------------------------------------------------*)
let foldConAbs (f : string -> conAbsInfo -> 'a -> 'a) (init:'a) =
(*---------------------------------------------------------------*)
   NamedTypeInfo.foldConAbs f init


(* Generate a definition for a comp's representation type *)
(*---------------------------------------------*)
let getRepTypeDef (t : Cil.typ) : Cil.global list =  
(*---------------------------------------------*)
   let doit type_name =
      try
         match NamedTypeInfo.getRepresentationType type_name with 
            TComp (ci,attr) -> [GCompTag (ci,Cil.locUnknown)]   
          | t_rep           ->  []
         with 
            Not_found -> []
   in 
   if isAbstractType t then 
      if isUpdatable t then 
         let type_name = Updatecilutil.getAbstractTypeName t in
         doit type_name
      else []
   else []





(* Recursively turns arrays into pointers  *)
(*----------------*)
let makePtrOfArray t =
(*----------------*)
   let id x = x in 
   let rec makePtrOfArray' t' (f:Cil.typ->Cil.typ) = 
   match t' with
      TArray(bt, sz, attr) -> TPtr(makePtrOfArray' bt id, attr)
    | TNamed(ti,attr) -> 
         makePtrOfArray' ti.ttype (fun t'' -> TNamed({ti with ttype = f(t'')}, attr))
    | _ -> f(t')
   in 
   makePtrOfArray' t id



(* determine the representation type of a structure *)
(*----------------------------------------*)
let rec getCompRepType (t:Cil.typ) : Cil.typ = 
(*----------------------------------------*)
   let type_name = Updatecilutil.getAbstractTypeName t in
   try 
      NamedTypeInfo.getRepresentationType type_name 
   with 
      Not_found -> 
         Updatedebug.abort (Updatedebug.bug "getCompRepType: Tried to get rep type for %a but failed" d_plaintype t)  


(* Unroll an abstract type to its representation type.
   Might drop some attributes !! *)
(*--------------------------*)
and unrollTypeOnce (t:Cil.typ) = 
(*--------------------------*)
   match t with 
     TNamed (r, _) -> r.ttype
   | TComp (ci,attrs) as t -> getCompRepType t
   | x -> x


(* 
   Return a pair of CIL ASTs for abs and con coercions for the given named type  
*)
(*-------------------------*)
and getConAbs loc (t:Cil.typ) = 
(*-------------------------*)
  
  (* check to see if t is abstract *)
  if not (isAbstractType t) 
  then  
     Updatedebug.abort (Updatedebug.bug "(%a) Tried to get con/abs coercion for the non abstract type:\n%a\n" d_loc loc d_plaintype t)  
  else 
  begin
     let type_name = Updatecilutil.getAbstractTypeName t in
     try
        NamedTypeInfo.getConAbs type_name 
     with Not_found ->
        let nt = Commoncilutil.sanitizeTypeName (Commoncilutil.cTN t) in
            
        let createRepType t = 
          match t with 
            TComp(ci,attr) ->
	      TComp (copyCompInfo ci (Updateconfig.makeConcreteCompName ci),
		     attr)
           | TNamed(ti,attr) -> ti.ttype
           | _ -> t in 
   
        let abs_rep_type = createRepType t in
             
        let conName = Updateconfig.makeConName nt in
        let absName = Updateconfig.makeAbsName nt in
        
        
        (* generate an empty function -- will be repalced at wrapper stage with real thing *)
        (*------------------------------------------------------------------------------------------------*)
        let makeConAbs (in_t:Cil.typ) (out_t:Cil.typ) (name:string) (conOrAbs : con_or_abs) : Cil.fundec = 
        (*------------------------------------------------------------------------------------------------*)
        
           (* Return types can't be arrays *)
           let out_t' = makePtrOfArray out_t in
           let newFunction  = Cil.emptyFunction name in 
	   if Updateargs.args.Updateargs.arg_dostaticconabs then 
	     (newFunction.svar.vstorage <- Static;
	      newFunction.svar.vinline <- true);
           let theBody = 
              begin match conOrAbs with 
                 MAKE_CON -> 
                    let newFormal = Cil.makeFormalVar newFunction ~where:"^" "abs" in_t in
                    newFunction.svar.vtype <- TFun(out_t', Some [("abs",in_t,[])], false, []);
                    (* This will be replaced when adding wrappers so that it returns .data *)
                    [Commoncilutil.mkStmt 
                       (Return (Some
                           (CastE( out_t', (Lval(var newFormal)))), Cil.locUnknown))] 
        
               | MAKE_ABS -> 
		   (* We need to treat abs functions for TComps and NamedT differently. *)
		   begin 
		     match t with TComp _ -> 
		       let newFormal = Cil.makeFormalVar newFunction ~where:"$" "abs" (TPtr(out_t,[])) in
		       newFunction.svar.vtype <- TFun(Cil.voidType, Some [("abs",(TPtr(out_t,[])),[])], false, []);
                       [Commoncilutil.mkStmt (Return (None, Cil.locUnknown))]
		     | _ ->
		       let newFormal = Cil.makeFormalVar newFunction ~where:"^" "con" in_t in
                       let formalToCoerce = Cil.makeFormalVar newFunction ~where:"$" "abs" (TPtr(out_t,[])) in
                       newFunction.svar.vtype <- TFun(Cil.voidType, Some [("con",in_t,[]); ("abs",(TPtr(out_t,[])),[])], false, []);
                       [Commoncilutil.mkStmt (Instr [Set( ( Mem( Lval( var formalToCoerce ) ), NoOffset ), 
							  Lval(
							  Mem(
                                                          CastE(TPtr(out_t,[]), Cil.mkAddrOf(var newFormal))), 
                                                          NoOffset), 
							  Cil.locUnknown)]);
			Commoncilutil.mkStmt (Return (None, Cil.locUnknown))]
		   end
              end
           in
        
           newFunction.sbody <- {battrs=[]; bstmts = theBody};
         
           ignore (Updatedebug.log [Updatedebug.BTNamedtypes] 
                                   "Will generate coercion <%s> (%s:%d) with type %a\n" 
                                   name !(Cil.currentLoc).file !(Cil.currentLoc).line 
                                   d_plaintype newFunction.svar.vtype);
           newFunction
        
        in
        
                
        let t' = match Cil.unrollType t with
           TFun _ -> TPtr(t, [])
          | _ -> t 
        in
        
        let conFunction : Cil.fundec = makeConAbs (TPtr(t,[])) (TPtr(abs_rep_type,[])) conName MAKE_CON in 
        let absFunction : Cil.fundec = makeConAbs abs_rep_type t' absName MAKE_ABS in
        let ca_info =   {ca_info_con = conFunction; ca_info_abs = absFunction} in
        NamedTypeInfo.setConAbs type_name ca_info t abs_rep_type; 
        (*ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                   "Generated coercions for <%s>\n Abstract type is %a\n Concrete type is %a \n\n" 
                                   type_name 
                                   d_plaintype t
                                   d_plaintype abs_rep_type);*)

        ca_info
   end



(* Get con/abs functions *)
(*------------------*)
let getCon loc (t:typ) =  
(*------------------*)
   (getConAbs loc t).ca_info_con;;

(*------------------*)
let getAbs loc (t:typ) = 
(*------------------*)
   (getConAbs loc t).ca_info_abs;;




(* Mark a named type as non-updatable; if recursive = true, it goes down in TNamed chains 
   and marks them nonupdatable too *)
let rec setTypeNonUpdatableInternal t added_type recursive loc reason = 
   if Forcedpragma.isForcedUpdatable (Updatecilutil.getAbstractTypeName t) 
   then (* ignore *)
      ignore (Updatedebug.log [Updatedebug.BTAlways] 
                 "(%a) : Warning : setTypeNonUpdatable(%s) (%s) request dropped because this type was forced updatable \n" d_loc loc (Updatedebug.ppType t) reason)
         
   else
      begin
         (* Continue pinning down recursively if the next in the chain is abstract *)
         let type_name =  Updatecilutil.getAbstractTypeName t in 
         let added_type' = NamedTypeInfo.setTypeNonUpdatable type_name in

         ignore (Updatedebug.log [Updatedebug.BTAlways] 
                    "(%a) setTypeNonUpdatable(%s) (%s)\n" d_loc loc (Updatedebug.ppType t) reason);

         if recursive then 
            begin
               match t with
                     TNamed(ti,attr) when isAbstractType ti.ttype -> 
                        setTypeNonUpdatableInternal  ti.ttype added_type' true loc reason
                  | _ -> ()
            end
      end;; 


(* Mark a named type as non-updatable; goes down in TNamed chains and marks them nonupdatable too *)
(*-----------------------------*)
let setTypeNonUpdatable t loc reason = 
(*-----------------------------*)
   setTypeNonUpdatableInternal t false true loc reason

(* Mark a named type as non-updatable; doesn not go down in TNamed chains *)
(*-----------------------------*)
let setTypeNonUpdatableNonRecursive t loc reason = 
(*-----------------------------*)
   setTypeNonUpdatableInternal t false false loc reason


(* Set an abstract type name non-updatable.  This abstract type name must be of the form returned by getAbstractTypeName,
   e.g. for a structure T it returns struct_T. *)
(*-----------------------------------------------------*)
let setAbsTypeNameNonUpdatable (abs_t_name:string) loc reason : unit = 
(*-----------------------------------------------------*)
   if Forcedpragma.isForcedUpdatable abs_t_name
   then (* ignore *)
      ignore (Updatedebug.log [Updatedebug.BTAlways] 
                 "(%a) : Warning : setTypeNonUpdatable(%s) (%s) request dropped because this type was forced updatable \n" d_loc loc abs_t_name reason)

   else
      begin
         if (NamedTypeInfo.setTypeNonUpdatable abs_t_name) then
            ignore (Updatedebug.log [Updatedebug.BTAlways] 
                       "(%a) type non-updateable (%s): %s\n"  d_loc loc reason abs_t_name);
         ()
      end
 
(* Given a type t, it transitively  concretises t *)
let concretize t = Cil.unrollType t


(* Test to see if a given fucntion name is a con/abs function *)
let isConAbsFuncName conAbsName = 
   try
      ignore(NamedTypeInfo.getAbstractType conAbsName);
      true
   with 
      Not_found -> false


(* Given a con/abs function name return the abstract type associated with it.
   Might raise Not_found.*)
(* was: getConAbsTypeByName *)
(*--------------------------------*)
let getAbstractTypeByConAbsName conAbsName = 
(*--------------------------------*)
   let res = NamedTypeInfo.getAbstractType conAbsName
   in
    begin
    match res with
         TNamed _ 
       | TComp _ -> ()
       | _ -> Updatedebug.abort (Updatedebug.bug "setConAbs : tried to add junk type %a as abs_typ !" d_plaintype res)
     end;
   res



   


(* return true if t is updatable.  If this type hasn't been seen before set it updatable and 
   return true *)
let isUpdatableOrNotSeen t = 
    if isAbstractType t then
      let type_name =  Updatecilutil.getAbstractTypeName t in  
      try
         NamedTypeInfo.isUpdatable type_name
      with 
         Not_found -> true
   else 
      Updatedebug.abort (Updatedebug.bug "isUpdatableOrNotSeen: Tried to see if non-abstract type %a was updatable (or not seen)!\n" d_plaintype t)  




(* determine if a named type should be treated tranparently (i.e. should we unroll it when looking at its type) *)
(*---------------*)
let isTransparent t = 
(*---------------*)
   (* external/non-updatable types should be transparent *)
   try not (isUpdatableOrNotSeen t) 
   with Not_found -> false



(* Given a con/abs function name determine if the con/abs function is not needed *)
(*----------------------------*)
let isRemovableConAbs conAbsName =
(*----------------------------*)
  if isConAbsFuncName conAbsName
  then 
    let res = not (isUpdatable (NamedTypeInfo.getAbstractType conAbsName)) in
    ignore (Updatedebug.log [Updatedebug.BTNamedtypes] "isRemovable (%s) : %b\n" conAbsName res);
    res
  else 
     false
    
    
(* a type is wrapable iff it exists in the conAbsHash and is updatable *)
(*------------*)
let isWrapable t =
(*------------*)
   (*ignore (Updatedebug.log [Updatedebug.BTNamedtypes] "is wrapable: %a\n" d_plaintype t);*)
   if isAbstractType t && (not (isExternalType t)) then 
      try 
         isUpdatable t
      with 
         Not_found -> 
         (*ignore (Updatedebug.log [Updatedebug.BTNamedtypes] "is wrapable bailed not found\n"); *)
         false
   else false

(* Traverse the structure of a type calling f for all named types and types of array *)
(*----------------------------------------------------------------------------*)
let namedTypeActionIter (follow_ptr : bool) (f:Cil.typ -> 'a) (t:Cil.typ) : unit =
(*----------------------------------------------------------------------------*)
   let compHash = H.create 7 in 

   let rec doit t' = 
      match t' with 
       | TInt _
       | TFloat _
       | TVoid _                  -> ()
       | TEnum (enum, a)          -> ()
       | TPtr (t'', a)             -> if follow_ptr then doit t'' else ()
       | TArray (t'',l,a)          -> f t''; ()
       | TComp (ci, a)            -> 
            if not (H.mem compHash ci.cname) 
            then 
            begin 
               H.add compHash ci.cname true;
               f t'; 
               List.iter (fun fi -> doit fi.ftype) ci.cfields
            end
            
       | TFun(rt,args,isva,a)     -> doit rt; List.iter doit (Commoncilutil.argsToTypesList args) 
       | TNamed(ti, a)            -> f t'; doit ti.ttype 
       | TBuiltin_va_list al      -> f t'; ()
   in 
      doit t


(* Set t non-updatable and proceed recursively on the structure of t setting all
   those abstract types we come accross non-updatable *)
(*------------------------------*)
let rec setTypeNonUpdatableDeep t loc reason =
(*------------------------------*)
   namedTypeActionIter true (fun t' -> if isAbstractType t' then  setTypeNonUpdatable t' loc reason) t

(* same as setTypeNonUpdatableDeep, but don't follow pointers *)
(*------------------------------*)
let rec setTypeNonUpdatableDeepNoPtr t loc reason =
(*------------------------------*)
   namedTypeActionIter false (fun t' -> if isAbstractType t' then setTypeNonUpdatable t' loc reason) t


(* Similar to setTypeNonUpdatableDeep but set it external *)
(*-------------------------*)
let rec setTypeExternalDeep t loc reason = 
(*-------------------------*)
   namedTypeActionIter true (fun t' -> if isAbstractType t' then setTypeExternal t' loc reason) t; ()


(* Determine if a type contains any updatable named types *)
(*------------------------------------------------*)
let containsUpdatableNamedTypes (typ:Cil.typ) : bool = 
(*------------------------------------------------*)
   let compHash = H.create 7 in 
   let rec containsNamedTypes' (typ':Cil.typ) : bool = 
      match typ' with 
         TVoid a -> false
       | TInt (ik, a) -> false
       | TFloat (fk, a) -> false
       | TEnum (enum, a) -> false
       | TPtr (t, a) -> containsNamedTypes' t
       | TArray (t, l, a) -> containsNamedTypes' t
       | TFun (t, args, isva, a) -> (containsNamedTypes' t) || (List.exists containsNamedTypes' (Commoncilutil.argsToTypesList args))
       | TComp (ci, a) -> if H.mem compHash ci.cname then false 
                          else if isUpdatableOrNotSeen typ' then true
                          else 
                             begin 
                             H.add compHash ci.cname ();
                             List.exists (fun fi -> containsNamedTypes' fi.ftype) ci.cfields
                             end
       | TNamed (ti, a) -> 
            let res = isUpdatableOrNotSeen typ'  in
            (*ignore( Updatedebug.log [Updatedebug.BTAlways] "containsUpdatableNamedTypes: notseen %b" res);*)
            res || containsNamedTypes' ti.ttype
       | TBuiltin_va_list a -> false
   in
   let res = containsNamedTypes' typ in 
(*   ignore( Updatedebug.log [Updatedebug.BTAlways] "containsUpdatableNamedTypes: %a = %b\n" d_plaintype typ res);*)
   res

(* iter over the forced hash and kill all types that the user has marked non-updatable *)
let setForcedNonUpdatableTypes () = 
   let setForcedNonUpdatableTypes_iter typeName updatability = match updatability with
         Forcedpragma.ForcedNonUpdatable -> 
            ignore( Updatedebug.log [Updatedebug.BTAlways] "User-forced nonupdatable : %s\n" typeName);
            ignore(NamedTypeInfo.setTypeNonUpdatable typeName)
      | _ -> ()
   in
   H.iter setForcedNonUpdatableTypes_iter Forcedpragma.forcedPragmaHash
