(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*  Compilation of patches

    Given a file containing a patch and some information about the program to patch,
    we attempt to produce a dynamic patch in the form of a .so library.

*)

open Cil

module H = Hashtbl



type patch_global =
{
   pg_fun_hash   : (string, Cil.global) H.t;   (* fun name to global fundec *)
   pg_type_hash  : (string, Cil.global) H.t;   (* fun name to global type decl *)
   pg_data_hash  : (string, Cil.global) H.t;   (* fun name to global data def/decl *)
}


(* information about the patch file *)
type patch_info = 
{
   pi_new           : patch_global;              (* globals that did not exist before and will be added by this patch *)
   pi_replaced      : patch_global;              (* globals that existed before and will be replaced by this patch *)
   pi_extern        : patch_global;              (* globals that existed before and have an external decl here *)
   pi_vd            : Versiondata.version_data;  (* information about running program *)
   pi_tt_arg_types  : (string,string) H.t;       (* map from tt arg type names to the associated abstract type being transformed *)
}


type patch_elm_status = PESNew | PESReplaced | PESExtern
type patch_elm_type = PETFun | PETType | PETData


let create_patch_info vd = 
   let create_patch_global () = 
      { pg_fun_hash = H.create 4099;
        pg_data_hash = H.create 4099;
        pg_type_hash = H.create 4099; }
   in
   { pi_new = create_patch_global ();
     pi_replaced = create_patch_global (); 
     pi_extern = create_patch_global (); 
     pi_vd = vd; 
     pi_tt_arg_types = H.create 4099; }


(************************* patch data structure manipulation functions *****************************)


(* Given the status and type of an element we are interested in, give us the hash for it *)
let getHash elm_status elm_type pi : ('a,'b) H.t = 
   let gp : patch_global = 
      match elm_status with 
         PESNew       -> pi.pi_new
       | PESReplaced  -> pi.pi_replaced
       | PESExtern    -> pi.pi_extern
   in
   match elm_type with 
      PETFun   -> gp.pg_fun_hash
    | PETType  -> gp.pg_type_hash
    | PETData  -> gp.pg_data_hash



let piAdd elm_status elm_type (name:string) (g:Cil.global) pi : unit = 
   H.add (getHash elm_status elm_type pi) name g

let piExists elm_status elm_type (name:string) pi : bool = 
   H.mem (getHash elm_status elm_type pi) name

let piFind elm_status elm_type (name:string) pi : Cil.global = 
   H.find (getHash elm_status elm_type pi) name

let piRemove elm_status elm_type (name:string) pi : unit = 
   H.remove (getHash elm_status elm_type pi) name



(************************************** Data collection phase **************************************)

let collectFunNames (file:Cil.file) : (string,unit) H.t = 
   let funNameHash = H.create 4099 in 
   let collectFunNames_iter (g:Cil.global) = 
      match g with 
         GFun(fd,loc) -> H.add funNameHash fd.svar.vname ()
       | _ -> ()
   in
   Cil.iterGlobals file collectFunNames_iter;
   (funNameHash)

(* given a hash of forward declarations decide if vname is a forward declaration *)
let isForwardFunDeclaration (funNameHash : (string,unit) H.t) (name:string) : bool = 
   H.mem funNameHash name

let genPatchInfoGlobal 
   (funNameHash : (string,unit) H.t) 
   (pi:patch_info) 
   (g:global) : unit =

   let vd = pi.pi_vd in 

   let () = ignore (Updatedebug.log [Updatedebug.BTPatchcomp] "genPatchInfoGlobal(%s)...\n" (Commoncilutil.getGlobalName g)) in
   let addToTypeHash t glob = 
      let tname = Updatecilutil.getAbstractTypeName t in
      (* if it has a type transformer, then we assume it's been changed.  Otherwise, we assume it's the same as before. *)
      let hasTT = begin 
         try 
            ignore(Ttpragma.getTTData (Ttpragma.getTTName tname));
            true
         with Not_found -> 
            false
      end in
      try 
         (* if this is an argument type of a type transformer, then we should associate 
            this with the abstract type it transforms *)
         let t_abs_type = Ttpragma.getAbsTypeFromTTArgType tname in
         if H.mem pi.pi_tt_arg_types tname then 
            Updatedebug.abort (Updatedebug.error "genPatchInfoGlobal: adding TT argument type %s twice!\n" tname)
         else
            H.add pi.pi_tt_arg_types tname t_abs_type
      with 
         Not_found -> 
            (* tname is the name of a regular type, so deal with it as normal *)
            if Versiondata.isExistingType vd tname then
               (* this type existed before, but may be included just for extern use not for update *)
               try
                  if Versiondata.isUpdatableType vd tname then
		    if hasTT then
                      (* this is a replacement type as it existed before *)
                      piAdd PESReplaced PETType tname glob pi
		    else (* make sure the types are equal if no TT *)
                      (let t_orig = Versiondata.getAbsType vd tname in
                      if false (* not (Typeequality.uAreTypesEqualNoAttrs t_orig t) *) then (* FIXME: gross hack: why do we crash here ?*)
                        Updatedebug.abort 
			  (Updatedebug.error 
			     "genPatchInfoGlobal: No type transformer for '%s' which changed from %a to %a\n"
                             tname d_typsig (typeSig t) d_typsig (typeSig t_orig)))
                  else begin
                     (* This is a non-updatable, pre-existing type.  If they don't match
			the user needs to change them all by hand.  Issue warning. *)
		     (* FIX: This isn't working, since types that are changed are not generating
			a warning. *)
                     let t_orig = Versiondata.getAbsType vd tname in   
                     if not (Typeequality.uAreTypesEqualNoAttrs t_orig t) then
                       ignore 
			 (Updatedebug.log [Updatedebug.BTAlways]
			    "Non-updatable type %s changed; must update instances by hand\n" tname)
		  end
               with
                  Not_found -> 
                     Updatedebug.abort (Updatedebug.bug "genPatchInfoGlobal: really shouldn't be here\n")
            else 
               piAdd PESNew PETType tname glob pi
   in 
  
   match g with 
    | GFun(fd,l) ->
         if (not (!Versiondata.isInternalFunction fd.svar.vname)) && not (Cil.startsWith Updateconfig.state_xform_function_name fd.svar.vname) then 
            begin
               if Versiondata.isExistingFunction vd fd.svar.vname 
               then 
                  (* this is a replacement function, as it existed before and is not an internal update function *)
                  (* we have to check wether its signature has changed or not *)
                  begin
                     let oldFtype = Versiondata.getFunctionType vd fd.svar.vname
                     and newFtype = fd.svar.vtype in
                  
                        if true (* Commoncilutil.isEqualFuntypeQuick oldFtype newFtype *) (* ruled out because we now do subtyping/hold the patch application until it's safe to do so *)
                        then
                           (* no sig change, report as replaced *)
                           piAdd PESReplaced PETFun fd.svar.vname g pi
                        else
                           (* sig changed, report as new *)
                           piAdd PESNew PETFun fd.svar.vname g pi 
                  end
               else            
                  (* this is a new function *)
                  piAdd PESNew PETFun fd.svar.vname g pi
            end
         else 
            (* ignore internal functions *)
            ignore (Updatedebug.log [Updatedebug.BTPatchcomp] "Ignoring internal function %s\n" fd.svar.vname)
            
    | GType(ti,l) ->
         addToTypeHash (TNamed(ti,[])) g
         
    | GEnumTag(ei,l) -> ()
    
    | GEnumTagDecl(ei,l) -> ()
    
    | GCompTag(ci,l) ->
         addToTypeHash (TComp(ci,[])) g
    
    | GCompTagDecl(ci,l) -> ()
    
    | GVarDecl(vi,l) -> 
         (* We don't allow global variables to be updated -- this is done in the state transformer --
         but we should check to see that this NEW variable is not already defined *)
         
         (* fix existing, extern function names (i.e. append the version) *) 
         if Versiondata.isExistingFunction vd vi.vname &&
            not (!Versiondata.isInternalFunction vi.vname) &&
            not (isForwardFunDeclaration funNameHash vi.vname)
         then 
            begin
               let existingFunctionVersion = Versiondata.getFunctionVersion vd vi.vname in
               let newViName = Updateconfig.makeVersionedFunctionName vi.vname existingFunctionVersion in
               ignore (Updatedebug.log [Updatedebug.BTPatchcomp] "Fixing existing extern fun decl : %s -> %s\n" vi.vname newViName);
               vi.vname <- newViName
            end
         else
         if Versiondata.isExistingVar vd vi.vname then begin
           ignore (Updatedebug.log [Updatedebug.BTPatchcomp] "%s is an existing var\n" vi.vname);
            if vi.vstorage = Cil.Extern then 
               begin
               (* variable is declared extern in the patch and it exists in the program.  Remember it
                  for later use *)
               ignore (Updatedebug.log [Updatedebug.BTPatchcomp] "Adding %s to extern var decls\n" vi.vname);
               piAdd PESExtern PETData vi.vname g pi
               end
            else if isForwardFunDeclaration funNameHash vi.vname then 
               ()
            else
               (* bail as we are shadowing existing variable *)
               ()
(*                Updatedebug.abort (Updatedebug.bug  *)
(*                "genPatchInfoGlobal: patch contains global variable %a which already exists in the running program (1)\n" d_global g) *)
	 end
         else begin
            (* Brand new variable *)
           ignore (Updatedebug.log [Updatedebug.BTPatchcomp] "Adding %s to new vardecls\n" vi.vname);
           piAdd PESNew PETData vi.vname g pi 
	 end

    | GVar(vi,ii,l) -> 
         (* We don't allow global variables to be updated -- this is done in the state transformer --
         but we should check to see that this NEW variable is not already defined *)
         if Versiondata.isExistingVar vd vi.vname then
            (* alright, the static initializer for this global var has changed *)
            piAdd PESReplaced PETData vi.vname g pi
         else begin
            (* Brand new variable *)
            ignore (Updatedebug.log [Updatedebug.BTPatchcomp] "Adding %s to new var defs\n" vi.vname);
            piAdd PESNew PETData vi.vname g pi 
	 end


    | GAsm _ | GPragma _ | GText _ -> ()



            


let genPatchInfo (file:Cil.file) (pi : patch_info) : unit = 
   let funNameHash = collectFunNames file in
   Cil.iterGlobals file (genPatchInfoGlobal funNameHash pi)









(************************** Generate patch installation code ********************************)


let generatePatchInstallCode (vd : Versiondata.version_data) (pi : patch_info) : (Cil.instr list * string list) = 

   (* Process each global type definition that already exists in the current program.  For each one
      increment the latest version variable and set the type transformer *)
   let genSetLatestTypeVersion_fold name g ((acc_il, acc_changed_types) : (Cil.instr list * string list)) : (Cil.instr list * string list) = 
      (* get the type of the global *)
      let g_t = 
         match g with 
            GType(ti,loc) -> TNamed(ti,[])
          | GCompTag(ci,loc) -> TComp(ci,[])
          | _ -> Updatedebug.abort (Updatedebug.bug "genSetLatestTypeVersion: expected global type definition, but go %a\n" d_global g)
      in
      
      let tname = Updatecilutil.getAbstractTypeName g_t in 

      (* Not all types have type transformers (e.g. because they didn't change) *)
      let hasTT = begin 
         try 
            ignore(Ttpragma.getTTData (Ttpragma.getTTName tname));
            true
         with Not_found -> 
            false
      end in
         
      if hasTT then begin
	ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "genSetLatest: type %s has TT\n" tname);
         let bang s = Updatedebug.abort (Updatedebug.bug "genSetLatestTypeVersion(type version) : (%s) %s\n" tname s) in

         (* create the instruction to increment the current version number *)
         let current_version = 
            try
               Versiondata.getCurrentTypeVersion pi.pi_vd tname
            with 
               Not_found -> bang "type does not exist in current version, but should ..."
         in 
      
         (* What is the name of the variable that stores the latest version of this type in the current program? *)
         let latest_type_ver_var_name = Updateconfig.makeLatestTypeVersionVarName tname in
      
         (* Instructions to set the variable that holds latest version number of this type one higher *)
         let incr_latest_version_il = 
            try
               (* look up the global in the patch that defines the latest version variable as extern *)
               let g_latest_ver_extern = piFind PESExtern PETData latest_type_ver_var_name pi in 
                  begin match g_latest_ver_extern with 
                        GVarDecl(vi,loc) -> 
                           [ Set(var vi, Cil.integer current_version, Cil.locUnknown)]
                           
                     | _ -> bang "the latest type version variable is not defined extern in patch" 
                  end
            with
                  Not_found -> bang "expected latest version variable to be defined extern"
         in

         let bang_tt s = Updatedebug.abort (Updatedebug.bug "genSetLatestTypeVersion(tt) : (%s) %s\n" (Ttpragma.getTTName tname) s) in

         (* insttructions to set the function pointer that points to the highest versioned type transformer *) 
         let set_tt_il = 
            let process_g g = match g with
                     GFun(fd,_) ->
                        let g_tt_vi = Typeindirection.getLatestTypeTransformer tname in
                        let tt_wrapper_name = Ttpragma.getTTName tname in
                        let tt_wrapper = Ttwrapinfo.getTTWrapper (Ttpragma.getTTName tname) in
                        [Set(var g_tt_vi, Cil.mkAddrOf (var tt_wrapper), Cil.locUnknown)]
                  
                  | _ -> 
                       bang_tt "latest TT is not a function"
            in

            let g = 
               begin
                  try 
                     piFind PESNew PETFun (Ttpragma.getTTName tname) pi

                  with Not_found ->
                     begin
                        try 
                           piFind PESReplaced PETFun (Ttpragma.getTTName tname) pi
                        with
                           Not_found -> bang_tt "Couldn't find type transformer in PESNew/PESReplaced"
                     end
               end
            in
            process_g g
         in
         (incr_latest_version_il @ set_tt_il @ acc_il, tname::acc_changed_types)
      end
      else
         (acc_il, acc_changed_types)
   in

   (* given a replacing function, do two things :
      generate the code to set the function indirection variable to point to it 
      add the function name to the "forbidden functions list" if its type has changed.
   *) 
   let genUpdateFunIndirection_fold name g ((acc_il, acc_forbidden_fun) : (Cil.instr list * string list)) : (Cil.instr list * string list) = 
      if (Indirection.shouldConsider name) && not (Ttpragma.isTT name) && not (Cil.startsWith Updateconfig.state_xform_function_name name) then
         begin
            let fd = match g with
                  GFun(f,loc) -> f
               | _ ->  Updatedebug.abort (Updatedebug.bug "genUpdateFunIndirection: expected global function definition, but got %a\n" d_global g)
            in
            let fun_name = fd.svar.vname in
            let bang s = Updatedebug.abort (Updatedebug.bug "genUpdateFunIndirection(%s) %s\n" fun_name s) in

            (* we're in patch mode and Indirection has already run at this point, 
               hence we might know the indirection var from Indirection *)
            let g_indirection_var = 
               try
                  let vi = Indirection.FunctionInfo.getIndirectionVar fun_name in
                     GVarDecl(vi, Cil.locUnknown)  
               with
                     Not_found -> 
                        begin try 
                           let indirection_var_name = Versiondata.getFunctionIndirectionVarName vd fun_name in
                              begin try 
                                 piFind PESExtern PETData indirection_var_name pi
                              with
                                    Not_found -> bang "piFind : Can't find indirection var for this function"
                              end
                        with
                              Not_found -> bang "Can't find indirection var in Versionadata for this function"
                        end
            in 
               begin match g_indirection_var with 
                     GVarDecl(vi,l_var) ->
                        let addFVersionList_svar = Indirection.FunctionInfo.getAddFVersionList_vi () in
                        let indirectVar = vi in
                        let functionName = name in
                        let functionVersion = Versiondata.getFunctionVersion vd fd.svar.vname in

                        (* -1 because we haven't bumped up the global version yet *)
                        let patchVersion = (Versiondata.getCurrentVersion vd) - 1 in

                        let addFVersionList_exp = Lval (var addFVersionList_svar) in

                        let addrOfIndirectionVar_exp = Cil.mkAddrOf (var indirectVar) in
                        let functionName_exp = Const(CStr functionName) in
                        let functionVersion_exp = Cil.integer functionVersion in
                        let patchVersion_exp = Cil.integer patchVersion in

                        (*        DSU_addFVersionList(indirectVar, functionName, functionVersionNumber, patchVersionNumber)  *)
                        let callToAddFVersion = Call( None, 
                                                      addFVersionList_exp, 
                                                      [ addrOfIndirectionVar_exp; functionName_exp; functionVersion_exp; patchVersion_exp ], 
                                                      Cil.locUnknown) in
                        let new_acc_il = (callToAddFVersion :: acc_il) in

                        (* Did the signature of the function change ? 
                           If YES, add it to the 'forbidden_functions'
                           If NO, do nothing
                        *)   
                        let new_acc_forbidden_fun = 
                           let oldFtype = Versiondata.getFunctionType vd fd.svar.vname
                           and newFtype = fd.svar.vtype in

                           (* isEqualFuntypeQuick shouldn't distinguish between wrapper names and non-wrapper names *)   
                           let wrapper_oblivious_name_comparator = fun x y t1 t2 -> 
                              (x = y) ||
                              (let cn t = Commoncilutil.cTN t
                               and wn n = Updateconfig.makeWrapperName n in
                               let we n1 t2 = 
                                  let wn2 = wn (cn t1) in 
                                  (wn2 = n1) 
                               in
                                  match t1, t2 with
                                       TComp (ci1,_),  TComp (ci2,_) -> (we ci1.cname t2 || we ci2.cname t1)
                                     | TComp _,  TNamed (ti2, _)     -> we ti2.tname t1
                                     | TNamed (ti1,_), TComp (ci1,_) -> we ti1.tname t2
                                     | _,_ -> false
                              )
                           in
                           let isEqualFuntype = 
                              let res = Commoncilutil.isEqualFuntypeQuick ~comparator:wrapper_oblivious_name_comparator oldFtype newFtype in
                              (*ignore(Updatedebug.log [Updatedebug.BTAlways] " isEqualFuntypeQuick(%s)\n  %s\n  %s\n  = %b \n" 
                                        fd.svar.vname 
                                        (Updatedebug.ppType oldFtype)
                                        (Updatedebug.ppType newFtype)
                                        res);*)
                              res
                           in
                           if isEqualFuntype
                           then acc_forbidden_fun
                           else (fd.svar.vname :: acc_forbidden_fun)

                        in
                        (new_acc_il, new_acc_forbidden_fun)
                              
                  | _ -> bang "Indirection variable is not a GVarDecl ??"
               end
         end
      else
         (acc_il, acc_forbidden_fun)
   in

(* TODOafter:enable this for gvar types   (\* given a GVar whose static initializer has changed since last version, *)
(*       generate code that will alpha rename the var so it doesn't collide w/the old declaration. *)
(*       dlopen() prefers the old one *)
(*     *\)  *)
(*    let genUpdateGlobalVar_fold name g (acc:Cil.instr list) : Cil.instr list = *)
(*       let vi = match g with *)
(*          GVar(v, _, _) -> v *)
(*        | _ ->  Updatedebug.abort (Updatedebug.bug "genUpdateGlobalVar_fold: expected global var definition, but got %a\n" d_global g) *)
(*       in *)
(*       let bang s = Updatedebug.abort (Updatedebug.bug "genUpdateGlobalVar_fold(%s) %s\n" vi.vname s) in *)
(*       (\* What we have to do is alpha rename this var to avoid a clash     *\) *)

(*       let newName = Updateconfig.makeVersionedVarName vi.vname ((Versiondata.getCurrentVersion vd) + 1) in *)
(*       let () = *)
(*          vi.vname <- newName      (\* alpha rename the new var to avoid the clash *\) *)
(*       in *)
(*          acc *)

(*    in *)

 (* given a GVar whose static initializer has changed since last version,
    * generate code that will set the old var to the new initializer
   *) 
(* FIX:
   let genUpdateGlobalVar_fold name g (acc:Cil.instr list) : Cil.instr list = 
      let vi = match g with
         GVar(v, _, _) -> v
       | _ ->  Updatedebug.abort (Updatedebug.bug "genUpdateGlobalVar_fold: expected global var definition, but got %a\n" d_global g)
      in
      let bang s = Updatedebug.abort (Updatedebug.bug "genUpdateGlobalVar_fold(%s) %s\n" vi.vname s) in
      (* What we have to do is locate the old var, alpha rename this var to avoid a clash 
       * and emit the set instr: 
       *      new_var = old_var
       *)
      let old_vi = begin try 
         match piFind PESExtern PETData (Updateconfig.makeOldExternVarName vi.vname) pi with
              GVarDecl(v, _) -> v
            | _ ->  Updatedebug.abort (Updatedebug.bug "genUpdateGlobalVar_fold: expected extern var decl, but got %a\n" d_global g)
         with 
              Not_found ->  bang 
		  (Printf.sprintf "getExistingVar : Can't find old var (%s) for this GVar"
		     (Updateconfig.makeOldExternVarName vi.vname))
      end in
      let newName = Updateconfig.makeVersionedVarName vi.vname ((Versiondata.getCurrentVersion vd) + 1) in
      let () = 
         old_vi.vname <- vi.vname; (* get rid of the '_old_extern' suffix *)
         vi.vname <- newName;      (* alpha rename the new var to avoid the clash *)
      in
         [Set(var old_vi, Lval(var vi), Cil.locUnknown)] @ acc
   in
*)
   let genUpdateGlobalVar_fold name g acc : string =
     let vi = match g with
       GVar(v, _, _) -> v
     | _ ->  Updatedebug.abort (Updatedebug.bug "genUpdateGlobalVar_fold: expected global var definition, but got %a\n" d_global g)
     in
     acc ^ "  " ^ vi.vname ^ "\n" in

   let setLatestTypeVersions_il, changed_types = H.fold genSetLatestTypeVersion_fold pi.pi_replaced.pg_type_hash  ([],[]) in
   let updateFunIndirections_il1, forbidden_functions1 =  H.fold genUpdateFunIndirection_fold pi.pi_replaced.pg_fun_hash ([],[]) in
   let updateFunIndirections_il2, forbidden_functions2 =  H.fold genUpdateFunIndirection_fold pi.pi_new.pg_fun_hash ([],[]) in
  (* FIX:
   let updateGlobalVars_il =  H.fold genUpdateGlobalVar_fold pi.pi_replaced.pg_data_hash [] in
*)
   let changedVars = H.fold genUpdateGlobalVar_fold pi.pi_replaced.pg_data_hash "Global variables that changed:\n" in
   ignore(Updatedebug.log [Updatedebug.BTAlways] "%s" changedVars);

   (setLatestTypeVersions_il @ updateFunIndirections_il1 @ updateFunIndirections_il2 (* PUT ME BACK ASAP @ updateGlobalVars_il *), 
    forbidden_functions1 @ forbidden_functions2 @ changed_types)



(* generate the patch install function and add it to the globals *)
let makePatchInstallFunction (vd : Versiondata.version_data) (pi : patch_info) (file : Cil.file) : unit =
   let il, forbidden_functions = generatePatchInstallCode vd pi in 
   let il = Cil.foldGlobals file
       (fun il g -> 
	 match g with
	   GFun (fd,loc) -> 
	     if fd.svar.vname = Updateconfig.state_xform_function_name then
	       if (isFunctionType fd.svar.vtype) then
		 (match (Updatecilutil.getTypesInFunDec fd) with
		   (TVoid _,[]) -> 
              fd.svar.vstorage <- Static;
              fd.svar.vinline <- true;
		     il @ [Call (None,Lval (var fd.svar),[],loc)]
		 | _ -> 
		     Updatedebug.abort (Updatedebug.error "makePatchInstallFunction: user state transformer function %s must have type void -> void\n" Updateconfig.state_xform_function_name))
	       else Updatedebug.abort (Updatedebug.error "makePatchInstallFunction: user state transformer function %s must have type void -> void\n" Updateconfig.state_xform_function_name)
	     else il
	 | _ -> il) il in

   let patchInstallFunction = 
      let install_fun = Cil.emptyFunction Updateconfig.patch_install_function_name in 
      install_fun.sbody <- {bstmts = [Commoncilutil.mkStmt (Instr il)]; battrs = []};
      GFun(install_fun, Cil.locUnknown)

      (*
           char * __DSU_metadata = "f1 f2 f3 .... fn";
           
      *)
   and metadataVar = 
      let metadata_vi = Cil.makeGlobalVar Updateconfig.patch_metadata_var_name Updateconfig.patch_metadata_var_type in 

      (* make a big string w/all the function names*)   
      let function_names_sausage = List.fold_left (fun x acc -> x ^ " " ^ acc) "" forbidden_functions in 

      (* set metadata = string *)   
      GVar(metadata_vi, { init = Some (SingleInit (Cil.mkString function_names_sausage))}, Cil.locUnknown)

   in
   file.globals <- file.globals @ [patchInstallFunction; metadataVar];;




(****************************** Merge patch info with version data *********************)


(* merge the given patch info, pi, with given version data, vd.  The version data is
   modified in-place so that it represents the version data of a a program conforming to
   the original vd that had a patch conforming to pi applied to it *)
let mergePIintoVD (pi : patch_info) (vd : Versiondata.version_data) (file : Cil.file) (funTypeHash : Analysis.funTypeHash_t) : unit =
  
  (* merge a new global function into the version data *)
  let mergeNewFunctions (funName : string) (g : Cil.global) : unit =
     let () = ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "mergeNewFunctions : %s\n" funName) in
     let fd = 
        match g with 
          GFun(fd',loc) -> fd'
        | _ -> Updatedebug.abort (Updatedebug.bug "mergeNewFunctions : tried to merge non-function global %a as function''%s'!\n" d_global g funName)
     in
     if Versiondata.isExistingFunction vd funName then 

        (* it really did exist before, so bump up the version number as we're replacing it *)
        let () = ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "mergeReplacedFunctions : signature changed : %s\n" funName) in
        Versiondata.incFuncVersionNumber vd funName

     else
        (* Really new to VD, so add it, make a fresh indirection var *)
        (* we're in patch mode and Indirection has already run at this point, 
           hence we might know the indirection var from Indirection *)
        begin
           let indirection_var = 
              try
                 Indirection.FunctionInfo.getIndirectionVar funName 
              with Not_found -> 
                 Updatedebug.abort (Updatedebug.bug "mergeNewFunctions(%s) : can't find indirection var for this function\n" funName) 
           in
           Versiondata.addFunction vd fd indirection_var.vname funTypeHash;
           Versiondata.addVar vd indirection_var;   
        end
  in 

  (* merge a replaced global function into the version data *)
  let mergeReplacedFunctions (funName : string) (g : Cil.global) : unit =
     let () = ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "mergeReplacedFunctions : %s\n" funName) in
     let fd = 
        match g with 
          GFun(fd',loc) -> fd'
        | _ -> Updatedebug.abort (Updatedebug.bug "mergeReplacedFunctions : tried to merge non-function global %a as function''%s'!\n" d_global g funName)
     in
     if Versiondata.isExistingFunction vd funName then
        (* it really did exist before, so bump up the version number as we're replacing it *)
        let () = ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "mergeReplacedFunctions : %s\n" funName) in
        Versiondata.incFuncVersionNumber vd funName
     else
        (* We're in an inconsistent state as we have a function in the patch to replace a non-existant function *)
        Updatedebug.abort (Updatedebug.bug "mergeReplacedFunctions : '%s' allegedly existing, but not found in version data ?\n" funName)
  in

  (* merge a new type into the version data *)
  let mergeNewTypes (typeName : string) (g : Cil.global) : unit =
    let () = ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "mergeNewTypes : %s\n" typeName) in
    let t = 
      match g with 
          GType(ti, _) -> TNamed(ti, Commoncilutil.noAttr) 
        | GCompTag(ci, _) -> TComp(ci, Commoncilutil.noAttr)
        | _ -> Updatedebug.abort (Updatedebug.bug "mergeNewTypes : tried to merge non-type global %a as type '%s'!\n" d_global g typeName)
    in
      if Versiondata.isExistingType vd typeName then 
        (* It existed before, but it is being added as a new type: bad *)
        Updatedebug.abort (Updatedebug.bug "mergeNewTypes : '%s' allegedly new, but it exists in version data ?\n" typeName)
      else
        (* Really new to VD, so add it *)
        Versiondata.addType vd (Namedtypes.isUpdatable t) t
  in 


  (* merge a replaced global type into the version data *)
  let mergeReplacedTypes (typeName : string) (g : Cil.global) : unit =
     let () = ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "mergeReplacedTypes : %s\n" typeName) in
     let t = 
        match g with 
          GType(ti, _) -> TNamed(ti, Commoncilutil.noAttr) 
        | GCompTag(ci, _) -> TComp(ci, Commoncilutil.noAttr)
        | _ -> Updatedebug.abort (Updatedebug.bug "mergeReplacedTypes : tried to merge non-type global %a as type '%s'!\n" d_global g typeName)

     in
     if Versiondata.isExistingType vd typeName then
        (* it really did exist before, so bump up the version number as we're replacing it *)
        Versiondata.incTypeVersionNumber vd typeName
     else
        (* We're in an inconsistent state as we have a type in the patch to replace a non-existent type *)
        Updatedebug.abort (Updatedebug.bug "mergeReplacedTypes : '%s' allegedly existing, but not found in version data ?\n" typeName)

  in

  (* merge a new global variable into the version data *)
  let mergeNewVars (varName: string) (g : Cil.global) : unit = 
     let () = ignore(Updatedebug.log [Updatedebug.BTPatchcomp] "mergeNewVars : %s\n" varName) in
     let vi = 
        match g with 
          GVar(vi',_,loc) | GVarDecl(vi',loc) -> vi'
        | _ -> Updatedebug.abort (Updatedebug.bug "mergeNewVars : tried to merge non-variable global %a as variable''%s'!\n" d_global g varName)
     in
     if Versiondata.isExistingVar vd vi.vname then 
       ignore(Updatedebug.log 
		[Updatedebug.BTPatchcomp] "mergeNewVars : global var exists : %s\n" varName)
     else
       begin
         Versiondata.addVar vd vi;
       end
  in

     genPatchInfo file pi;

     (* the order these are called in doesn't matter *)
     H.iter mergeNewFunctions pi.pi_new.pg_fun_hash; 
     H.iter mergeNewTypes pi.pi_new.pg_type_hash; 
     H.iter mergeNewVars pi.pi_new.pg_data_hash; 
     H.iter mergeReplacedFunctions pi.pi_replaced.pg_fun_hash;
     H.iter mergeReplacedTypes pi.pi_replaced.pg_type_hash; 


     (* finally bump up the global version *)
     Versiondata.incrementCurrentVersion vd 


(* Make sure old and new versions make consistent asumptions
   about the types (e.g. updatability) *)
let ensureOldAndNewTypeConsistency (vd : Versiondata.version_data) : unit =
   let vd_ensure_consist_iter typeName typeData =
      begin
         if (not typeData.Versiondata.vd_abs_updatable) 
         then
            begin
               if Forcedpragma.isForcedUpdatable typeName 
               then
                  (* Error. We got a force updatable request for a non-updatable type *)
                  Updatedebug.abort (Updatedebug.error "ensureOldAndNewTypeConsistency : Trying to force updatable type '%s', but it exists in VD as non-updatable !\n" typeName)
               else (* was non-updatable in last version, force non-updatable here as well *)
                  ignore(Namedtypes.setAbsTypeNameNonUpdatable typeName Cil.locUnknown "was non-updatable in previous version") 
            end
         else (* was updatable *)
            begin
               if Forcedpragma.isForcedNonUpdatable typeName 
               then
                  (* Error. We got a force non-updatable request for an updatable type *)
                  Updatedebug.abort (Updatedebug.error "ensureOldAndNewTypeConsistency : Trying to force non-updatable type '%s', but it exists in VD as updatable !\n" typeName)
            end
      end
   in
   H.iter vd_ensure_consist_iter vd.Versiondata.vd_abs_type_hash
                                                                               
                                                                               

