(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*   We need to split the patch into separate file for each function, 
     in order to be able to unload them individually (function garbage collection)

     If the patch has say, 3 functions, before the spitting we have:

     .patch.c
     --------
     tt_typedef_T() {}
     foo() {}          
     bar() {}          

     and after the splitting it's gonna look like this:  


     .patch.c                       foo.c               bar.c
     ----------------               ------              -----
     tt_typedef_T() {}              foo() {}            bar() {}

*)

open Cil

module H = Hashtbl


let getSplinterFileOutputChannel fileName =
  try 
    open_out fileName 
  with (Sys_error msg) ->
    Updatedebug.abort (Updatedebug.error "Can't open '%s' for writing the splinter file: %s\n" fileName msg)


(* Given a list of globals that form a splinter, dump them into a file *)
let dumpGlobalsToSplinterFile globalsList splinterFileName isRoot =
  let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "dumpGlobalsToSplinterFile(%s)\n" splinterFileName) in
  let splinterFile = { globals = globalsList;
                       fileName = splinterFileName;
                       globinit = None;
                       globinitcalled = false} in
  let outChan = getSplinterFileOutputChannel splinterFileName in 
  output_string outChan ("/* Splinter file, generated by CIL in patch compilation mode\n");

  (* throw away all unneeded externs *)   
(*   Rmtmps.removeUnusedTemps ~isRoot:isRoot splinterFile;   *)

  (* and finally write it to disk *)
  Cil.dumpFile Cil.defaultCilPrinter outChan splinterFile


(* given a patch file and a function name, turn all definitions but that function
   into external declarations. Dump the file to disk too. 
*)
let makeSplinterFile (patchfile : Cil.file) 
                     (funName : string) 
                     (isNew : bool) 
                     (isIndirectionVarForThisFunction : string -> string -> bool)
                     (isSplinterFunction : string -> bool) 
                     (getNonVersionedName : string -> string) 
                     (outputDir : string) : unit = 

   (* we're gonna set these *once* in the fold, so I don't bother carrying them around *)
   let funRootName = ref ""
   and indirectVarRootName = ref "" in

   (* Remove temps uses isRoot to ask whether this is a root in the transitive closure or not 
      Only the function and its indirection var definition are roots; all the rest can be thrown away
      if not needed or left external if needed.
   *)   
   let isRoot = fun g -> match g with
        GFun(fd,loc) when fd.svar.vname = !funRootName -> true
      | GVar(vi,ii,loc) when vi.vname = !indirectVarRootName -> true
      | _ -> false
   in

   let splinterGlobals = 
      let turnDefsIntoDecls_fold (acc : Cil.global list) (g : Cil.global) : Cil.global list = 
         match g with

              GFun(fd,loc) when isSplinterFunction fd.svar.vname && funName = getNonVersionedName fd.svar.vname ->
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "splinter turnDefsIntoDecls_fold (%s) : keep def\n" funName) in
                 funRootName := fd.svar.vname;   
                 g :: acc

            | GFun(fd,loc) when Updateconfig.isWrapperFunctionForMe funName fd.svar.vname ->
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "splinter turnDefsIntoDecls_fold (%s) : keep def\n" funName) in
                 g :: acc

            | GFun(fd,loc) ->
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "splinter turnDefsIntoDecls_fold (%s) : def -> decl\n" fd.svar.vname) in
                 let declVi = Cil.makeGlobalVar fd.svar.vname fd.svar.vtype in
                 declVi.vstorage <- Extern;   
                 GVarDecl(declVi, loc) :: acc

              (* indirection var definitions for new functions must stay in the patch, not in splinter 
                 (otherwise we get an undefined symbol in __DSU_install_patch()  *)      
            | GVar(vi,ii,loc) when isNew && not (Cil.startsWith "__DSU_indirect_tt" vi.vname) ->

                 let b = isIndirectionVarForThisFunction funName vi.vname in
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "********* isIndirectionVarForThisFunction fun %s var %s : %b\n" funName vi.vname b) in
                 indirectVarRootName := vi.vname;   
                 let declVi = Cil.makeGlobalVar vi.vname vi.vtype in
                 declVi.vstorage <- Extern;   
                 GVarDecl(declVi, loc) :: acc

   
            | GVar(vi,ii,loc) ->
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "splinter turnDefsIntoDecls_fold (%s) : def -> decl\n" vi.vname) in
                 let declVi = Cil.makeGlobalVar vi.vname vi.vtype in
                 declVi.vstorage <- Extern;   
                 GVarDecl(declVi, loc) :: acc

            | _ -> 
                 g :: acc              
      in
      List.rev (List.fold_left turnDefsIntoDecls_fold [] patchfile.globals)
   in
   dumpGlobalsToSplinterFile splinterGlobals (outputDir ^"/" ^ funName ^ ".patch.cil.c") isRoot   

(* Given a file and PI info, spin off all functions into separate files 
   Turn their definitions into extern declaration in the original patch 
*)
let doPatchSplitting (patchfile : Cil.file) (pi : Patchcomp.patch_info) (versionedFunHash : (string, string) H.t) : unit = 

   let outputDir = 
      match Updateargs.args.Updateargs.arg_patchDirectoryOut with
            "" -> 
               Updatedebug.abort (Updatedebug.bug "doPatchSplitting : no output directory was specified, so where should the patches be stored ? \nPlease specify an output dir with --patchdirectory-out=<dir>\n")
         | _ as res -> res
   in
   let isSplinterFunction = (fun name -> H.mem versionedFunHash name) in
   let getNonVersionedName = (fun name -> H.find versionedFunHash name) in 

   let isTTthing name = 
      if (Ttpragma.isTT name) || (Ttwrapinfo.isTTWrapper name)  || (Cil.startsWith Updateconfig.state_xform_function_name name)
      then true
      else
         if isSplinterFunction name 
         then 
            let nvName = getNonVersionedName name in
               (Ttpragma.isTT nvName) || (Ttwrapinfo.isTTWrapper nvName)  
         else
            false
   in


   let addedFunsIndirectionVars = 
      let res = H.create 127 in
      let putAddedFunIntoHash_iter funName g = 
         if Indirection.shouldConsider funName then
            begin
               try
                  let indirectionVarName = (Indirection.FunctionInfo.getIndirectionVar funName).vname in
                     H.add res indirectionVarName funName
               with _ ->
                  Updatedebug.abort (Updatedebug.bug "doPatchSplitting : function '%s' doesn't have an indirection var ??\n" funName)
            end
         else
            ()
      in
      H.iter putAddedFunIntoHash_iter pi.Patchcomp.pi_new.Patchcomp.pg_fun_hash;
      res
   in
   let isIndirectionVarForThisFunction = (fun funName varName -> 
                                             try
                                                funName = (H.find addedFunsIndirectionVars varName)
                                             with Not_found -> 
                                                false
                                         )
   in

   let isIndirectionVarForNewFunction = (fun varName -> H.mem addedFunsIndirectionVars varName) in

   (* create splinter files first *)
   let lReplaced = H.fold (fun funName g acc -> (funName, false) :: acc) pi.Patchcomp.pi_replaced.Patchcomp.pg_fun_hash [] in
   let lNew =   H.fold (fun funName g acc -> (funName, true) :: acc) pi.Patchcomp.pi_new.Patchcomp.pg_fun_hash      [] in


   let splinterFunctionNames = List.filter (fun (x,y) -> Indirection.shouldConsider x && not (isTTthing x)) (lReplaced @ lNew) in   
   let () = List.iter (fun (funName,isNew) -> makeSplinterFile patchfile funName isNew isIndirectionVarForThisFunction isSplinterFunction getNonVersionedName outputDir) splinterFunctionNames in


   (* then take care of the patch itself *)
   let patchGlobals = 
      let turnDefsIntoDecls_fold (acc : Cil.global list) (g : Cil.global) : Cil.global list = 
         match g with

              GFun(fd,loc) when isSplinterFunction fd.svar.vname && not (isTTthing fd.svar.vname) ->
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "patch turnDefsIntoDecls_fold (%s) : def -> decl\n" fd.svar.vname) in
                 fd.svar.vstorage <- Extern;   
                 GVarDecl(fd.svar, loc) :: acc

            | GFun(fd,loc)  ->
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "patch turnDefsIntoDecls_fold (%s) : keep def\n" fd.svar.vname) in
                 g :: acc

              (* definitions for new indirection variables must stay in the patch, and not in the splinter files *)      
            | GVar(vi,ii,loc) when isIndirectionVarForNewFunction vi.vname && not (Cil.startsWith "__DSU_indirect_tt" vi.vname)->
                 let () = ignore(Updatedebug.log [Updatedebug.BTPatchsplit] "patch turnDefsIntoDecls_fold (%s) : def -> decl\n" vi.vname) in
                 let newIi = 
                    begin
                       match ii with
                             {init = Some (CompoundInit(indirectionVarTyp, [itemBaseFunName; itemFirst; itemActiveFunction; itemOldVersionTotalCount; itemLength]))} ->
                                let zeroAddr = Cil.makeZeroInit (snd Updateconfig.fv_itemActiveFunction) in
                                   { init = Some (CompoundInit(indirectionVarTyp, [itemBaseFunName; itemFirst; (fst itemActiveFunction, zeroAddr); itemOldVersionTotalCount; itemLength])) }
                          | _ -> Updatedebug.abort (Updatedebug.bug "doPatchSplitting : function w/indirect var '%s' has non-conform indirection var ??\n" vi.vname)

                    end
                 in GVar(vi,newIi,loc) :: acc

            | _ -> 
                 g :: acc              
      in
      List.rev (List.fold_left turnDefsIntoDecls_fold [] patchfile.globals)
   in
   patchfile.globals  <- patchGlobals;;
