(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(****************************** Determine TT argument types *************************************)

open Cil

module H = Hashtbl

(*
   When generating patches the user specifies transformations between types.  The types that the 
   type transformer converts between should be left alone by con/abs insertion, wrapping and compilation 
   in general.  We identify type transformers and their associated types by a #pragma directive:

     #pragma DSU_TT("type_transformer", "T", "t_old", "t_new")

   In this phase we gather these directives and store them for later use.

   FIXME: at present T has to be struct_T/enum_T/union_T/typedef_T so that it coincides with 
   getAbstractTypeByName.  We have to keep the struct/enum/union/typedef because they are 
   separate name spaces, but we could munge in the underscore.
*)



(* ttHash:  map from typetransformer names to pairs representing the names of the old and new types
            it converts between *)
type ttHashData = {tt_abs_type_name : string; tt_old : string; tt_new : string}
type ttHash_t = (string,ttHashData) H.t
let ttHash : ttHash_t = H.create 4099 

(* ttArgsHash: a map from the abstract type names of the argument to the abstract type name to replace *)
type ttArgsHash_t = (string,string) H.t
let ttArgsHash : ttArgsHash_t = H.create 4099 

(* map from type names to type transformer names *)
let ttNameHash : (string, string) H.t = H.create 4099


let isTTArgType (t:Cil.typ) : bool =
   H.mem ttArgsHash (Updatecilutil.getAbstractTypeName t)

let getAbsTypeFromTTArgType (t_arg_name:string) : string = 
   H.find ttArgsHash t_arg_name

let isTT (name:string) : bool = 
   H.mem ttHash name

let getTTName (tname:string) = 
   H.find ttNameHash tname

let getTTData (tname:string) : ttHashData = 
   H.find ttHash tname 

(* extract all of the info in the TT hash pragma directives *)
let collectTTPragmas (vd : Versiondata.version_data) (file:Cil.file) : unit = 
   
   let addArgTypeNameToHash (t_arg_name:string) (t_abs_name:string) : unit = 
      if H.mem ttArgsHash t_arg_name then
         Updatedebug.abort (Updatedebug.bug "addTTNameToHash: adding arg name %s to hash for the second time!\n" t_arg_name)
      else
         H.add ttArgsHash t_arg_name t_abs_name;
   in

   (* Associate a TT name with the names of the types it transformes between *)
   let addTTNameToHash (ttname:string) (ttdata:ttHashData) :unit = 
      if H.mem ttHash ttname then 
         Updatedebug.abort (Updatedebug.bug "addTTNameToHash: adding TT name %s to hash for the second time!\n" ttname)
      else 
         H.add ttHash ttname ttdata
   in
   
   let addTTToTTNameHash (tname:string) (ttname:string) : unit = 
      if H.mem ttNameHash tname then 
         Updatedebug.abort (Updatedebug.bug "addTTToTTNameHash: adding type name %s to hash for second time!\n" tname)
      else
         H.add ttNameHash tname ttname
   in
   
   let processPragma (g : Cil.global) : unit =
      match g with 
           GPragma(Attr (name,[AStr ttname; AStr tname; AStr t_old; AStr t_new]),loc) when name = Updateconfig.ttPragmaTypeName ->
              begin 
                 try
                    if Versiondata.isUpdatableType vd tname then
                       begin
                          addArgTypeNameToHash t_old tname;
                          addArgTypeNameToHash t_new tname;
                          addTTNameToHash ttname {tt_abs_type_name = tname; tt_old=t_old; tt_new=t_new};
                          addTTToTTNameHash tname ttname;

                          (* Set the argument types non-updatable *)
                          try
                             Namedtypes.setAbsTypeNameNonUpdatable t_old loc "user #pragma"; 
                             Namedtypes.setAbsTypeNameNonUpdatable t_new loc "user #pragma" 

                          with Not_found -> 
                             Updatedebug.abort (Updatedebug.error "processPragma(%s) Namedtypes doesn't know about this type ?!\n" tname)
                       end
                    else
                       (* Warning: We got a type transformer for a non-updatable type; ignore *)
                       Updatedebug.abort (Updatedebug.error "Found type transformer for non-updatable type '%s' (auto-generated?)\n" tname)
                 with Not_found -> 
                    Updatedebug.abort (Updatedebug.error "processPragma(%s) Versiondata doesn't know about this type ?!\n" tname)
              end
         | GPragma(Attr (name,[AStr ttname; AStr tname; AStr t_old; AStr t_new]),loc) when name = Updateconfig.ttPragmaGVarName ->
              begin
                 ignore (Updatedebug.log [Updatedebug.BTAlways] "processPragma(%s) : ttname '%s' tname '%s' \n" name ttname tname);
                 
                  addTTNameToHash ttname {tt_abs_type_name = tname; tt_old=t_old; tt_new=t_new};
                  addTTToTTNameHash tname ttname

              end  
            
         | _ -> ()
   in
   Cil.iterGlobals file processPragma


