(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil

module H = Hashtbl

(*
   The Type transformers that the user supplies are of the form:
   
      void tt(Told* to, Tnew* tn, T* new)
      
   But the runtime expects transformations to be done in-place by a TT of the form:
   
      void __DSU_tt(__DSU_wrapper_T *w)
      
   So we create this function that wraps the users TT:
   
      void __DSU_tt(__DSU_wrapper_T *w) {
         Told to;
         Tnew *tn;

         if (w->version < __DSU_latest_type_version_T) {
            to = w->udata.data;     /* make shallow structural copy of old data */
            tn = &(w->udata.data);  /* point new data to old storage space */
            tt(&to,tn,w);             /* call users TT */
         }

         return;
      }
*)

(* Create a TT that does the coercion in place given the user's TT *)
let makeInplaceTT (vd : Versiondata.version_data) (userTT : Cil.fundec) (memcpy_vi : Cil.varinfo) : Cil.fundec * (Cil.global option) = 
   (* make wrapper function type *)
   let newFunction  = Cil.emptyFunction (Updateconfig.makeInplaceTTName userTT.svar.vname) in 
   let absTypeName = (Ttpragma.getTTData userTT.svar.vname).Ttpragma.tt_abs_type_name in
   let absType = 
      try
         Namedtypes.getAbsTypeFromName absTypeName 
      with
         Not_found -> Updatedebug.abort (Updatedebug.bug "makeInplaceTT : Namedtypes.getAbsTypeFromName failed for type '%s' !\n" absTypeName)
   in 
   let wrapperType = Wrappers.getWrapperType vd absType in 
   let newFunctionType = TFun(TVoid [], Some [("w", TPtr(wrapperType,[]), [])], false, []) in
   let newFormal = Cil.makeFormalVar newFunction ~where:"^" "w" (TPtr(wrapperType,[])) in
   let () = Cil.setFunctionType newFunction newFunctionType in

   (* decl for last version of userTT *)
   let lastF = 
     let version = Versiondata.getFunctionVersion vd userTT.svar.vname in
     if version != 0 then
       let vi = makeGlobalVar 
	   (Updateconfig.makeVersionedFunctionName 
	      (Updateconfig.makeInplaceTTName userTT.svar.vname) (version-1))
	   newFunctionType in
       vi.vstorage <- Extern;
       Some (GVarDecl (vi,locUnknown))
     else None in

   (* add the formals of userTT to the local vars of the function *)
   let t_old_vi = List.nth userTT.sformals 0 in 
   let t_new_vi = List.nth userTT.sformals 1 in
   let to_local = 
      match t_old_vi.vtype with 
         TPtr(t,_) -> Cil.makeLocalVar newFunction t_old_vi.vname t 
       | _ -> Updatedebug.abort (Updatedebug.bug "makeInplaceTT: argument to type transformer is not a pointer")
   in
   let tn_local = Cil.makeLocalVar newFunction t_new_vi.vname t_new_vi.vtype in 
   let version_fi = Commoncilutil.getFieldinfoByName wrapperType Updateconfig.wVersionFieldName None in
   let udata_fi = Commoncilutil.getFieldinfoByName wrapperType Updateconfig.wDataUnionName None in 
   let data_fi = Commoncilutil.getFieldinfoByName udata_fi.ftype Updateconfig.wDataFieldName None in 
   let latest_type_version_vi = Typeindirection.getVersionVar absTypeName in
   let cast_to_T_old_via_ptr lval = Lval( Mem (CastE(t_old_vi.vtype, Cil.mkAddrOf lval)), NoOffset ) in
   let cast_to_T_new_ptr e = CastE(t_new_vi.vtype, e) in

   let make_old_data_copy_instr = 
      let dest = (var to_local) in
      let src = cast_to_T_old_via_ptr ((Mem (Lval( var newFormal ) ) ), Field(udata_fi, Field( data_fi, NoOffset )) ) in
      if (Cil.isArrayType to_local.vtype)
      then
         (* memcpy (to, w->udata.data, sizeof(to)) *)
         Call(None, Lval(var memcpy_vi), [ Lval dest; src; SizeOfE(Lval dest)], Cil.locUnknown)
      else
         (* to = w->udata.data *)
         Set(dest, src, Cil.locUnknown)
   in
   let point_new_data_to_old_storage_instr = Set( var tn_local, 
                                                  cast_to_T_new_ptr ( Cil.mkAddrOf ( Mem (Lval( var newFormal ) ),  Field(udata_fi, Field( data_fi, NoOffset ))) ), 
                                                  Cil.locUnknown )
   in
   let call_tt_instr = Call(None, Lval (var userTT.svar), [Cil.mkAddrOf (var to_local); Lval (var tn_local); (Lval( var newFormal ))], Cil.locUnknown) in

   let then_block = 
     let rest = 
       Commoncilutil.mkStmt 
	 (Instr 
	    [ make_old_data_copy_instr;
	      point_new_data_to_old_storage_instr;
	      call_tt_instr]) in
     (match lastF with
       Some (GVarDecl (vi,_)) ->
	 let call_lasttt_instr =
	   Call(None, Lval (var vi), [Lval( var newFormal )], Cil.locUnknown) in
	 let if_stmt =
	   Commoncilutil.mkStmt
	     (If (BinOp(Ne, Lval( (Mem( Lval( var newFormal ) ) ), Field(version_fi, NoOffset)), 
			BinOp(MinusA, Lval(var latest_type_version_vi), 
			      (Cil.integer 1), Cil.intType), Cil.intType ),
		  Cil.mkBlock [Commoncilutil.mkStmt (Instr [call_lasttt_instr])],
		  Cil.mkBlock [], Cil.locUnknown)) in
	 Cil.mkBlock [if_stmt; rest]
     | None -> 
	 Cil.mkBlock [rest]
     | _ -> failwith "Impossible!")
   in
   let body : Cil.block = 
      Cil.mkBlock
      [ Commoncilutil.mkStmt (
           If( BinOp(Lt, Lval( (Mem( Lval( var newFormal ) ) ), Field(version_fi, NoOffset)), Lval(var latest_type_version_vi), Cil.intType ), 
               then_block,
               Cil.mkBlock [], Cil.locUnknown) 
        ) ] 
   in 
   newFunction.sbody <- body;
   (newFunction,lastF)
      

(* Given a file, it identifies all type transformers and adds wrappers for them;
   also, it inserts them wrappers in ttToTTWrapperHash *)
let genTTWrappers (file : Cil.file) (vd : Versiondata.version_data) (memcpy_vi : Cil.varinfo) : unit =
   let genTTWrapper_fold g acc = 
      match g with
         GFun(fd,loc) when Ttpragma.isTT fd.svar.vname -> 
            let (tt_wrapper,g_last_tt_wrapper) = makeInplaceTT vd fd memcpy_vi in
            let g_tt_wrapper = GFun(tt_wrapper, Cil.locUnknown) in
            let () = Indirection.FunctionInfo.addStaticFunction (Some vd) tt_wrapper.svar.vname  g_tt_wrapper tt_wrapper.svar in
            let indirectionVar = Indirection.FunctionInfo.getIndirectionVar tt_wrapper.svar.vname in
            let bogusFuntypeHash = H.create 1 in
            let () = H.add bogusFuntypeHash tt_wrapper.svar.vname tt_wrapper.svar.vtype in
            H.add Ttwrapinfo.ttToTTWrapperHash fd.svar.vname tt_wrapper.svar;
            H.add Ttwrapinfo.ttWrapperHash tt_wrapper.svar.vname ();

            (if not (Versiondata.isExistingVar vd indirectionVar.vname)
            then Versiondata.addVar vd indirectionVar);

            (if not (Versiondata.isExistingFunction vd tt_wrapper.svar.vname)
             then Versiondata.addFunction vd tt_wrapper indirectionVar.vname bogusFuntypeHash
             else Versiondata.incFuncVersionNumber vd tt_wrapper.svar.vname);

	    (match g_last_tt_wrapper with
	      Some g' ->
		g::g'::g_tt_wrapper::acc
	    | None -> 
		g::g_tt_wrapper::acc)
         
        | _ -> g::acc
   in      
   file.globals <- List.fold_right genTTWrapper_fold file.globals []




