(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)




open Cil


(* Type structures in CIL are truely recursive -- be careful when comparing them! 
   Cil uses type signatures to compare types as t=t' iff sig(t)=sig(t').
   However, CIL's type signatures don't have named types -- they unroll them when 
   converting to typesigs!  This is not good for us, so we have our own type signatures. *)
type utypsig = 
    UTSArray of utypsig * exp option * attribute list
  | UTSNamed of string * utypsig * attribute list
  | UTSPtr of utypsig * attribute list
  | UTSComp of bool * string * attribute list
  | UTSFun of utypsig * utypsig list * bool * attribute list
  | UTSEnum of string * attribute list
  | UTSBase of typ

let rec ppUTypeSig (usig:utypsig) : string = 
   match usig with 
      UTSArray(ut,eo,al)  ->  "[" ^ (ppUTypeSig ut) ^ "]"
    | UTSNamed(s,ut,al)  -> s 
    | UTSPtr(ut,al)   -> "(" ^ (ppUTypeSig ut) ^ ")*"
    | UTSComp(b,s,al)  -> s
    | UTSFun(ut,utl,b,al)  -> (List.fold_left (fun s ut' -> s ^ (ppUTypeSig ut') ^ ",") "" utl) 
                              ^ " -> " ^ (ppUTypeSig ut)
    | UTSEnum(s,al)  -> s
    | UTSBase(t)  -> "BASE"


(* Convert a type to a type signature.  Copied from CIL, but don't unroll named types *)
let rec uTypeSigWithAttrs (doattr:Cil.attributes -> Cil.attributes) (t:Cil.typ) =  
  let typeSig = uTypeSigWithAttrs doattr in
  match t with 
  | TInt (ik, al) -> UTSBase (TInt (ik, doattr al))
  | TFloat (fk, al) -> UTSBase (TFloat (fk, doattr al))
  | TVoid al -> UTSBase (TVoid (doattr al))
  | TEnum (enum, a) -> UTSEnum (enum.ename, doattr a)
  | TPtr (t, a) -> UTSPtr (typeSig t, doattr a)
  | TArray (t,l,a) -> 
       (* Arrays contain expressions, whcih contain types, which can be cyclic!
          When we compare array types we ignore the length -- is this right? *)
       UTSArray(typeSig t, None, doattr a)
  | TComp (comp, a) -> 
      UTSComp (comp.cstruct, comp.cname, doattr (Cil.addAttributes comp.cattr a))
  | TFun(rt,args,isva,a) -> 
      UTSFun(typeSig rt, 
            List.map (fun (_, atype, _) -> (typeSig atype)) (argsToList args),
            isva, doattr a)
  | TNamed(ti, a) -> 
       if Namedtypes.isTransparent t then 
          typeSig ti.ttype
       else
          UTSNamed(ti.tname, uTypeSigAddAttrs (doattr a) (typeSig ti.ttype), doattr a)

  | TBuiltin_va_list al -> UTSBase (TBuiltin_va_list (doattr al))      

and uTypeSigAddAttrs a0 t = 
  if a0 == [] then t else
  match t with 
    UTSBase t -> UTSBase (Cil.typeAddAttributes a0 t)
  | UTSNamed (name,ts,a) -> UTSNamed(name,ts, Cil.addAttributes a0 a)
  | UTSPtr (ts, a) -> UTSPtr (ts, Cil.addAttributes a0 a)
  | UTSArray (ts, l, a) -> UTSArray(ts, l, Cil.addAttributes a0 a)
  | UTSComp (iss, n, a) -> UTSComp (iss, n, Cil.addAttributes a0 a)
  | UTSEnum (n, a) -> UTSEnum (n, Cil.addAttributes a0 a)
  | UTSFun(ts, tsargs, isva, a) -> UTSFun(ts, tsargs, isva, Cil.addAttributes a0 a)


(* produce a type signature removing attributes *)
let uTypeSigNoAttrs t = uTypeSigWithAttrs (fun al -> []) t

(* produce a type signature maintaining attributes *)
let uTypeSig t = uTypeSigWithAttrs (fun al -> al) t

(* determine if two types are equal modulo their attributes *)
let uAreTypesEqualNoAttrs t1 t2 = 
  let ts1 = (uTypeSigNoAttrs t1) in
  let ts2 = (uTypeSigNoAttrs t2) in 
  ts1 = ts2

(* determine if two types are equal *)
let uAreTypesEqual t1 t2 = (uTypeSig t1) = (uTypeSig t2)

