(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Manuel Oriol        <moriol@inf.ethz.ch> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* $Id: typeprocessing.ml,v 1.7 2004/11/17 22:12:59 oriol Exp $ *)
(* This feature is made to generate type transformers that transform values of an old version of the type to the new one.
By default the functionning of typetransformer.ml is as follows:

cilly --dotypetransformer --merge  test0-1.c test0-2.c 

The standard output should be filled in with the pretty printed version of the type transformers generated functions.
Note that as there are error messages in cil when redefining types the standard output should probably be redirected into a file to have a very good presentation.

In the code, the function doOneType is the most important.

 doOneType  (originatingFunction:Cil.fundec) (t1:Cil.typ) (t2:Cil.typ) (input: lval)(output: lval) : Cil.stmt list 

It is a recursive function that takes an originating function which is intended to be the type transformer function itself.
t1 is the old version of the type.
t2 is the new version of the type.
input is the input variable (having type t1).
output is the output variable (having type t2).

It appeared to be useful to have the separation from t1,t2 and input,output for the simple reason that we may have strange transtyping things that happen in C.
This function returns a stmt list that constitutes the statements of the type transformer.

TODO: semantic unification of variables (see discussion with Iulian)
TODO: verify unions

Manuel
*)


open Cil


module H = Hashtbl
 

type hashOfGlobals = (string, global) H.t
type hashOfTT = (Cil.typ,string) H.t


let uniqueId : int ref = ref 1


let oldVersionTag:string  = "old"
let newVersionTag:string  = "new"

(* code used to build the names differentiated for versions *)
let versionSeparator = "_"
let typeTransformerBaseName = "tt_"

(* a function that builds the type transformer name *)
let buildTypeTransformerName (baseTypeName: string)(oldVersionTag : string) (newVersionTag : string):string = 
   (typeTransformerBaseName ^ baseTypeName(* ^ versionSeparator ^ oldVersionTag ^ versionSeparator ^ newVersionTag*))



(* for structs in structs treatment, we make struct transforming calls 
 * Manuel's code*)
let structTT: hashOfTT = H.create 4099


(* here begin the functions that make the stmtkind *)
(* this function makes a return stmtkind *)
let mkReturnStmtkind(v:exp) : stmtkind = 
   Return(Some(v), Cil.locUnknown)

(* this function returns a call stmtkind e.g. printf *)
let mkCallStmtkindNoReturn (funName: string)(funtype: Cil.typ)(args: Cil.exp list) (*originatingFunction: Cil.location*):stmtkind =
     let i:Cil.instr = Call ( None, (Lval(((Var(makeGlobalVar funName funtype)),NoOffset))), args, Cil.locUnknown) in
     Instr([i])



(* this function returns a call stmtkind e.g. int2float *)
let mkCallStmtkind (result:Cil.lval)(funName: string)(funtype: Cil.typ)(args: Cil.exp list) (*originatingFunction: Cil.location*):stmtkind =
        let i:Cil.instr = Call (Some(result), (Lval(((Var(makeGlobalVar funName funtype)),NoOffset))), args, Cil.locUnknown) in
   Instr([i])  

(* this function returns a cast stmtkind (*e.g. (int *)x *)
let mkCastStmtkind (newTyp: Cil.typ) (valueToAssign: Cil.lval) (assignedValue: Cil.lval): stmtkind =
   let castExp:Cil.exp=(Cil.mkCast (Lval(valueToAssign)) newTyp) in
   let instCast: Cil.instr = Set(assignedValue,castExp,Cil.locUnknown) in
   Instr([instCast])
 
(* this function returns a cast stmtkind (*e.g. (int *)x *)
let mkAssStmtkind (valueToAssign: Cil.lval) (assignedValue: Cil.lval): stmtkind =
   let instCast: Cil.instr = Set(assignedValue,Lval(valueToAssign),Cil.locUnknown) in
   Instr([instCast])

 (* this function returns a cast stmtkind (*e.g. (int *)x *)
let mkAssStmtkindExp (valueToAssign: Cil.exp) (assignedValue: Cil.lval): stmtkind =
   let instCast: Cil.instr = Set(assignedValue,valueToAssign,Cil.locUnknown) in
   Instr([instCast])
 

(* this function builds a fresh int variable *) 
let freshIntVar (originatingFunction: Cil.fundec): Cil.lval =
 let name= "i"^(string_of_int !uniqueId) in
 let _ = (uniqueId := (!uniqueId)+1) in 
 (Var(makeLocalVar originatingFunction ~insert:true name Cil.intType),NoOffset)   

(* this function builds a fresh void * variable *) 
let freshVoidPtrVar (originatingFunction: Cil.fundec): Cil.lval =
 let name= "vp"^(string_of_int !uniqueId) in
 let _ = (uniqueId := (!uniqueId)+1) in 
 (Var(makeLocalVar originatingFunction ~insert:true name Cil.voidPtrType),NoOffset)   

(* this function builds a fresh void * variable *) 
let freshPtrVar (originatingFunction: Cil.fundec)(t:Cil.typ): Cil.lval =
 let name= "p"^(string_of_int !uniqueId) in
 let _ = (uniqueId := (!uniqueId)+1) in 
 (Var(makeLocalVar originatingFunction ~insert:true name t),NoOffset)   

 
(* this function builds a fresh int variable *) 
let freshIntVarinfo (originatingFunction: Cil.fundec): Cil.varinfo =
 let name= "i"^(string_of_int !uniqueId) in
 let _ = (uniqueId := (!uniqueId)+1) in 
 makeLocalVar originatingFunction ~insert:true name Cil.intType  

(* this function builds a fresh int variable *) 
let freshFloatVarinfo (originatingFunction: Cil.fundec): Cil.varinfo =
 let name= "f"^(string_of_int !uniqueId) in
 let _ = (uniqueId := (!uniqueId)+1) in 
 makeLocalVar originatingFunction ~insert:true name Cil.doubleType  

(* this function builds a fresh void * variable *) 
let freshPtrVarinfo (originatingFunction: Cil.fundec)(t:Cil.typ): Cil.varinfo =
 let name= "p"^(string_of_int !uniqueId) in
 let _ = (uniqueId := (!uniqueId)+1) in 
 (makeLocalVar originatingFunction ~insert:true name t)  


(* function used to make one type transformer function 
 * this function is recursive and should generate the instruction
 * that makes the effective transformation. It is used by doOneTransformer
 *)
   let rec doOneType  (originatingFunction:Cil.fundec) (t1:Cil.typ) (t2:Cil.typ) (input: lval)(output: lval) (nochange: bool ref) (pass: int) (generateStmt: bool): Cil.stmt list =
     (* we match the types and return the corresponding instruction list *)
      match t1,t2 with
       (* we put equal complex types first to be able to test equality afterwards
        * simple types
        **)
       (* well, this a complex type: what do we do by default?
        * T* ->T'*    ==  (T->T')*  ?
        * char* -> int* == ?
        *) 
       TPtr (t00, _), TPtr (t0, _) ->
           (*let _=  Printf.printf "Ptr2Ptr: the type transformer should be generated and the ref along if it is not the same attribute type\n" in *)
           (*this is the code for producing *t'=*t *)
              (* first we build the input value *)

              (*
              let inp:lval = (Mem(Lval(input)),NoOffset) in
              let out:lval = (Mem(Lval(output)),NoOffset) in

              doOneType originatingFunction t00 t0 inp out nochange (pass+1) generateStmt
              *)

           (*this is the code for producing t'=t *)
           (* WE PREFER THAT ONE AT THE MOMENT *)
              if (not generateStmt) then [] else
              let k2:stmtkind = mkAssStmtkind input output  in 
              let treat: stmt = Cil.mkStmt k2 in
              [treat] 
 


       (* well, this a complex type: what do we do by default?
        * [T] ->[T']    ==  [T->T']
        * 
        *) 
       | TArray (t00, Some l00,_), TArray (t0,Some l0,_) -> 
           (*let _=  Printf.printf "Array2Array: the type transformer [T] ->[T']    ==  [T->T'] should be generated and the ref along if it is not the same attribute type\n" in *)

           (* then we make the recursive call*)
           (* with for *)
           let index:Cil.varinfo = freshIntVarinfo originatingFunction in
           (* this is here that we will have the recursive call *)
           (* first we build the input value *)
           let inp:lval = Cil.addOffsetLval (Index((Lval(Var(index),NoOffset)),NoOffset)) input in
           (* then we make the output value *)
           let outp:lval = Cil.addOffsetLval (Index((Lval(Var(index),NoOffset)),NoOffset)) output in
           let l001 = Cil.constFold true l00 in
           let l01 = Cil.constFold true l0 in

           (match l001,l01 with
               (* we verify that the array size is affected *)
               Const(CInt64(max00,IInt,_)),Const(CInt64(max0,IInt,_)) when max00>=max0->
                (* if the array to affect is smaller than the previous version 
                 * We do a simple copy
                 *) 
                  let l= Const(CInt64(max0,IInt, None)) in
                  let _ = if (max0 <> max00) then nochange:= false in
                  (*we finally build the body for the For expression*)
                  let bodyFor:stmt list = doOneType originatingFunction t00 t0 inp outp nochange (pass+1) generateStmt
 in
                  let f= Cil.integer 0 in
                    
                  if (not generateStmt) then [] else
                  (* then we make the for loop that will be encoded by a while loop *)
                  let forList = mkForIncr ~iter:index ~first:f ~stopat:l ~incr:(Cil.integer 1) ~body:bodyFor in
                  forList
                
                |_,_ -> 
                  let _ = (nochange := false) in
                  (* if the array is bigger we initialize it first *)
                  (* we make an initialization for the array first *)
                  let rec baseElement(t:Cil.typ)(originatingFunction:fundec):(exp * varinfo) =
                     match t with
                       TInt (_,_) -> ((Const(CInt64(Int64.zero,IInt,None))),(freshIntVarinfo originatingFunction))
                       | TFloat (_,_) -> ((Const(CReal(0.0,FFloat,None))),(freshFloatVarinfo originatingFunction))

                       | TArray (t,_,_) -> (baseElement t originatingFunction)
                       | _ -> ((Const(CStr("We do not know any Array a priori initialization for this type: we do it only for basic types"))),(freshPtrVarinfo originatingFunction Cil.charType))
                  in

                  let rec baseType(t:Cil.typ):Cil.typ=
                     match t with
                       | TArray (t,_,_) -> (baseType t)
                       | _ -> t
                  in

                  let tmp= baseElement t0 originatingFunction in
                  let init: stmt list= 
                    (match tmp with
                       (Const(CStr(s)),_) -> 
                          let k2:stmtkind =   mkCallStmtkindNoReturn ("printf") (TVoid([])) ([(Const(CStr(s)))]) in 
                          let treat: stmt = Cil.mkStmt k2 in
                          [treat]
                       | (x,v) -> 
                           let assignement:stmtkind = mkAssStmtkindExp x ((Var(v)),NoOffset) in
                           let assi: stmt = Cil.mkStmt assignement in
                           assi :: (doOneType originatingFunction (baseType t00) t00 (Var(v),NoOffset) outp nochange (pass+1) generateStmt
) 
                     )
                  in


                  let bElse: block ={battrs=[];bstmts=init} in
                  let bThen: block ={battrs=[];bstmts=doOneType originatingFunction t00 t0 inp outp nochange (pass+1) generateStmt
} in
                  let ifInit: stmtkind = If((BinOp(Lt,(Lval(Var(index),NoOffset)),l00,Cil.intType)),bThen,bElse,Cil.locUnknown) in
                  let assi: stmt = Cil.mkStmt ifInit in

                  let f= Cil.integer 0 in

                  if (not generateStmt) then [] else
                  (* then we make the for loop that will be encoded by a while loop *)
                  let forList = mkForIncr ~iter:index ~first:f ~stopat:l0 ~incr:(Cil.integer 1) ~body:[assi] in
                 forList

              (* this is the case where we do not know anything *)
(*              | _,_ -> 
                 let l:exp = l00 in
 
                 let _ = (nochange := false) in
                 (*we build the body for the For expression in the general case*)
                 let bodyFor:stmt list = doOneType originatingFunction t00 t0 inp outp nochange in
                 let k2:stmtkind =   mkCallStmtkindNoReturn ("printf") (TVoid([])) ([Const(CStr("//REMOVE: We are not able to compute secure upper bound. Verify that uperbound is correct and initialization is correct. Finally decide if the type transformer should be kept\n"))]) in 
                 let treat: stmt = Cil.mkStmt k2 in

                 let f= Cil.integer 0 in
                 (* then we make the for loop that will be encoded by a while loop *)
                 let forList = mkForIncr ~iter:index ~first:f ~stopat:l ~incr:(Cil.integer 1) ~body:bodyFor in
                 (treat::forList)*)

         )
       | TComp (info1,_), TComp (info2,_) ->
           (* We first check if they are structs or unions true== struct, false == union*)
           (match info1.cstruct,info2.cstruct with
               true, true ->(
                (* First we check that the type is not into the hashtable structTT  
                 *
                 * This let us support the recursive structs.
                 * Basically the idea is that if we already defined a type transformer 
                 * for this struct, we only put a message.
                 *)
                
                if (pass=0) then
                ( 
                let _= H.add structTT t2 (originatingFunction.svar.vname) in 
                (* If we did not find any matching names, we use a much relaxed
                 * condition: one of them is the beginning of the other.
                 * 
                 * Discussion with Iulian: 
                 * we can do better with the application bijection and
                 * know exactly which goes with which
                 *)
                let matchFieldNames (s1:string) (s2:string): bool =
                   (Cil.startsWith s1 s2) || (Cil.startsWith s2 s1)
                in

               let listInst: stmt list ref= ref [] in
               (* these are two structs: we consider the name as the part that should
                * not change/have few changes 
                * we iterate on both structs and make assignements to identically 
                * named fields*) 
                  for i = 0 to ((List.length info2.cfields)-1) do
                     let isAssigned:bool ref = ref false in 
                        for j = 0 to ((List.length info1.cfields)-1) do
                           let f1:fieldinfo= (List.nth info1.cfields j) in
                           let f2:fieldinfo= (List.nth info2.cfields i) in         
                              if ((f1.fname) = (f2.fname)) then 
                                 let _= (isAssigned := true) in
                                    (*let _ = Printf.printf "i=%d,j=%d, f1.name = %s, f2.name = %s \n" i j f1.fname f2.fname in*)
                                    (* We know that we will have to transform the type *)
                                 let inp:lval = Cil.addOffsetLval (Field(f1,NoOffset)) input in
                                 let outp:lval = Cil.addOffsetLval (Field(f2,NoOffset)) output in 
                                    listInst := ((!listInst) @ (doOneType originatingFunction (f1.ftype) (f2.ftype) inp outp) nochange (pass+1) generateStmt) 
                        done;
                        (* if we did not find any correct transformation we make approximations *)
                        if (!isAssigned = false) then 
                           let _ = (nochange:=false) in
                              for j = 0 to ((List.length info1.cfields)-1) do
                                 let f1:fieldinfo= (List.nth info1.cfields j) in
                                 let f2:fieldinfo= (List.nth info2.cfields i) in         
                                    if ((matchFieldNames f1.fname f2.fname)) then
                                       let _= (isAssigned := true) in
                                          
                                       (*let _ = Printf.printf "i=%d,j=%d, f1.name = %s, f2.name = %s \n" i j f1.fname f2.fname in*)
                                       (* We know that we will have to transform the type *)
                                       let inp:lval = Cil.addOffsetLval (Field(f1,NoOffset)) input in
                                       let outp:lval = Cil.addOffsetLval (Field(f2,NoOffset)) output in 
                                          listInst := ((!listInst) @ (doOneType originatingFunction (f1.ftype) (f2.ftype) inp outp) nochange (pass+1) generateStmt
) 
                              done;
                              (* if the field cannot have a std init, we put a TODO statement *)
                              if (!isAssigned = false) then 
                                 let f2:fieldinfo= (List.nth info2.cfields i) in         
                                 let outp:lval = Cil.addOffsetLval (Field(f2,NoOffset)) output in 
                                 let k2:stmtkind =   mkCallStmtkind outp "printf" (TVoid([])) ([Const(CStr("//TODO We should probably put an initialization here\n"))]) in 
                                 let treat: stmt = Cil.mkStmt k2 in
                                    listInst := ((!listInst) @ [treat]) 
                  done;
      
                  if (not generateStmt) then [] else
                  (!listInst)) else
                     (* we are in the second pass: we make a simple assignement *)
                     let name = (match t1 with 
                             TComp(cinfo,_) -> cinfo.cname 
                             | _ -> "")
                     in

      
                     let doType = not (Cil.startsWith "__anon" name) in
 
                     if (doType=false) then
                     (
                        let tmp = doOneType originatingFunction t1 t2 input output nochange 0 generateStmt in
        
                        if (not generateStmt) then [] else tmp
                     )
                     else
                     (
                     if (not generateStmt) then [] else
                     let k2:stmtkind = mkAssStmtkind input output  in 
                     let treat: stmt = Cil.mkStmt k2 in
                     [treat] 
                     )
                 )
 
               | false, false -> 
               (* these are two unions: we consider the values that 
                * should have only few changes because they represent the same value
                * we compare all fields and if one is different we will include the TT
                * however, we decide to include only one of the possibilities fot the TT
                * we choose to have the longest one included in the TT.
                *)
               (let listInst: stmt list ref= ref [] in
                for i = 0 to ((List.length info2.cfields)-1) do
                    let isAssigned:bool ref = ref false in 
                    for j = 0 to ((List.length info1.cfields)-1) do
                      let f1:fieldinfo= (List.nth info1.cfields j) in
                      let f2:fieldinfo= (List.nth info2.cfields i) in         
                      if ((f1.fname) = (f2.fname)) then 
                        let _= (isAssigned := true) in
                        let inp:lval = Cil.addOffsetLval (Field(f1,NoOffset)) input in
                        let outp:lval = Cil.addOffsetLval (Field(f2,NoOffset)) output in
                        let l = (doOneType originatingFunction (f1.ftype) (f2.ftype) inp outp nochange (pass+1) generateStmt) in
                        if ((List.length l) > (List.length !listInst)) then listInst:= l
                    done;
                    if (not (!isAssigned))
                    then (nochange:=false) 
                done;
 
                if (not generateStmt) then [] else
                let k2:stmtkind =  mkCallStmtkindNoReturn "printf" (TVoid([])) ([Const(CStr("//TODO: verify that the transfer is ok/needed\n"))]) in 
                let treat: stmt = Cil.mkStmt k2 in

                (treat::!listInst))
               | _,_ -> (

                  let _ = (nochange:=false) in
                  if (not generateStmt) then [] else
                  let k2:stmtkind =   mkCallStmtkind output "printf" (TVoid([])) ([Const(CStr("//TODO type transformation from %d: we do not know what to do here\n"));Lval(input)]) in 
                  let treat: stmt = Cil.mkStmt k2 in
                  [treat])
         )
           
       | TNamed (_,_), TNamed (_,_) -> 
           (* let _=  Printf.printf "tnamed2tnamed(x)\n" in *)
           if (not generateStmt) then [] else
           let k2:stmtkind =   mkCastStmtkind t2 input output in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 

       (* This initializes a table with the value*)
       | t00,TArray (t0, Some Const(CInt64(max00,IInt,_)),_)->

           let _ = (nochange:=false) in
           (* then we make the recursive call*)
           (* with for *)
           let index:Cil.varinfo = freshIntVarinfo originatingFunction in
           let l= Const(CInt64(max00,IInt, None)) in
 
           (* then we make the output value *)
           let outp:lval = Cil.addOffsetLval (Index((Lval(Var(index),NoOffset)),NoOffset)) output in
           (*we finally build the body for the For expression*)
           let bodyFor:stmt list = doOneType originatingFunction t00 t0 input outp nochange (pass+1) generateStmt
 in
           let f= Cil.integer 0 in
           if (not generateStmt) then [] else
           (* then we make the for loop that will be encoded by a while loop *)
           let forList = mkForIncr ~iter:index ~first:f ~stopat:l ~incr:(Cil.integer 1) ~body:bodyFor in
           forList



      (* this should treat the case for simple type non-transformation w.g.:
        * int -> int, float -> float
        * in that case we do not generate any type transformer. 
        *)
       | t1,t2 when t1=t2 -> 
           if (not generateStmt) then [] else
           let k2:stmtkind = mkAssStmtkind input output  in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 
       (* when we have a pointer on one type transformed in the type, we de-alias it*)
       | TPtr(t3,_),t4 when t3=t4 -> 
           if (not generateStmt) then [] else
           let _ = (nochange:=false) in
           let k2:stmtkind =   mkCastStmtkind t4 (Mem(Lval(input)),NoOffset) output in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 


       (* Now we begin the type transforming functions.
        * we consider the following types: TVoid, TInt, TFloat, TPtr,
        * TArray, TFun, TNamed, TComp, TEnum, TBuiltin_va_list
        *)
       (* x -> void == we remove the data, should probably only be used with ptr->
        *)
       | _, TVoid (_) ->
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           let k2:stmtkind =   mkCallStmtkind output "printf" (TVoid([])) ([Const(CStr("//TODO type transformation from %d: we do not know what to do here\n"));Lval(input)]) in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat]
       (* x -> void == we initialize the data, should probably only be used with ptr->
        *)
       | TVoid (_), _ ->
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           let k2:stmtkind =   mkCallStmtkind output "printf" (TVoid([])) ([Const(CStr("//TODO type transformation from %d: we do not know what to do here\n"));Lval(input)]) in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat]
       (* int -> float == int2float
        *)
       | TInt (_, _), TFloat (_, _) -> 
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           (*let _=  Printf.printf "int2float(x)\n" in *)
           let k2:stmtkind =   mkCallStmtkind output "int2float" (TVoid([])) ([Lval(input)]) in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 
       (* int -> ptr == id
        *)
       | TInt (_, _), TPtr (t, _) -> 
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           (*let _=  Printf.printf "int2float(x)\n" in *)
           let k2:stmtkind = mkCastStmtkind t2 input output  in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 
       (* int -> ptr == id
        *)
       | TPtr (_, _), TInt (_, _) -> 
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           (*let _=  Printf.printf "int2float(x)\n" in *)
           let k2:stmtkind =   mkCastStmtkind t2 input output in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 

       (* any int type char -> int
        *)
       | TInt(_,_), TInt(_,_) ->
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           (* put here a matching with all int types or duplicate to have the correct predefined 
            * transtyping rules or leave it like that *) 
           let k2:stmtkind =   mkCastStmtkind t2 input output in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 


       (* float -> int == float2int
        *)
       | TFloat (FFloat, _), TInt (IInt, _) ->
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           (*let _=  Printf.printf "float2int(x)\n" in *)
           let k2:stmtkind =   mkCallStmtkind output "float2int" (TVoid([])) ([Lval(input)]) in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat] 
       (* function to function with arguments... *)
       | TFun(t0,Some at0,_,_), TFun(t00,Some at00,_,_) ->
           let rec listComp (l1: (string*Cil.typ*Cil.attributes) list) (l2: (string*Cil.typ*Cil.attributes) list) : bool =
             if ((List.length l1)<>(List.length l2)) then false
             else 
              if ((List.length l1)=0) then true
              else
               (match (List.hd l1),(List.hd l2) with 
                  (_,t01,_),(_,t002,_) -> 
                    if (t01 <> t002) then false
                    else listComp (List.tl l1) (List.tl l2)
               )
           in
           if ((t0 = t00)&&(listComp at0 at00)) then(
             if (not generateStmt) then [] else
             let k2:stmtkind =   mkCastStmtkind t2 input output in 
             let treat: stmt = Cil.mkStmt k2 in
             [treat] )
           else
             let _ = (nochange:=false) in
             if (not generateStmt) then [] else
             let k2:stmtkind =   mkCallStmtkindNoReturn ("printf") (TVoid([])) ([Const(CStr("//TODO we don't know what to do with differently typed functions\n"))]) in 
             let treat: stmt = Cil.mkStmt k2 in
             [treat]

       (* function to function with no argument... *)
       | TFun(t0,None,_,_), TFun(t00,None,_,_) ->
            if ((t0 = t00)) then(
             if (not generateStmt) then [] else
             let k2:stmtkind =   mkCastStmtkind t2 input output in 
             let treat: stmt = Cil.mkStmt k2 in
             [treat] )
           else
             let _ = (nochange:=false) in
             if (not generateStmt) then [] else
             let k2:stmtkind =   mkCallStmtkindNoReturn ("printf") (TVoid([])) ([Const(CStr("//TODO we don't know what to do with differently typed functions\n"))]) in 
             let treat: stmt = Cil.mkStmt k2 in
             [treat]

       | TEnum(_,_), TEnum(_,_) ->
           if (not generateStmt) then [] else
           let k2:stmtkind =   mkCallStmtkind output "printf" (TVoid([])) ([Const(CStr("//TT We do not do any transformation from one enumeration to the other one as cil already inlines/do not treat them.\n"))]) in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat]


       (* we really don't know what to do at this point...
        * example: int -> (int->int) (PB: this case raises an exception)
        *)
       |_,_ -> 
           let _ = (nochange:=false) in
           if (not generateStmt) then [] else
           let k2:stmtkind =   mkCallStmtkind output "printf" (TVoid([])) ([Const(CStr("//TODO type transformation from %d: we do not know what to do here\n"));Lval(input)]) in 
           let treat: stmt = Cil.mkStmt k2 in
           [treat]
  ;; 


   (* function used to make one type transformer function 
    * this function is the one that effectively makes the function body
    * that will transform variables of one named type into its new version
    *)
   let rec doOneTransformer (t1:Cil.typ) (t2:Cil.typ) (name:string) : Cil.fundec=
      (* first of all we build the function body *)
      let funtype: Cil.typ = TVoid [] in

      (* we build here the name of the function that we are building *)

      (* extract the labelled name of the type. e.g. "type \"myInt\"" -> "myInt" *)
      let typeName =    
         let base = Commoncilutil.extractGlobalName name in
            (* differentiate between struct foo and typedef X foo  *)
            match t2 with 
                  TNamed _ -> "typedef_" ^ base
               | _ -> base
      in
      
      let v: varinfo = makeGlobalVar (buildTypeTransformerName typeName oldVersionTag newVersionTag) funtype in
      (* PB: this init is very shitty can't we have better????? *)
      let f: Cil.fundec={
          svar=v;sformals=[];slocals = []; smaxid=0; sbody={battrs=[];bstmts=[]}; smaxstmtid=None;sallstmts=[]
      } 
      in

      let output: lval = (Mem(Lval(Var(makeLocalVar f ~insert:false "xout"  (TPtr(t2,[]))),NoOffset)),NoOffset) in
      let input:lval= (Mem(Lval(Var(makeLocalVar f ~insert:false "xin" (TPtr(t1,[]))),NoOffset)),NoOffset) in

      (* input and output are concrete data.  If you want to reference what the TT converts 
         outside of the TT then you need to point to the wrapper.  Therefore, we give the TT a pointer to the wrapper 
      let wrapped_output:lval = (Mem(Lval(Var(makeLocalVar f ~insert:false "xnew" ( (TPtr(t2,[])) )),NoOffset)),NoOffset) in
      *)

      (* we set the return type of the function if it is an array, we return a 
       * pointer on the same type*)
      let _= Cil.setFunctionType f (TFun((TVoid([])),Some [], false, []))(*TFun((TPtr(t2,[])),Some [], false, [])*) in
      (* we build the type of the input variable and insert it in the parameters of the type transformer *)
      let varInput:varinfo = makeVarinfo true "xin" (TPtr(t1,[])) in
      let varOutput:varinfo = makeVarinfo true "xout" (TPtr(t2,[])) in
      let varNew:varinfo = makeVarinfo true "xnew" (TPtr(t2,[])) in
      let _= Cil.setFormals f [varInput;varOutput;varNew] in

      

      (* Then we make the call to the recursive function that makes the
       * type transformer instruction list
       *)
      let nochange: bool ref = ref true in

      let doType = not (Cil.startsWith "__anon" name) in 

(*       if (doType=false) then
          try 
           E.s (E.log "Omitting type transformer because this is an anonymous type:\n %s \n" typeName) 
          with  Errormsg.Error -> ()
       else*)
      let inst: Cil.stmt list = doOneType f t1 t2 input output nochange 0 true in
    
      (* make no return statement  *) 
      (* integrate it in the list of statements, at the end *)
      let b: block ={battrs=[];bstmts=inst} in

      (* this list of statements is put as body of the type transformer *)
      f.sbody <- b;
      f

   
