(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Manuel Oriol        <moriol@inf.ethz.ch> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* This feature is made to generate type transformers that transform values of an old version of the type to the new one.
By default the functionning of typetransformer.ml is as follows:

cilly --dotypetransformer --merge  test0-1.c test0-2.c 

The standard output should be filled in with the pretty printed version of the type transformers generated functions.
Note that as there are error messages in cil when redefining types the standard output should probably be redirected into a file to have a very good presentation.

In the code, the function doOneType is the most important.

 doOneType  (originatingFunction:Cil.fundec) (t1:Cil.typ) (t2:Cil.typ) (input: lval)(output: lval) : Cil.stmt list 

It is a recursive function that takes an originating function which is intended to be the type transformer function itself.
t1 is the old version of the type.
t2 is the new version of the type.
input is the input variable (having type t1).
output is the output variable (having type t2).

It appeared to be useful to have the separation from t1,t2 and input,output for the simple reason that we may have strange transtyping things that happen in C.
This function returns a stmt list that constitutes the statements of the type transformer.

TODO: semantic unification of variables (see discussion with Iulian)

Manuel
*)


open Cil

module H = Hashtbl
 

type hashOfGlobals = (string, global) H.t
type hashOfTT = (Cil.typ,string) H.t


let stage : int ref = ref 0

let out : out_channel ref = ref stdout

let oldVersionTag:string  = "old"
let newVersionTag:string  = "new"

type typeDirective = 
  ReplacedType of (string*string)
  | UnchangedType of string
  | NewType of string
  | DeletedType of string

type listOfDirectives = typeDirective list

(* this hashtable type will store the types
 * the first string is the name under which it is existing in the new version
 *    if it does not exist anymore the first string will be prefixed old_
 * the string is the old name (NOT prefixed by old_)
 * the typ is the old type declaration
 * the typ is the new type declaration
 * the boolean indicates if it has changed/new or did not changed/disappeared 
 * the fundec is the generated type transformer
 *)
type hashOfTransformedTypes = (string, (string option* typ option* typ option * bool * fundec option)) H.t



(* code used to build the names differentiated for versions *)
let versionSeparator = "_"
let typeTransformerBaseName = "TT_"

(* a function that builds the type name *)
let buildTypeName (baseTypeName: string)(versionTag : string) :string = 
   (baseTypeName ^ versionSeparator ^ versionTag)


(* a function that builds the type transformer name *)
let buildTypeTransformerName (baseTypeName: string)(oldVersionTag : string) (newVersionTag : string):string = 
   (typeTransformerBaseName ^ baseTypeName (*^ versionSeparator ^ oldVersionTag ^ versionSeparator ^ newVersionTag*))


(*
 * Hashtables that contains the globals from the two versions.
 *)
let globalsFile1: hashOfGlobals  = H.create 65535
let globalsFile2: hashOfGlobals  = H.create 65535

(* filter on globals; only type definitions and functions will pass *)
let filterOutGlobalDeclarations (g : Cil.global) : bool = 
  match g with
      GCompTagDecl _ | GEnumTagDecl _ | GVarDecl _ | GAsm _ | GPragma _ | GText _ -> false
    | _ -> true 

(* Given a file, fill out the hash will all the globals that pass the given filter *)
let fillGlobalHash (file : Cil.file) (filterFun : Cil.global -> bool) (hash : hashOfGlobals) = 
  Cil.iterGlobals 
    file 
    (fun g -> if filterFun g then H.add hash (Commoncilutil.getGlobalName g) g)


(*
 * This function does the analysis of the transformed
 * types and builds a transforming function from an old type to the new one.
 *  
 **)
let constructTypeTransformers (l : listOfDirectives) (globalsFile1: hashOfGlobals) (globalsFile2: hashOfGlobals) : hashOfTransformedTypes  =
  let hResult: hashOfTransformedTypes = H.create 65535 in
  
   
   (* first we will build the list of all types names *)
   let lTypNames: (string, unit) H.t = H.create 4099 in

   (*
    * this function is here to add the name of the types to a pool of globals to treat
    * name: the name of the functions
    * global: the processed global 
    * 
    *)
   let addToNames (name: string)(g: Cil.global) =
    (* extract the labelled name of the type. e.g. "type \"myInt\"" -> "myInt" *)
     let tname = Commoncilutil.extractGlobalName name in 
     begin
       if not (Commoncilutil.isAnonymous tname) then 
         match g with
             (* differentiate between struct foo and typedef X foo  *)
             GType(_,_)
           | GCompTag (_, _) 
           | GEnumTag(_,_)  -> Commonutil.hashtbl_add_once lTypNames name ()

           | _ -> ()
       
     end
   in
  
   let _ = H.iter addToNames globalsFile1 in
   let _ = H.iter addToNames globalsFile2 in

   (* we remove the names that are used in the directives *)
   let removeFromNames_iter (fd: typeDirective) =
     let removeNoQuestions s = 
       try
         H.remove lTypNames s
       with 
         Not_found -> ()
     in
       match fd with 
           ReplacedType (s1,s2) -> 
             removeNoQuestions s1; 
             removeNoQuestions s2

         | UnchangedType(s1) 
         | NewType (s1) 
         | DeletedType (s1) ->
             removeNoQuestions s1  
   in
   let _ = List.iter removeFromNames_iter l in
   
   let getTypeFromGlobal (g : Cil.global) : Cil.typ = match g with
       GType(ti,_) -> TNamed(ti,[]) (* ti.ttype *)
     | GCompTag(ci,_) -> TComp(ci,[]) 
     | GEnumTag(ei,_) -> TEnum(ei,[]) 

     |  _ -> TVoid([])
   in

   let oldTypList : global list ref = ref [] in
   (* initialize hResult with users directives *)
    let initializeHashTable (fd: typeDirective) =
      (
       match fd with 
          ReplacedType(s1,s2) -> 
             (try
               let g1 = H.find globalsFile1 s1 in
               let g2 = H.find globalsFile2 s2 in
               let _ = (oldTypList:= g1::!oldTypList) in
  
               let t1 = getTypeFromGlobal g1 in
               let t2 = getTypeFromGlobal g2 in
               let tt = Typeprocessing.doOneTransformer t1 t2 s2 in
               H.add hResult s2 (Some s1,Some t1,Some t2,true, Some tt)
             with Not_found -> ())
          | UnchangedType(s)  ->  
            (try
               let g1 = H.find globalsFile1 s in
               let _ = (oldTypList:= g1::!oldTypList) in
               let t = getTypeFromGlobal g1 in
               H.add hResult s (Some s,Some t,Some t,false, None)
             with Not_found -> ())

          | NewType(s2) -> 
            (try
               let g2 = H.find globalsFile2 s2 in
               let t2 = getTypeFromGlobal g2  in
               H.add hResult s2 (None,None,Some t2,true, None)
             with Not_found -> ())

          | DeletedType(s1) ->
             (try
               let g1 = H.find globalsFile1 s1 in
               let _ = (oldTypList:= g1::!oldTypList) in
               let t1 = getTypeFromGlobal g1  in
               H.add hResult ("old_"^s1) (Some s1,Some t1,None,false,None)
               with Not_found -> ())


      ) in
        

   let _ = List.iter initializeHashTable l in


   (* complete the hashtable with unspecified types *) 
   let completeHashTable_iter (name: string) () =
       (
        try 
          let g1 = H.find globalsFile1 name in
          let g2 = H.find globalsFile2 name in
          (  
             let t1 = getTypeFromGlobal g1  in
             let t2 = getTypeFromGlobal g2  in
             let _ = (oldTypList:= g1::!oldTypList) in
             let isModified =(not (Modifiedfunctions.isEqualTyp t1 t2)) in
             let tt = Typeprocessing.doOneTransformer t1 t2 name in
             H.add hResult name (Some name,Some t1,Some t2, isModified, Some tt)
             
          )
        with Not_found ->(
         try 
          let g1 = H.find globalsFile1 name in
          let _ = (oldTypList:= g1::!oldTypList) in
          (
             let t1 = getTypeFromGlobal g1 in
             H.add hResult ("old_"^name) (Some name,Some t1,None, false, None)
          )
        with Not_found -> (
          let g2 = H.find globalsFile2 name in
          (
             let t2 = getTypeFromGlobal g2  in
             H.add hResult name (None,None, Some t2, true, None)
          )
      ))) 
   in
   (* we eventually iterate on all the functions from globalsFile1 *)
   let _ = H.iter completeHashTable_iter lTypNames in
   hResult
   
   
(*
 * Initialization of the analysis.
 *)

let pre () =  stage := !stage + 1


(*
 * Finalization of the analysis.
 * 
 * 
 **)
let post () =  
  begin
    if (2 = !stage) 
    then 
   let l: listOfDirectives = [] in
   let statsNew = ref 0 in
   let statsChanged = ref 0 in
   let statsDisappeared = ref 0 in
   let statsUnchanged = ref 0 in
   let statsUnexplained = ref 0 in
   let ttGenerated = ref 0 in
  
   let printResult(name:string)(change:(string option* typ option* typ option * bool *fundec option)) = 
       match change with
         (None, None, Some(t),true,None) ->
                let _= (statsNew:= !statsNew +1) in
                (Printf.printf "The type %s is new\n" name)
         | (Some(s1), Some(t1), Some(t2),true,Some tt) ->
                let _= (statsChanged:= !statsChanged +1) in
                let _= (ttGenerated:= !ttGenerated +1) in
                let _ = try Pretty.fprint !out ~width:80 (d_global () (GFun(tt,locUnknown))) with Errormsg.Error -> () in
                (Printf.printf "The type %s is a new version of %s\n" name s1)
         | (Some(s1), Some(t), None,false, None) ->
                let _= (statsDisappeared:= !statsDisappeared +1) in
                (Printf.printf "The type %s has disappeared\n" s1)
         | (Some(s1), Some(t1), Some(t2),false, Some tt) ->
                let _= (statsUnchanged:= !statsUnchanged +1) in
                let _= (ttGenerated:= !ttGenerated +1) in
                (Printf.printf "The type %s has not changed\n" s1)
         | _ -> 
                let _= (statsUnexplained:= !statsUnexplained +1) in
                (Printf.printf "hum... that is strange...")

 
   in 
     let hashResult = constructTypeTransformers l globalsFile1 globalsFile2 in 
     let _ = H.iter printResult hashResult in
      
     (Printf.printf "Total result:\n------------------------- \n New Types: %d\n Types Modified: %d\n Types Removed: %d\n Types Unchanged: %d\n Types Unexplained: %d\n Type Transformers Generated %d\n" !statsNew !statsChanged !statsDisappeared !statsUnchanged !statsUnexplained !ttGenerated)
      
   end

(* this is called for each file in turn; 
   for the second file, post() will do the TT inference and printing 
 *)
let ttProcessFile (f : Cil.file) : unit = 
  pre ();
  begin
    if !stage = 1 
    then fillGlobalHash f filterOutGlobalDeclarations globalsFile1
    else fillGlobalHash f filterOutGlobalDeclarations globalsFile2
  end;
  post ()

  
(*
 * Type transformer feature descriptor
 *)
let feature : featureDescr = 
  { fd_name = "typetransformer";
    fd_enabled = Cilutil.typetransformerEnabled;
    fd_description = "generate type transformers from an old version to a new version";
    fd_extraopt = [];
    fd_doit = (function (f: file) -> ttProcessFile f);
    fd_post_check = true;
  } 

