(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* misc util functions that perform generic operations on CIL structures 
   They're only needed for update; the rest of them needed by semsdiff as well are in
   common/commoncilutil
*)

open Cil

module H = Hashtbl

(* extend an lval with an offset *)
let rec addOffsets o1 o2 = 
   match o1 with 
      Field (fi,o) -> Field(fi,addOffsets o o2)
    | Index (e,o) -> Index(e,addOffsets o o2)
    | NoOffset -> o2




(* CIL's lval structure is hard to work with, so we have our own *)
type ulval = 
   UVar of Cil.varinfo
 | UMem of Cil.exp 
 | UField of ulval * Cil.fieldinfo
 | UIndex of ulval * Cil.exp


(* convert one of CIL's lvals to one of ours *)
let lvalToUlval lval =
   let rec convertOffset offset base f = 
      match offset with
         Field (fi,o) -> convertOffset o base (fun ul -> UField(f ul, fi))
       | Index (e,o) -> convertOffset o base (fun ul -> UIndex(f ul, e))
       | NoOffset -> f base
   in
   let id = fun x -> x in
   match lval with 
      (Var vi, o) -> convertOffset o (UVar vi) id
    | (Mem e, o) ->  convertOffset o (UMem e) id

(* convert one of out lvals to a CIL lval *)
let rec ulvalToLval ulval =
   match ulval with 
     UVar vi -> (Var vi, NoOffset)
   | UMem e -> (Mem e, NoOffset)
   | UField (ulval',fi) ->
      let (host,offset) = ulvalToLval ulval' in 
      (host, addOffsets offset (Field(fi,NoOffset)))
   | UIndex (ulval',e) ->
      let (host,offset) = ulvalToLval ulval' in 
      (host, addOffsets offset (Index(e,NoOffset)))





(* Split the given global list into two list, the first the globals
   before the functions and the second the list of functions *)
let splitGlobalsBeforeFuncs (globals : Cil.global list) : (Cil.global list) * (Cil.global list) = 
   let rec splitGlobalsBeforeFuncs' before after =
     match after with
        [] -> (before,after)
      | h::t -> 
        begin 
        match h with
           GFun(_, _) -> (before,after)
        |  _ -> splitGlobalsBeforeFuncs' (before @ [h]) t
        end
   in 
   splitGlobalsBeforeFuncs' [] globals

(* Split the given global list into two list, the first the globals
   before the var decls/defs and the second the rest *)
let splitGlobalsBeforeVars (globals : Cil.global list) : (Cil.global list) * (Cil.global list) = 
   let rec splitGlobalsBeforeVars' before after =
     match after with
        [] -> (before,after)
      | h::t -> 
        begin 
        match h with
           GVar _ | GVarDecl _ | GFun _ -> (before,after)
        |  _ -> splitGlobalsBeforeVars' (before @ [h]) t
        end
   in 
   splitGlobalsBeforeVars' [] globals


let typeOfEnum = (TInt(IInt,[]))

(* Compute the number of unrolling needed to reach a concrete type *)
let typeHeight t1 = 
   let rec typeHeight' t1 n = 
      match t1 with
         TNamed (r, _) -> typeHeight' r.ttype (n+1)
       | _ -> n
   in 
   typeHeight' t1 0


(**** Compute the type of an expression; partly stolen from CIL ****)
let rec properTypeOf (e: exp) : typ = 

  match e with
  | Const(CInt64 (_, ik, _)) -> TInt(ik, [])

    (* Character constants have type int.  ISO/IEC 9899:1999 (E),
     * section 6.4.4.4 [Character constants], paragraph 10, if you
     * don't believe me. *)
  | Const(CChr _) -> intType

    (* The type of a string is a pointer to characters ! The only case when 
     * you would want it to be an array is as an argument to sizeof, but we 
     * have SizeOfStr for that *)
  | Const(CStr s) -> !stringLiteralType

  | Const(CWStr s) -> TPtr(!wcharType,[])

  | Const(CReal (_, fk, _)) -> TFloat(fk, [])
  | Lval(lv) -> properTypeOfLval lv
  | SizeOf _ | SizeOfE _ | SizeOfStr _ -> !Cil.typeOfSizeOf
  | AlignOf _ | AlignOfE _ -> !Cil.typeOfSizeOf
  | UnOp (_, _, t) -> t
  | BinOp (_, _, _, t) -> t
  | CastE (t, _) -> t
  | AddrOf (lv) -> TPtr(properTypeOfLval lv, [])
  | StartOf (lv) -> begin
      match  (properTypeOfLval lv) with
          TArray (t,_, a) -> TPtr(t, a)
        | TNamed _ as typ -> typ 
        | _ -> Cil.typeOf e (* FIXME: why doesn't the code below do the job ? *)
(*          begin *)
(*            match  (Cil.typeOfLval lv) with *)
(*                TArray (t,_, _) -> TPtr(t, []) *)
(*              | _ -> Updatedebug.abort (Updatedebug.bug "properTypeOf: StartOf on a non-array (%s:%d)" !currentLoc.file !currentLoc.line) *)
(*          end *)
    end

and properTypeOfInit (i: init) : typ = 
  match i with 
    SingleInit e -> properTypeOf e
  | CompoundInit (t, _) -> t

and properTypeOfLval lv = 
  let _ = if Updateconfig.debug_properType 
          then ignore (Updatedebug.log [Updatedebug.BTAlways] "properTypeOfLval (%s:%d) [%a]\n" !currentLoc.file !currentLoc.line d_plainlval lv) in
  let res = 
    match lv with
        Var vi, off -> properTypeOffset vi.vtype off
      | Mem addr, off -> 
          let pa = properTypeOf addr in
          let rec deref typ = 
             begin match typ with
                 TPtr (t, _) -> properTypeOffset t off
               | TArray (t, _, _) -> properTypeOffset t off
               | TNamed (ti,_) -> 
                  (* The named type must unroll to a pointer, if not this is an error. *)
                  deref ti.ttype
               | _ -> Updatedebug.abort (Updatedebug.bug 
                         "properTypeOfLval: Mem on non-pointer, non-array type : %a, type %a\n" 
                          d_plainlval lv d_plaintype pa)
             end
          in
          deref pa
  in
    if Updateconfig.debug_properType then ignore (Updatedebug.log [Updatedebug.BTAlways] "  %s\n" (Updatedebug.ppType res));
    res

and properTypeOffset basetyp off = 
  let _ = 
    if Updateconfig.debug_properType 
    then ignore (Updatedebug.log [Updatedebug.BTNamedtypes] "properTypeOffset (%s:%d) [%a] [%a] = " 
                   !currentLoc.file !currentLoc.line d_plaintype basetyp Commoncilutil.my_d_plainoffset off) 
  in
  let res = 
    match off with
             NoOffset -> basetyp
      | Index (_, o) -> begin
               match basetyp with
              TArray (t, _, _) -> properTypeOffset t o
                 | TNamed (bt, _) -> properTypeOffset bt.ttype off
                 | t -> Updatedebug.abort (Updatedebug.bug "(%s:%d) properTypeOffset: Index on a non-array (type %a)\n" 
                                             !currentLoc.file !currentLoc.line d_plaintype basetyp)
             end 
      | Field (fi, o) -> properTypeOffset fi.ftype o
  in
    if Updateconfig.debug_properType then ignore (Updatedebug.log [Updatedebug.BTNamedtypes] "  [%a]\n" d_plaintype res);
    res



(* Extract the return type and argument types from a fundec *)
let getTypesInFunDec (fd : Cil.fundec) : Cil.typ * (Cil.typ list) =
  match fd.svar.vtype with
      TFun(rt,Some args, isva, a) -> (rt, Commoncilutil.argsToTypesList (Some args))
    | TFun(rt,None,isva,a) -> (rt, []) 
    | _ -> Updatedebug.abort (Updatedebug.bug "getTypesInFunDec: type of fundec not TFun!")

let argsFromType t = 
  match t with
      TFun(rt,Some args, isva, a) -> args
    | TFun(rt,None,isva,a) -> [] 
    | _ -> Updatedebug.abort (Updatedebug.bug "argsFromType: type not TFun!")

(* get the return type of a function *)
let getReturnType (fd : Cil.fundec) : Cil.typ =
  fst (getTypesInFunDec fd)

(* given an expression (e.g pointer to function), try to make the function type out and return it *)
let getTypesInFunExp (exp:Cil.exp) : Cil.typ * (Cil.typ list) * bool =
  let rec getTypesInternal t = 
  match t with
      TFun(rt,Some args,isva,a) -> (rt, Commoncilutil.argsToTypesList (Some args),isva) 
    | TFun(rt,None,isva,a) -> (rt,[],isva) 
    | TPtr(t',_) | TArray(t',_,_) -> getTypesInternal t'
    | TNamed (bt, _) -> getTypesInternal bt.ttype
    | _ ->
       Updatedebug.abort (Updatedebug.bug "getTypesInFunExp: exp doesn't unroll to a TFun!")
  in
    getTypesInternal (Cil.typeOf exp)





let getAbstractTypeName (t : Cil.typ) : string =
   match t with 
      TComp (comp, attr) -> 
         if comp.cstruct then
            "struct_" ^ comp.cname
         else
            "union_" ^ comp.cname

    | TEnum (enum, attr) -> "enum_" ^ enum.ename

    | TNamed (bt, attr) -> bt.tname

    | TBuiltin_va_list _ -> "TBuiltin_va_list"

    | _ -> 
       Updatedebug.abort (Updatedebug.bug 
           "Tried to calculate the canonical type name for the non abstract type %a" 
           d_plaintype t)


(* determine if the expression exp is a null initialiser for some data of 
   type typ (which should be a pointer) *)
let rec isNullInitialiser loc (typ : Cil.typ) (exp : Cil.exp) : bool =
   match typ,exp with
      t, CastE (t',e) -> isNullInitialiser loc t e
      
    | t, Const (CInt64(n,_,_)) when (Int64.compare n Int64.zero) = 0 ->
         begin match (Cil.unrollType t) with 
            TPtr _ -> 
(*             ignore (Updatedebug.log [Updatedebug.BTUpdateutil] "(%a) set null initialiser type = %s\n" d_loc loc (Updatedebug.ppType t)); *)
            true
          | _ -> false
         end

    | _ -> false



let isCompType (typ: Cil.typ) : bool = 
  match typ with
    TComp _ -> true
  | _ -> false

(* ---- Visitor to remove effect attributes ---- *)
class removeEffectAttrVisitor(effName) : cilVisitor = object(self)
  inherit nopCilVisitor

   method vattr attr = 
      match attr with 
         Attr(name,_) when name = effName ->
            ChangeTo []
       | _ -> DoChildren
end

(* ---- Creates the update function ---- *)
let createUpdateFunDecl (name:string) (f:Cil.file) : Cil.varinfo = 

   let updateFunctionType = Updateconfig.updateFunctionType in
   let updateFunction = Cil.makeGlobalVar name updateFunctionType in
   let () = updateFunction.vstorage <- Extern in
   let updateFunctionGlobal = GVarDecl (updateFunction, Cil.locUnknown) in

   (* the update() function will come after global definitions and right before the other functions 
      we could actually insert it at the very beginning *)
   let splitglobals = splitGlobalsBeforeFuncs f.globals in
   f.globals <- (fst splitglobals) @ [updateFunctionGlobal] @ (snd splitglobals);
   updateFunction



