(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* $Id: updatemain.ml,v 1.29 2004/11/16 17:24:25 neamtiu Exp $ *)

open Cil

module H = Hashtbl

let initInternalFun () =
  let f (fname:string) : bool = 
    try 
      Namedtypes.isConAbsFuncName fname 
    || Updateconfig.isUpdateFunc fname 
    || Updateconfig.isEmptyTT fname 
    || Indirection.FunctionInfo.isWrapper fname 
    with Not_found -> false
  in
  Versiondata.isInternalFunction := f
;;

let doRemoveStaticSpecifiers (file : Cil.file) : unit = 
  Removestatics.removeStaticSpecifiers file;;

let doEffectsAnalysis (file : Cil.file) (mallocFunHash : Mallocpragma.mallocHash_t) : unit = 
  Effects.doEffectsAnalysis file mallocFunHash;;

let doElimStructTypedefs (file : Cil.file) : unit = 
  if Updateconfig.enableElimStructTypedefs 
  then Elimtypedef.elimStructTypedefs file
  else ();;

let doGlobalVarsUniqTypes (file : Cil.file) (vd : Versiondata.version_data) : unit = 
  if Updateargs.args.Updateargs.arg_dogvartypes 
  then Gvartypes.genGlobalVarUniqTypes file vd
  else ();;

let doConabsinsertion (mallocFunHash : Mallocpragma.mallocHash_t) 
                      (vd : Versiondata.version_data) 
                      (file : Cil.file) : unit = 
   Conabsinsertion.default_insert_coercions mallocFunHash vd file;;

let doPtraliastracking (mallocFunHash : Mallocpragma.mallocHash_t) (file:Cil.file) : unit =
   Ptraliastracking.doNamedTypesTracking mallocFunHash file

let doExternaltypeselim (vd:Versiondata.version_data) (file:Cil.file) : unit = 
   visitCilFile (new Externaltypeselim.externalTypesElimVisitor vd) file;;

let doNonUpdateableConAbsRemoval (file:Cil.file) : unit = 
   visitCilFile (new Ptraliastracking.nonUpdatableConAbsRemovalVisitor) file;;

let doBitfieldPinDownVisitor (file:Cil.file) : unit = 
   visitCilFile (new Bitfieldpindown.bitfieldPinDownVisitor) file;;

let doSizeofPinDownVisitor (file:Cil.file) : unit = 
   if Updateconfig.sizeofPinDownEnabled then
      visitCilFile (new Sizeofpindown.sizeofPinDownVisitor) file;;

let doInsertWrappers (file:Cil.file) (vd : Versiondata.version_data) : Cil.varinfo = 
  (* side-effecting, replaces f.globals in place *)
   let memcpy_vi = Wrappers.insertWrapperTypes file vd in
   file.globals <- Wrappers.fixWrapperInits file;
   memcpy_vi;;

let doIndirection (vd : Versiondata.version_data) (file:Cil.file) : unit = 
  (* side-effecting, replaces f.globals in place *)
  Indirection.insertIndirections vd file;;

let doInsertUpdatePoints (file:Cil.file) : unit = 
  (* side-effecting, replaces f.globals in place *)
  Updatepoints.insertUpdatePoints file;;

let doAddFunctionGCinfrastructure (vd : Versiondata.version_data) (file:Cil.file) : unit = 
  (* side-effecting, replaces f.globals in place *)
  Indirection.addFunctionGCinfrastructure vd file;;

let doExternalFunctionsVoidStarArgs (file:Cil.file) : unit = 
  (* read-only *)
  Externalvoidstarfunctions.default_external_functions_void_star_args(file);;

let doLoops (file:Cil.file) : varinfo =
  (* side-effecting *)
  Loops.default_update_loops file;;

let doVoidTypedefElim (file:Cil.file) : unit = 
   Voidtypedefelim.voidTypedefElim file;;


let doAnalysis (mallocHash : Mallocpragma.mallocHash_t) 
               (funEffectHash : Analysis.funEffectHash_t) 
               (funTypeHash : Analysis.funTypeHash_t) 
               (deltaSolutionHash : Analysis.deltaSolutionHash_t) 
               (setjmpHash : Setjmp.setjmpHash_t) 
               (file:Cil.file) 
               (suspect_result : Suspectalias.SAM.mo_result)
               (internVarHash : (string, unit) H.t)
               : Analysis.analysis_result  = 
   Analysis.runAnalysis mallocHash funEffectHash funTypeHash deltaSolutionHash setjmpHash file suspect_result internVarHash


let doTypeIndirection (vd : Versiondata.version_data) (file:Cil.file) : unit = 
   Typeindirection.genTypeInfastructure vd file;;

let doPopulateVersionData (vd : Versiondata.version_data) (file:Cil.file) (funTypeHash : Analysis.funTypeHash_t) : unit = 
   Versiondata.populateVersionData vd file funTypeHash;;

let doFunVersioning (vd : Versiondata.version_data) (file : Cil.file) : (string, string) H.t = 
   Funversioning.funVersioning vd file;;

let doCollectTTPragmas (vd : Versiondata.version_data) (file:Cil.file) : unit = 
   Ttpragma.collectTTPragmas vd file;;

let doCollectForcedPragmas (file:Cil.file) : unit = 
   Forcedpragma.collectForcedPragmas file;;

let doMarkForcedNonUpdatableTypes () : unit = 
   Namedtypes.setForcedNonUpdatableTypes ();;

let doCreatePatchInstallCode (vd : Versiondata.version_data) (pi : Patchcomp.patch_info) (file:Cil.file) : unit = 
   Patchcomp.makePatchInstallFunction vd pi file

let doGenTTWrappers (file:Cil.file) (vd : Versiondata.version_data) (memcpy_vi : Cil.varinfo) : unit = 
   Ttwrapper.genTTWrappers file vd memcpy_vi;;

let doInsertCallToUpdatelibInit (file:Cil.file) : unit = 
   ignore(Updatelibinit.insertCallToUpdatelibInit file);;

let doMergePIintoVD (pi : Patchcomp.patch_info) (vd : Versiondata.version_data) (file:Cil.file) (funTypeHash : Analysis.funTypeHash_t) : unit = 
   Patchcomp.mergePIintoVD pi vd file funTypeHash;;

let doSaveVersionData (filename:string) (vd : Versiondata.version_data) : unit = 
   Versiondata.saveVersionData vd filename;;

let doPatchSplitting (file : Cil.file) (pi : Patchcomp.patch_info) (versionedFunHash : (string, string) H.t) : unit = 
   if Updateargs.args.Updateargs.arg_dopatchsplitting  
   then Patchsplit.doPatchSplitting file pi versionedFunHash
   else ();;

let doCollectSetjmps (file : Cil.file) : Setjmp.setjmpHash_t = 
   Setjmp.collectSetjmpPoints file;;

let doSpeculativeToReal (analysis_result : Analysis.analysis_result) (file : Cil.file) (updFun : varinfo) : unit = 
   Updatepoints.speculativeToReal analysis_result file updFun

(* let doMarkInvalidSubtypesNonUpdatable (mallocFunHash : Mallocpragma.mallocHash_t) (file:Cil.file) : unit = 
   Cil.visitCilFileSameGlobals (new Ptraliastracking.markInvalidSubtypesNonUpdatableGlobal mallocFunHash) file;;
*)

let doEnsureOldAndNewTypeConsistency (vd : Versiondata.version_data) (file:Cil.file) : unit = 
   begin  
   match Updateargs.args.Updateargs.arg_currentMode with 
      Updateargs.PatchMode -> Patchcomp.ensureOldAndNewTypeConsistency vd
     | _ -> ()
   end;;

let doSuspectAliasAnalysis (f: Cil.file) : Suspectalias.SAM.mo_result = 
  if Updateargs.args.Updateargs.arg_dosuspect then
    Suspectalias.doSuspectAliasAnalysis f
  else
  { Suspectalias.SAM.sar_LPC_constraints = Deltasolver.DeltaConstraintSet.empty;
    Suspectalias.SAM.sar_fname_to_LPS_hash = H.create 11;
  }

let doMakeInternVarHash (file : Cil.file) (vd : Versiondata.version_data) : (string, unit) H.t = 
   Internvar.makeInternVarHash file vd;;

let currentTestVisitor = new Testvisitors.printVaTypesVisitor;;

(* 'debugFeatures' is very convenient for debugging and performance tuning. 
   Set it to true and you'll get nice green prints about the current stage. *)

let debugFeatures = true;;


let runFeature str f arg = 
   (if debugFeatures then ignore(Updatedebug.log [Updatedebug.BTAlways] "\027[32m%s\027[0m\n" str) else ());
   Stats.time ("Stats : " ^ str) f arg

let doUpdateFeature (file:Cil.file) : unit = 
   initInternalFun ();
   let (>>) f1 f2 = fun x -> f2 (f1 x) in 
   let funEffectHash = Hashtbl.create 4099 in 
   let funTypeHash = Hashtbl.create 4099 in 
   let analysisSolutionHash = Hashtbl.create 4099 in 
   let mallocFunHash = Mallocpragma.collectMallocPragmas file in 

   let patch_and_normal (vd : Versiondata.version_data) : unit = 
(* Aggressive GC'ing. Turn on if running out of physical memory *)      
(*       let () = Gc.set { (Gc.get()) with Gc.space_overhead = 20 } in *)
(*       let len = List.length file.globals in   *)
(*       let () = ignore(Updatedebug.log [Updatedebug.BTAlways] "Total globals: %d\n" len) in *)

      let () = runFeature "doCollectForcedPragmas - doDSU" doCollectForcedPragmas file in
      let updatefun = runFeature "doLoops - doDSU" doLoops file in
      let () = runFeature "doGlobalVarsUniqTypes - doDSU"  doGlobalVarsUniqTypes file vd in
      let () = runFeature "doVoidTypedefElim - doDSU" doVoidTypedefElim file in
      let () = runFeature "doSizeofPinDownVisitor - doDSU" doSizeofPinDownVisitor file in
      let () = runFeature "doBitfieldPinDownVisitor - doDSU" doBitfieldPinDownVisitor file in
      let () = runFeature "doExternaltypeselim - doDSU"  doExternaltypeselim vd file in
      let () = runFeature "doPtraliastracking - doAN"  doPtraliastracking mallocFunHash file in
      let () = runFeature "doMarkForcedNonUpdatableTypes - doDSU" doMarkForcedNonUpdatableTypes () in 
      let () = runFeature "doEnsureOldAndNewTypeConsistency - doDSU" doEnsureOldAndNewTypeConsistency vd file in 
      let () = runFeature "doEffectsAnalysis - doAN" doEffectsAnalysis file mallocFunHash in
      let suspect_result = runFeature "doAliasAnalysis - doAN" doSuspectAliasAnalysis file in
      let () = runFeature "doConabsinsertion - doDSU"  doConabsinsertion mallocFunHash vd file in
      let () = runFeature "doNonUpdateableConAbsRemoval - doDSU"  doNonUpdateableConAbsRemoval file in 
      let () = runFeature "doInsertUpdatePoints - doDSU"  doInsertUpdatePoints file in
      let internVarHash = runFeature "doMakeInternVarHash - doDSU"  doMakeInternVarHash file vd in
      let setjmpHash = runFeature "doCollectSetjmps - doDSU" doCollectSetjmps file in
      let analysis_result = runFeature "doAnalysis - doAN"  doAnalysis mallocFunHash funEffectHash funTypeHash analysisSolutionHash setjmpHash file suspect_result internVarHash in
      let () = runFeature "doSpeculativeToReal - doDSU"  doSpeculativeToReal analysis_result file updatefun in
      let () = runFeature "doIndirection - doDSU"  doIndirection vd file in  (* Note:  vd is empty here *)
      let () = runFeature "doTypeIndirection - doDSU"  doTypeIndirection vd file in
      ()
   in

   match Updateargs.args.Updateargs.arg_currentMode with 

        (* only do loop transformation and exit if Updateargs.args.Updateargs.arg_stoploops is enabled *) 
        Updateargs.NormalMode when Updateargs.args.Updateargs.arg_stoploops ->
	  ignore(runFeature "doLoops - doDSU" doLoops file)

      | Updateargs.NormalMode ->

           (*       let () = doExternalFunctionsVoidStarArgs file in     *)
          let vd = runFeature "createVersionData - doDSU"  Versiondata.createEmptyVersionData funEffectHash analysisSolutionHash in 
          let () = runFeature "doRemoveStaticSpecifiers - doDSU" doRemoveStaticSpecifiers file in
          let () = runFeature "doElimStructTypedefs - doDSU" doElimStructTypedefs file in
          let () = runFeature "patch_and_normal"  patch_and_normal vd in
          let () = runFeature "doPopulateVersionData - doDSU"  doPopulateVersionData vd file funTypeHash in
          let memcpy_vi = runFeature "doInsertWrappers - doDSU"  doInsertWrappers file vd in
          let versionedFunHash = runFeature "doFunVersioning - doDSU"  doFunVersioning vd file in
          let () = runFeature "doInsertCallToUpdatelibInit - doDSU"  doInsertCallToUpdatelibInit file in 
          let () = runFeature "doSaveVersionData - doDSU" doSaveVersionData Updateargs.args.Updateargs.arg_versionDataOutFile vd in
(*          let _ = Versiondata.printVersionData vd in *)
          ()

    | Updateargs.PatchMode ->
         let vd =  runFeature "loadVersionData - doDSU"  Versiondata.loadVersionData Updateargs.args.Updateargs.arg_versionDataInFile in 
         let () =   runFeature "doRemoveStaticSpecifiers - doDSU" doRemoveStaticSpecifiers file in
         let () =   runFeature "doElimStructTypedefs - doDSU" doElimStructTypedefs file in
         let pi =   runFeature "create_patch_info"  Patchcomp.create_patch_info vd in 
         let () =   runFeature "doCollectTTPragmas - doDSU" doCollectTTPragmas vd file in
         let () =   runFeature "patch_and_normal" patch_and_normal vd in
         let () =   runFeature "doMergePIintoVD - doDSU" doMergePIintoVD pi vd file funTypeHash in
         let memcpy_vi = runFeature "doInsertWrappers - doDSU" doInsertWrappers file vd in
         let () =   runFeature "doGenTTWrappers - doDSU"  doGenTTWrappers file vd memcpy_vi in
         let () =   runFeature "doCreatePatchInstallCode - doDSU" doCreatePatchInstallCode vd pi file in
         let () =   runFeature "doAddFunctionGCinfrastructure - doDSU" doAddFunctionGCinfrastructure vd file in
         let versionedFunHash = runFeature "doFunVersioning - doDSU" doFunVersioning vd file in
         let () =   runFeature "doPatchSplitting - doDSU" doPatchSplitting file pi versionedFunHash in 
         let () =   runFeature "doSaveVersionData - doDSU" doSaveVersionData Updateargs.args.Updateargs.arg_versionDataOutFile vd in
         ()



    | Updateargs.TestMode ->
         ignore(Updatedebug.log [Updatedebug.BTAlways] ">>>>>>>>>>>>>> TESTMODE <<<<<<<<<<<<<<\n");
         Deltasolver.testDeltaSolver ()
         (*ignore( visitCilFileSameGlobals currentTestVisitor file)*)

    | Updateargs.PrintVDMode ->
         Versiondata.printVersionDataFile Updateargs.args.Updateargs.arg_versionDataInFile

let feature : featureDescr = 

  { fd_name = "update";
    fd_enabled = Cilutil.update;
    fd_description = "infer safe update points";
    fd_extraopt = [("--patchmode",
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_currentMode <- Updateargs.PatchMode),
                       "compile a patch \n");

                   ("--testmode",
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_currentMode <- Updateargs.TestMode),
                       "Run in test mode \n");

                   ("--printversiondata",
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_currentMode <- Updateargs.PrintVDMode),
                       "print the version data \n");

                   ("--update-points", 
                       Arg.String (fun s -> match s with
                                       "full"       -> Updateargs.args.Updateargs.arg_updatePointsPolicy <- Updateargs.UFull
                                     | "returnonly" -> Updateargs.args.Updateargs.arg_updatePointsPolicy <- Updateargs.UReturnonly
                                     | "none"       -> Updateargs.args.Updateargs.arg_updatePointsPolicy <- Updateargs.UNone
                                     | _ -> Updatedebug.abort (Updatedebug.error "Invalid argument for --update-points :'%s' ! \nIt must be [full | retonly | none]\n" s)),
                       "=[full | returnonly | none] insert update() calls in the source code (default is returnonly) \n" ^
                                 "\t full : before each statement\n" ^
                                 "\t returnonly : before return keywords only\n" ^
                                 "\t none : no update()s are inserted" );

                   ("--effect-printing", 
                       Arg.String (fun s -> match s with
                                       "none"       -> Updateargs.args.Updateargs.arg_effectPrintingPolicy <- Updateargs.EPNone
                                     | "functions" -> Updateargs.args.Updateargs.arg_effectPrintingPolicy <- Updateargs.EPFunctions
                                     | "types" -> Updateargs.args.Updateargs.arg_effectPrintingPolicy <- Updateargs.EPTypes
                                     | "full" -> Updateargs.args.Updateargs.arg_effectPrintingPolicy <- Updateargs.EPFull

                                     | _ -> Updatedebug.abort (Updatedebug.error "Invalid argument for --effect-printing :'%s' ! \nIt must be [full | functions | types | none]\n" s)),
                       "=[full | functions | types | none] insert inferred effects as comments in the source code (default is types) \n" ^
                                 "\t full : function labels and types \n" ^
                                 "\t types : types \n" ^
                                 "\t functions : function labels \n" ^
                                 "\t none : no effects are inserted" );

                   ("--assembly", 
                       Arg.String (fun s -> match s with
                                       "warn"         -> Updateargs.args.Updateargs.arg_assembly <- Updateargs.UnsoundWarn
                                     | "ignore"       -> Updateargs.args.Updateargs.arg_assembly <- Updateargs.UnsoundIgnore
                                     | "kill-shallow" -> Updateargs.args.Updateargs.arg_assembly <- Updateargs.UnsoundKillTypeShallow
                                     | "kill-deep"    -> Updateargs.args.Updateargs.arg_assembly <- Updateargs.UnsoundKillTypeDeep
                                     | "die"          -> Updateargs.args.Updateargs.arg_assembly <- Updateargs.UnsoundDie


                                     | _ -> Updatedebug.abort (Updatedebug.error "Invalid argument for --assembly :'%s' ! \nIt must be [ warn | ignore | die]\n" s)),
                       "=[ warn | ignore | die] what to do with inline assembly \n" );

                   ("--from-void", 
                       Arg.String (fun s -> match s with
                                       "warn"         -> Updateargs.args.Updateargs.arg_fromvoid <- Updateargs.UnsoundWarn
                                     | "ignore"       -> Updateargs.args.Updateargs.arg_fromvoid <- Updateargs.UnsoundIgnore
                                     | "kill-shallow" -> Updateargs.args.Updateargs.arg_fromvoid <- Updateargs.UnsoundKillTypeShallow
                                     | "kill-deep"    -> Updateargs.args.Updateargs.arg_fromvoid <- Updateargs.UnsoundKillTypeDeep
                                     | "die"          -> Updateargs.args.Updateargs.arg_fromvoid <- Updateargs.UnsoundDie


                                     | _ -> Updatedebug.abort (Updatedebug.error "Invalid argument for --from-void :'%s' ! \nIt must be [ warn | ignore | die]\n" s)),
                       "=[ warn | ignore | die] what to do when downcasting to T * from void * and [T] not in the void * effects \n" );

                   ("--patchdirectory-out", 
                       Arg.String (fun s -> Updateargs.args.Updateargs.arg_patchDirectoryOut <- s),
                       "directory where patches will be stored");

                   ("--versiondata-out", 
                       Arg.String (fun s -> Updateargs.args.Updateargs.arg_versionDataOutFile <- s),
                       "name of version data file output file");

                   ("--versiondata-in", 
                       Arg.String (fun s -> Updateargs.args.Updateargs.arg_versionDataInFile <- s),
                       "name of version data input file");

                   ("--no-suspect-alias", 
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_dosuspect <- false),
                       "don't do suspect alias analysis \n");

                   ("--no-static-conabs", 
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_dostaticconabs <- false),
                       "don't make con/abs functions static+inline \n");

                   ("--gvartypes", 
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_dogvartypes <- true),
                       "do type introduction for global variables \n");

                   ("--stop-loops", 
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_stoploops <- true),
                       "stop after extracting loops \n");

                   ("--no-patchsplit", 
                       Arg.Unit (fun () -> Updateargs.args.Updateargs.arg_dopatchsplitting <- false),
                       "don't split patches; by default, patches are split \n");

                   ("--restrict", 
                       Arg.String (fun s -> match s with 
                                     "f"  -> Updateargs.args.Updateargs.arg_restrictFuns <- true
                                    |"g"  -> Updateargs.args.Updateargs.arg_restrictGvars <- true
                                    |"t"  -> Updateargs.args.Updateargs.arg_restrictTypes <- true
                                    | _ -> Updatedebug.abort (Updatedebug.error "Invalid argument for --restrict :'%s' ! \nIt must be [f | g | t]+\n" s)),
                       "=[f | g | t] restriction on DSU_update (default is 't') \n" ^
                                 "\t f : restrict functions \n" ^
                                 "\t g : restrict global variables \n" ^
                                 "\t t : restrict types " );    

                   ("--type-transformers", 
                       Arg.String (fun s -> match s with
                                       "yes"       -> Updateargs.args.Updateargs.arg_gentts <- Updateargs.TTGYes
                                     | "no"        -> Updateargs.args.Updateargs.arg_gentts <- Updateargs.TTGNo
                                     | "exclusive" -> Updateargs.args.Updateargs.arg_gentts <- Updateargs.TTGExclusive
                                     | _ -> Updatedebug.abort (Updatedebug.error "Invalid argument for --type-transformers :'%s' ! \nIt must be [yes | no | exclusive]\n" s)),
                       "=[yes | no | exclusive] generate type transformers when in patch generation mode (default is yes) \n" ^
                                 "\t yes : generate full patch (functions and type transformers) \n" ^
                                 "\t no : no type transformers generated, patch contains functions and type definitions only \n" ^
                                 "\t exclusive : generate type transfromers only (no types and functions); useful for generating the .custom.c " );
                   ("--enable-debug", 
                       Arg.String (fun s -> Updatedebug.bugAction s true),
                       "Enable given debug printing flag \n")
];

    fd_doit = doUpdateFeature;
    fd_post_check = true;
  } 

