(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Update points insertion *)

open Cil

module H = Hashtbl

(* 
   We implement the 'update' keyword from theory by a call to the 'update' function.

   There are currently three policies for inserting 'update' keywords:

   UNone : no 'update' keywords are inserted
   -----

   UReturnonly : 'update' keywords are inserted right before the return statements 
   -----------   in functions. The cool thing about CIL is that it inserts return 
                 statements even if the original code lacked them, so we're guaranteed 
                 to get at least one return point per function.

   UFull : 'update' keywords are inserted before each CIL *statement*, i.e. at the 
   -----   beginning of control blocks 

   We might also consider having 'update' keywords inserted before each CIL *instruction*, 
   but that might be too much.

   Updateargs.args.Updateargs.arg_updatePointsPolicy stores the updatePointsPolicy 
   (updatePointsPolicy can be set in the cmd line via --update-points)

*)


(* module that stores information about update points in a file *)
(*====================*)
module UpdatePointsInfo =
(*====================*)
struct
     
   let theUpdateStatements : (int, bool) H.t = (H.create 65535)
   
   
   (*-----------------------------------------------------------------------------------*)
   let addUpdateStatement (sid : int) : unit =
   (*-----------------------------------------------------------------------------------*)
       try
         let _ = H.find theUpdateStatements sid in
         Updatedebug.abort (Updatedebug.bug "addUpdateStatement : tried to add already-existing update() statement \n")
   
       with Not_found -> 
         H.add theUpdateStatements sid true
   
  
   (* Remove an update statement from the hash *)
   let removeUpdateStatement (sid : int) : unit = 
      H.remove theUpdateStatements sid
  
end
  

(****************************** Insertion of speculative update points ************************)

class insertUpdatesSVisitor (updateFunction : Cil.varinfo) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vstmt s =

    let isReturn s = match s.skind with Return (_,_) -> true | _ -> false
    in
      match Updateargs.args.Updateargs.arg_updatePointsPolicy with
          Updateargs.UNone -> SkipChildren   (* no updates; do it fast, ignore children *)

        | Updateargs.UReturnonly when not (isReturn s) -> (* update before Returns only; if this is not a Return look at children *)
            DoChildren

        | _ -> (* UReturnOnly and this is a Return; or UFull *)
            begin
              let loc = !Cil.currentLoc in (* this is valid since we're in a visitor *)
                
              (* make an instruction containing the update() call *)
              let callToUpdate = Call(None, Lval(var updateFunction), [], loc) in

              let _ = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] "insertUpdatesSVisitor : inserting update at (%a)\n" d_loc loc) in

              (* pack update () into a statement *)
              let newSInstr = Commoncilutil.mkStmt (Instr [callToUpdate]) in
              let _ = ignore(Updatedebug.log [Updatedebug.BTUpdatepoints] "insertUpdatesSVisitor : Generated new Instr statement with ID=%d\n" newSInstr.sid)              in

              (* put the statement into the hash; needed later for computing # of update points *)
              let () = UpdatePointsInfo.addUpdateStatement newSInstr.sid in

              (* pack the update() and the original statement into a block *)
              let newSBlock = Commoncilutil.mkStmt (Block (mkBlock ([newSInstr; s]))) in
              let _ = ignore(Updatedebug.log [Updatedebug.BTUpdatepoints] "insertUpdatesSVisitor : Generated new Block statement with ID=%d\n" newSBlock.sid)              in
              ChangeTo newSBlock
            end
end


class insertUpdatesGVisitor (updateFunction : Cil.varinfo) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vglob g =
     match g with 
       GFun(fd,loc) when not (Namedtypes.isConAbsFuncName fd.svar.vname) -> 
          ChangeTo [GFun(Cil.visitCilFunction (new insertUpdatesSVisitor updateFunction) fd, loc)]
          
     | _            -> DoChildren

end;;


let insertUpdatePoints (f : file) : unit =
    (* create an empty update() function; we'll declare it extern later, 
       and it will be taken from updatelib *)    
    let updateFunction = Updatecilutil.createUpdateFunDecl Updateconfig.speculativeUpdateFunctionName f in
 
    (* insert update() statements in the program *)
    visitCilFile (new insertUpdatesGVisitor updateFunction) f;;



(************************** speculative to real **************************************)

   
class speculativeToRealVisitor 
   (analysis_solution : Analysis.analysis_result)
   (real_update : Cil.varinfo)
   (speculativeRemovalCount : int ref)
   : cilVisitor = 
object(self)
   inherit nopCilVisitor
   
   
   method vstmt s =
      match s with 
         { skind = Instr [Call(_, Lval(Var vi, NoOffset), _, loc)]; sid = sid } when  Updateconfig.isSpeculativeUpdateFunc vi.vname ->
            let mu_id = 
               begin
               try H.find analysis_solution.Analysis.ar_speculativeUpdateMuHash sid 
               with Not_found -> Updatedebug.abort (Updatedebug.bug 
                                 "speculativeToRealVisitor: couldn't find mu for speculative update %s at statement %d (%a)\n" vi.vname sid d_loc loc)   
               end
            in 
            let real_update = Instr [Call(None, Lval(Var real_update, NoOffset), [], loc)] in
            begin try
               if H.find analysis_solution.Analysis.ar_mu_solution mu_id then 
                  begin
                  (* the speculative update is valid, turn it into a real one *)
                  s.skind <- real_update;
                  DoChildren
                  end

               else begin
                  (* the speculative update is invalid, remove it *)
                  UpdatePointsInfo.removeUpdateStatement sid;
                  incr speculativeRemovalCount;
                  s.skind <- Instr [];
                  DoChildren
               end
            with 
               Not_found ->  (* no solution => unconstrained variable => U *) 
                  s.skind <- real_update;
                  DoChildren
               
                            
            end

       | _ -> DoChildren
       
end;; 

(* This will later turn into a dynamic check : if \Delta \cap \{types updated in the patch} = \varphi, 
   (i.e. we don't try to update any type that's in our current effects)
   then allow the update, otherwise delay it (well, actually 'deny it' at current point and pass it over to upcoming update points) 
*)
class elimUpdatePointsWithNonEmptyEffectsVisitor 
   (analysis_solution : Analysis.analysis_result)
   (real_update : Cil.varinfo)
   (nonEmptyEffectRemovalCount : int ref)
   (dckHash : (string, Analysis.deltaConstKind) H.t)
   : cilVisitor = 
object(self)
   inherit nopCilVisitor
   
   method vstmt s =
      match s with 
         { skind = Instr [Call(_, Lval(Var vi, NoOffset), _, loc)]; sid = sid } when 
            (Updateconfig.isSpeculativeUpdateFunc vi.vname) || (Updateconfig.isRealUpdateFunc vi.vname) ->

            let delta_id_in,delta_id_out = 
               begin
               try let din,dout = H.find analysis_solution.Analysis.ar_stmtid_to_delta_id sid in
              let _ = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] "elimUpdatePointsWithNonEmptyEffectsVisitor : hit update point (speculative = %b) at (%a) sid %d delta_id_in %d delta_id_out %d\n" (Updateconfig.isSpeculativeUpdateFunc vi.vname) d_loc loc sid din dout) in
(din,dout)
               with Not_found -> Updatedebug.abort (Updatedebug.bug 
                                 "elimUpdatePointsWithNonEmptyEffectsVisitor: couldn't find delta for update %s at statement %d (%a)\n" vi.vname sid d_loc loc)   
               end
            in 
            begin try
               let delta = H.find analysis_solution.Analysis.ar_delta_solution delta_id_out in
               (* 
                  We have to split the delta into two disjoint sets : 
                  typeSet and 
                  functionSet

                  typeSet will tell us whether this is a universal update point or not; in other words,
                  if typeSet is not empty, we ditch this candidate update point

                  functionSet has to be converted into a big string and passed as parameter to the update
                  function; updatelib, in turn, will test the functions in the patch against the functions
                  in this string and delay an update until the function set added in the patch and the function
                  set in this big string are disjoint.
               *)
               let make_real_update delta = 
                  let delta_as_big_string = Deltasolver.printDelta delta in
                  Instr [Call(None, Lval(Var real_update, NoOffset), [Cil.mkString delta_as_big_string], loc)] 
               in
               let isXname name knd = 
                  try
                     (H.find dckHash name) = knd
                  with 
                     Not_found -> false
	       in
               let isGvarName (name : string) : bool = 
		 let res = isXname name Analysis.DckGvar in
(* 		 ignore(Updatedebug.log [Updatedebug.BTAlways] "isGvarName %s ? %b\n" name res); *)
		 res
	       in
               let isTypeName (name : string) : bool = 
		 let res = isXname name Analysis.DckType in
(* 		 ignore(Updatedebug.log [Updatedebug.BTAlways] "isTypeName %s ? %b\n" name res); *)
		 res
               in
               let isFunctionName (name : string) : bool = 
		 let res = isXname name Analysis.DckFun in
(* 		 ignore(Updatedebug.log [Updatedebug.BTAlways] "isFunctionName %s ? %b\n" name res); *)
		 res
               in

               let typeSet = Deltasolver.CapabilitySet.filter isTypeName delta in
               let functionSet = Deltasolver.CapabilitySet.filter isFunctionName delta in
               let gvarSet = Deltasolver.CapabilitySet.filter isGvarName delta in
               if (Updateconfig.isRealUpdateFunc vi.vname)
               then (* If programmer-inserted update point, don't ditch it for having non-empty effect *)
                  begin 
                     let addSetIfNeeded (needed, set) accu = 
                        if needed 
                        then Deltasolver.CapabilitySet.union set accu 
                        else accu in

                     let l = [ (Updateargs.args.Updateargs.arg_restrictTypes, typeSet);
                               (Updateargs.args.Updateargs.arg_restrictFuns,  functionSet);
                               (Updateargs.args.Updateargs.arg_restrictGvars, gvarSet); ] in
   
                     (* collect only those sets instructed by Updateargs *)   
                     let setToPrint = List.fold_right addSetIfNeeded l Deltasolver.CapabilitySet.empty in

                     s.skind <- make_real_update setToPrint;
                     DoChildren 
                  end
               else (* speculative update *)
                  begin
                     if Deltasolver.CapabilitySet.is_empty typeSet then 
                        begin
                           (* the update is valid, turn it into a real one *)
                           s.skind <- make_real_update functionSet;
                           (* ignore (Updatedebug.log [Updatedebug.BTAlways] "Validating update %s at statement %d (%a)\n" vi.vname sid d_loc loc); *)
                           DoChildren
                        end

                     else begin
                        (* the update is invalid, remove it *)
                        UpdatePointsInfo.removeUpdateStatement sid;
                        incr nonEmptyEffectRemovalCount;
                        (* ignore (Updatedebug.log [Updatedebug.BTAlways] "Ditching update %s at statement %d (%a)\n" vi.vname sid d_loc loc); *)
                        s.skind <- Instr [];
                        DoChildren
                     end
                  end
            with 
               Not_found ->  (* this shouldn't happen *) 
                  Updatedebug.abort (Updatedebug.bug 
                                        "elimUpdatePointsWithNonEmptyEffectsVisitor: couldn't find solution for delta_id %d (%a)\n" delta_id_out d_loc loc)   
               
                            
            end

       | _ -> DoChildren
       
end;; 

let fixDckHash file dckHash = 
  let funNames = H.create 1023 in
  let addIfNeeded name = Commonutil.hashtbl_add_once funNames name () in
  let doOneGlobal g = match g with
	GFun(fd,_) -> addIfNeeded fd.svar.vname
      | _ -> ()
  in
  let () = Cil.iterGlobals file doOneGlobal in
  let newDckHash = H.create 1023 in
  let doOneDckEntry n d = 
    match d with 
	Analysis.DckGvar when H.mem funNames n -> H.add newDckHash n Analysis.DckFun
      | _ -> Commonutil.hashtbl_add_once newDckHash n d
  in
  H.iter doOneDckEntry dckHash;
  newDckHash
;;

(* Given a solution to the mu-constraints convert all of the speculative update points 
   to real ones iff their associated mu is U (true), otherwise remove them. *)
let speculativeToReal (analysis_solution : Analysis.analysis_result) (file : Cil.file) (real_update:Cil.varinfo) : unit = 
   (* create two statistics counters that are passed to visitors (need to pass by ref. as visitors can't return anything) *)
   let speculativeRemovalCount = ref 0 in
   let nonEmptyEffectRemovalCount = ref 0 in
   let totalBefore = Commonutil.hashtbl_size UpdatePointsInfo.theUpdateStatements in

   visitCilFile (new elimUpdatePointsWithNonEmptyEffectsVisitor analysis_solution real_update nonEmptyEffectRemovalCount (fixDckHash file analysis_solution.Analysis.ar_constNamesHash)) file;
   visitCilFile (new speculativeToRealVisitor analysis_solution real_update speculativeRemovalCount) file;

   (* visitors remove update points from the hash, se the leftovers are the safe ones *)
   let totalAfter = Commonutil.hashtbl_size UpdatePointsInfo.theUpdateStatements in
   ignore(Updatedebug.log [Updatedebug.BTStatistics] 
             "# Update points : Safe = %d, Speculative removed = %d, Non-empty effect removed = %d, Total candidates = %d\n" totalAfter !speculativeRemovalCount !nonEmptyEffectRemovalCount totalBefore);

