(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*        Version Data      *)


(*  
This module is responsible for version data management.
Version data is a persistent data structure containing the most
up to date information about types, global variables and functions.

It is created from scratch in normal mode and dumped to disk. All
subsequent patch compilations will load the vd from disk, update it according
to type/function/var changes introduced in the patch and save it back to disk.

The cool thing about VD is that it stores snapshots, not differences 
(w.r.t. dynamic updates) of the program. 
In contrast, for source code we take a 'differential' approach
by having an initial program on top of which we stack patches.

Another cool thing is that VD is self-contained. Previous source code 
and/or previous patches *don't* have to be available in order to properly
compile patches.

In normal mode, the output VD is specified using --versiondata-out=<file>
in command line.

In patch mode, previous VD is specified as --versiondata-in=<file>
and the output VD as --versiondata-out=<file>

*)

open Cil

module H = Hashtbl

(* we store info about types/funcs/global vars and results of the analysis *)

type vd_abs_type_data = 
{
   vd_abs_version        : int;       (* the version number in the currently running program *) 
   vd_abs_type           : Cil.typ;   (* the representation type in the current program *)
   vd_abs_updatable      : bool;      (* indicates if this type is updatable *)
   vd_abs_slopsize       : int;       (* slop size, i.e. (original type size * Updateconfig.wDefaultSlopRatio) *)
}


type vd_fun_data =
{
   vd_fun_version                 : int;          (* the version number in the currently running program *)      
   vd_fun_type                    : Cil.typ;      (* type of the function -- includes analysis annotations *)
   vd_fun_indirection_varname     : string;       (* indirection var name *)
}


type vd_var_data =
{
   vd_var_vi     : Cil.varinfo;     (* varinfo of a declaration of a global var in the current program *)
}

type vd_analysis_data =
{
   vd_an_funeffect_hash     :  Analysis.funEffectHash_t;         (* map from effect ids to effects *)
   vd_an_solution           :  Analysis.deltaSolutionHash_t;     (* map from delta ids to capability sets *)
}


type version_data = 
{
   vd_abs_type_hash   :  (string, vd_abs_type_data) H.t;   (* map from type name to information about the type *)
   vd_fun_hash        :  (string, vd_fun_data) H.t;        (* map from function name to information about the function *)
   vd_var_hash        :  (string, vd_var_data) H.t;        (* map from global var name to information about the variable *)
   vd_an_data         :  vd_analysis_data;                 (* various analysis results *)

   mutable vd_current_version :  int;                      (* global version number, like an ID; this increases with each update *) 
}


let createEmptyVersionData (funEffectHash : Analysis.funEffectHash_t) 
                           (analysisSolutionHash : Analysis.deltaSolutionHash_t) = 
   { vd_abs_type_hash = H.create 4099;
     vd_fun_hash = H.create 4099;
     vd_var_hash = H.create 4099; 
     vd_an_data = { vd_an_funeffect_hash = funEffectHash;  vd_an_solution = analysisSolutionHash; };
     vd_current_version = 0; }

(* write a given version data structure to disk *)
let persist_to_disk (file_name : string) (vd : version_data) : unit =
   let out_chan = 
      try 
         open_out file_name 
      with _ ->
         Updatedebug.abort (Updatedebug.error "persist_to_disk : cannot open '%s' for writing !\n" file_name)
   in
   Marshal.to_channel out_chan vd [];
   close_out out_chan


(* get a version data structure from disk *)
let retrieve_from_disk (file_name : string) : version_data =
   let in_chan = 
      try 
         open_in file_name 
      with _ ->
         Updatedebug.abort (Updatedebug.error "retrieve_from_disk : cannot open '%s' for reading !\n" file_name)
   in
   let in_chan = open_in file_name in
   let vd = Marshal.from_channel in_chan in
   close_in in_chan;
   vd


(* Query functions, mostly used in patch mode *)
let isExistingFunction (vd:version_data) (name:string) : bool = H.mem vd.vd_fun_hash name

let isExistingType (vd:version_data) (name:string) : bool = H.mem vd.vd_abs_type_hash name

let isExistingVar (vd:version_data) (name:string) : bool = H.mem vd.vd_var_hash name

let isUpdatableType (vd:version_data) (name:string) : bool = 
   (H.find vd.vd_abs_type_hash name).vd_abs_updatable
       
(* Given the name of an abstract type, return its Cil representation *)
let getAbsType (vd:version_data) (name:string) : Cil.typ = 
   (H.find vd.vd_abs_type_hash name).vd_abs_type

(* Given the name of a global variable, return the Cil varinfo *)
let getExistingVar (vd:version_data) (name:string) : Cil.varinfo = 
   (H.find vd.vd_var_hash name).vd_var_vi
   
(* Given the name of an abstract type, return its version in the current program *)
let getCurrentTypeVersion  (vd:version_data) (tname:string) : int =
   (H.find vd.vd_abs_type_hash tname).vd_abs_version

(* Given the name of an abstract type, return the slop size *)
let getSlopSize (vd:version_data) (tname:string) : int =
   (H.find vd.vd_abs_type_hash tname).vd_abs_slopsize

(* set slop size; we don't know vd_abs_slopsize in addType, so we have to do it afterwards *)
let setSlopSize (vd : version_data) (tname : string) (slopSize : int) = 
   let old_metadata = H.find vd.vd_abs_type_hash tname in
   let new_metadata = { old_metadata with vd_abs_slopsize = slopSize} in
   H.replace vd.vd_abs_type_hash tname new_metadata;;
   
(* Given the name of an abstract type, return its version in the current program *)
let getFunctionVersion (vd : version_data) (fname : string) = 
   (H.find vd.vd_fun_hash fname).vd_fun_version

(* Given the name of a function, return its type in the current program *)
let getFunctionType (vd : version_data) (fname : string) = 
   (H.find vd.vd_fun_hash fname).vd_fun_type

(* Given the name of a function, return its type in the current program *)
let getFunctionIndirectionVarName (vd : version_data) (fname : string) = 
   (H.find vd.vd_fun_hash fname).vd_fun_indirection_varname


(* Return the global version # *)
let getCurrentVersion (vd:version_data) : int =
   vd.vd_current_version


(* Modifiers for VD *)

(* Add a function to vd *)
let addFunction (vd : version_data) (fd : Cil.fundec) (indirectionVarname : string) (funTypeHash : Analysis.funTypeHash_t) = 
   begin try 
      let funtyp = H.find funTypeHash fd.svar.vname in
      H.add vd.vd_fun_hash fd.svar.vname 
      { vd_fun_version = 0; 
        vd_fun_type = funtyp; 
        vd_fun_indirection_varname = indirectionVarname; }
   with 
      Not_found -> Updatedebug.abort (Updatedebug.bug "addFunction : cannot find '%s' in funTypeHash !\n" fd.svar.vname)
   end

(* Add a type to vd *)
let addType (vd : version_data) (isUpdatable : bool) (tabs : Cil.typ) = 
   H.add vd.vd_abs_type_hash (Updatecilutil.getAbstractTypeName tabs)
      { vd_abs_version = 0;
        vd_abs_type = tabs;
        vd_abs_updatable = isUpdatable; 
        vd_abs_slopsize = 0; }

(* Add a variable to vd *)
let addVar (vd : version_data) (vi : Cil.varinfo) = 
   H.add vd.vd_var_hash vi.vname
      { vd_var_vi = vi }

(* Increment a function's version number *)
let incFuncVersionNumber (vd : version_data) (fname : string) : unit =
   let funData = H.find vd.vd_fun_hash fname in 
   H.replace vd.vd_fun_hash fname {funData with vd_fun_version = funData.vd_fun_version + 1}
   
(* Increment a type's version number *)
let incTypeVersionNumber (vd : version_data) (tname : string) : unit =
   let typeData = H.find vd.vd_abs_type_hash tname in 
   H.replace vd.vd_abs_type_hash tname {typeData with vd_abs_version = typeData.vd_abs_version + 1}

(* Increment the global version # *)
let incrementCurrentVersion (vd:version_data) : unit =
   vd.vd_current_version <-  vd.vd_current_version + 1


(********************* Initial population of version data structure ******************************)

(* hack to deal with mutual recursion; init in main() *)
let isInternalFunction : (string -> bool) ref = ref (function s -> true)

(* Sweep thru the globals and add functions, global vars and updatable types to VD *)
let populateVD_fun_var_iter (vd : version_data) (funTypeHash : Analysis.funTypeHash_t) (g : Cil.global)  : unit =
   match g with 
    | GFun(fd,l) -> 
        if (not (!isInternalFunction fd.svar.vname)) then 
	  addFunction vd fd (Updateconfig.makeIndirectionVarName fd.svar.vname 0) funTypeHash
    
    | GType(ti,l) -> 
         let t = TNamed(ti, Commoncilutil.noAttr) in
         begin try 
            addType vd (Namedtypes.isUpdatable t) t
         with 
            Not_found -> 
               (* This must be an internal type (like __data_T) that isn't added to the named types hash.
                  FIXME:  would be better to have a list of types to ignore. *)
               ()
         end
         
    | GEnumTag(ei,l) -> () 
    | GEnumTagDecl(ei,l) -> () 
    
    | GCompTag(ci,l) -> 
         let t = TComp(ci, Commoncilutil.noAttr) in
         begin try 
           addType vd (Namedtypes.isUpdatable t) t
         with 
            Not_found -> ()  (* see note in GType case *)
         end

    | GCompTagDecl(ci,l) -> (* extern: do nothing *) ()
    
    | GVarDecl(vi,l) -> 
        if not (!isInternalFunction vi.vname) then
	  addVar vd vi
    | GVar(vi,ii,l) -> addVar vd vi
    
    | GAsm(_,l) -> ()
    | GPragma(_,l) -> () 
    | GText(_) -> ()


(* Add information about the globals to the version_data *)
let populateVersionData (vd : version_data) (file : Cil.file) (funTypeHash : Analysis.funTypeHash_t) = 
   (* vd is modified in-place *)
   Cil.iterGlobals file (populateVD_fun_var_iter vd funTypeHash)

(* Save vd to disk *)
let saveVersionData (vd : version_data) (filename:string) = 
   persist_to_disk filename vd


(* Used in patch mode to load version data back in *)
let loadVersionData (filename:string) : version_data = 
   retrieve_from_disk filename

(* Used in normal/patch mode to write vd to disk *)
let saveVersionData (vd:version_data) (filename:string) : unit =
   persist_to_disk filename vd


(* useful for printing the vd at an intermediate stage *)
let printVersionData (vd : version_data)  : unit =

   let print_vd_abs_type_data (vd_abs_type_hash : (string, vd_abs_type_data) H.t) : unit = 
      let print_vd_abs_type_data_iter s vd_abs_type_data = 
         ignore(Updatedebug.log [Updatedebug.BTAlways] "Type '%s' : version = %d, abs_type = %s, updatable = %b, slop size = %d\n"
                  s
                  vd_abs_type_data.vd_abs_version 
                  (Updatedebug.ppType vd_abs_type_data.vd_abs_type)
                  vd_abs_type_data.vd_abs_updatable
                  vd_abs_type_data.vd_abs_slopsize)
      in
      ignore(Updatedebug.log [Updatedebug.BTAlways] "\n<<<<<<<<<<<<<<<<<<<<<<<<<  Abstract Types >>>>>>>>>>>>>>>>>>>>>>>>>>>\n\n");
      H.iter print_vd_abs_type_data_iter vd_abs_type_hash
   in
   let print_vd_fun_data (vd_fun_hash : (string, vd_fun_data) H.t) : unit = 
      let print_vd_fun_data_iter s vd_fun_data = 
         ignore(Updatedebug.log [Updatedebug.BTAlways] "Function '%s' : version = %d, type = %s\n"
                  s
                  vd_fun_data.vd_fun_version 
                  (Updatedebug.ppType vd_fun_data.vd_fun_type))
      in
      ignore(Updatedebug.log [Updatedebug.BTAlways] "\n<<<<<<<<<<<<<<<<<<<<<<<<<  Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n\n");
      H.iter print_vd_fun_data_iter vd_fun_hash 

   in
   let print_vd_var_data (vd_var_hash : (string, vd_var_data) H.t) : unit = 
      let print_vd_var_data_iter s vd_var_data = 
         ignore(Updatedebug.log [Updatedebug.BTAlways] "Variable '%s' : type = %s\n"
                  s
                  (Updatedebug.ppType vd_var_data.vd_var_vi.vtype))
      in
      ignore(Updatedebug.log [Updatedebug.BTAlways] "\n<<<<<<<<<<<<<<<<<<<<<<<<< Variables >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n\n");
      H.iter print_vd_var_data_iter vd_var_hash

   in
   let print_vd_global_version (vd : version_data) : unit = 
      let globalVersion = getCurrentVersion vd in
      ignore(Updatedebug.log [Updatedebug.BTAlways] "\n<<<<<<<<<<<<<<<<<<<<<<<<< Global version : %d >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n\n" globalVersion)
   in

      ignore(Updatedebug.log [Updatedebug.BTAlways] "\n------------------------- Version Data Start  -----------------------\n\n");
      print_vd_global_version vd;
      print_vd_abs_type_data vd.vd_abs_type_hash;
      print_vd_fun_data vd.vd_fun_hash;
      print_vd_var_data vd.vd_var_hash;
      ignore(Updatedebug.log [Updatedebug.BTAlways] "\n------------------------- Version Data End    -----------------------\n\n");;



(* print the contents of a vd file;
   useful for debugging purposes (since data is saved as Ocaml marshalled data, not text).  
   N.B. : whenever vd's structure changes this function must be kept up to date *)

(* cilly must be invoked like this:
  
   cilly --doupdate --printversiondata --versiondata-in=<file>.vd <some C file to force Cil into running features>
*)

let printVersionDataFile (filename : string) : unit =
   let vd = retrieve_from_disk filename in
   printVersionData vd;;   
