(* 
 *
 * Copyright (c) 2004-2006, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil

module H = Hashtbl

(* Pair (T, W_T) : abstract type signature * wrapper type 
   TODO: this should probably be hashed on names
*)
type wrapperHashData = 
   { 
      whd_wrapper_type : Cil.typ;   (* type of wrapper (structure) *)
      whd_data_union   : Cil.typ;   (* type of union used to store data in wrapper *)
   }

type wrapperHash_t = (Typeequality.utypsig, wrapperHashData) H.t
type isWrapperHash_t = (string, bool) H.t 


let wrapperTypes : wrapperHash_t = H.create 4099
let isWrapperHash : isWrapperHash_t = H.create 4099

let isWrapperType t =
   if Namedtypes.isAbstractType t then 
   begin
      let res = H.mem isWrapperHash (Updatecilutil.getAbstractTypeName t)  in 
      ignore(Updatedebug.log [Updatedebug.BTWrappers] "LOOKUP: %s %b\n" (Updatecilutil.getAbstractTypeName t) res);
      res
   end
   else false

let isPtrToWrapperType t = match t with
     TPtr(t,_) -> isWrapperType t
   | _ -> false

(* Typedefs to forward declared types:
   When we create wrappers for types typedefs are turned into structures (the wrapper).  If the 
   typedef was to a forward declared structure, for example:
      struct s;
      typedef s t;
   then the wrapper structure that is created for the typedef will contain an s-wrapper, but this
   won't be defined until s is defined.  This causes that C compiler to fail when computing the 
   size of the wrapper for t.  To fix this we need to remember the typedefs to undefined structs 
   and define their wrappers after the struct is defined. *)


type compDefinedStatus = Defined | Undefined of Cil.typ list

type forwardTypedefHash_t = (string, compDefinedStatus) H.t


let guardedBytesSizeof t callerFunName = 
   try
     (Cil.bitsSizeOf t) / 8
   with 
     Cil.SizeOfError (s,t) -> 
        Updatedebug.abort (Updatedebug.bug "%s -> guardedBytesSizeof : '%s' while computing sizeof type %a!\n" 
                              callerFunName
                              s
                              d_plaintype t);;

let getSlopSize vd t = 
   let type_name =  Updatecilutil.getAbstractTypeName t in

   (*  slop = max (wMinSlopSize, sizeof(type) * wDefaultSlopRatio) *)
   let computeSlopSize t' = 
      let bareTypeSize = guardedBytesSizeof t' ("computeSlopSize(" ^ type_name ^ ") ") in
      let ratioBasedSize = int_of_float ( ceil( (float_of_int bareTypeSize) *. Updateconfig.wDefaultSlopRatio) ) in
         max Updateconfig.wMinSlopSize ratioBasedSize
   in
      begin match Updateargs.args.Updateargs.arg_currentMode with
            Updateargs.NormalMode ->
               begin
                  let newSlopSize = computeSlopSize t in
                     if Versiondata.isExistingType vd type_name
                     then 
                        begin 
                           Versiondata.setSlopSize vd type_name newSlopSize;
                           ignore(Updatedebug.log [Updatedebug.BTWrappers] "setSlopSize (%s) -> %d\n" type_name newSlopSize);

                           newSlopSize
                        end
                     else 
                        Updatedebug.abort (Updatedebug.bug "getWrapperType (NormalMode) : Can't find %s metadata in the vd!\n" type_name) 

               end
         | Updateargs.PatchMode ->
              begin
                 if Versiondata.isExistingType vd type_name 
                 then 
                    let existingSlopSize = Versiondata.getSlopSize vd type_name in
(* 		    Updatedebug.log [Updatedebug.BTAlways] "old type %s: slop size = %d\n" *)
(* 		      type_name existingSlopSize; *)
                       if existingSlopSize <> 0 
                       then  (* old type *)
                          begin 
                             (* sanity check first : the new type's size can't be bigger than the slop *)
                             begin
                                let newTypeSize = guardedBytesSizeof t "getWrapperType : newTypeSize" in
(* 				Updatedebug.log [Updatedebug.BTAlways] "type %a: size = %d\n" *)
(* 				  Cil.d_type t newTypeSize; *)
                                   if existingSlopSize < newTypeSize 
                                   then
                                      let newTypeSize_f        = float_of_int newTypeSize in
                                      let oldTypeSize_f        = (float_of_int existingSlopSize) /. Updateconfig.wDefaultSlopRatio in
                                      let oldTypeSize          = int_of_float oldTypeSize_f in
                                      let recommendedSlopRatio = newTypeSize_f /. oldTypeSize_f in
                                         Updatedebug.abort (Updatedebug.error "getWrapperType : type size for '%s' went from %d to %d, so the slop is not enough. Please set wDefaultSlopRatio >= %.2f or wMinSlopSize >= %d in updateconfig.ml and try again.\n" 
                                                               type_name
                                                               oldTypeSize
                                                               newTypeSize
                                                               recommendedSlopRatio 
                                                               newTypeSize)
                             end;
                             existingSlopSize
                          end
                       else    (* new type, in fact *)
                          let newSlopSize = computeSlopSize t in
                             begin 
                                Versiondata.setSlopSize vd type_name newSlopSize;
                                ignore(Updatedebug.log [Updatedebug.BTWrappers] "setSlopSize (%s) -> %d\n" type_name newSlopSize);

                                newSlopSize
                             end
                 else
                    Updatedebug.abort (Updatedebug.bug "getWrapperType (PatchMode) : Can't find %s metadata in the vd!\n" type_name) 

              end

         | _ ->  Updatedebug.abort (Updatedebug.bug "getWrapperType should only be called in NormalMode or PatchMode !\n")
      end


(* Lazy wrapper creation *)
let rec getWrapperType (vd : Versiondata.version_data) (t : Cil.typ) : Cil.typ = 
  if (not (Namedtypes.isWrapable t)) || (isWrapperType t) then 
     t
  else
     begin
     let sig_t = Typeequality.uTypeSigNoAttrs t in

     try 
        (H.find wrapperTypes sig_t).whd_wrapper_type

     with Not_found -> 

        let canonicalTypeName = Commoncilutil.sanitizeTypeName (Commoncilutil.cTN t) in


        let data_union_type = 
           (* the data field,  *)
           (*  Get the representation type to use for .data *)
           let t_unrolled_once  = Namedtypes.unrollTypeOnce t in

           (* continue to wrap by unfolding named type *)
           let t_unrolled_once' = 
              if ((Commoncilutil.isNamedType t) && (Namedtypes.isWrapable t_unrolled_once)) then 
                 getWrapperType vd t_unrolled_once 
              else 
                 t_unrolled_once 
           in

           let udata_ci = Commoncilutil.makeCompInfo false (Updateconfig.makeDataUnionName canonicalTypeName) in 

           let fPData = Commoncilutil.makeField udata_ci Updateconfig.wPDataFieldName (TPtr(t_unrolled_once',[])) in

           (* deal with TFun typedefs *)   
           let t_unrolled_once'' = 
              match t_unrolled_once' with
                   TFun _ -> TPtr(t_unrolled_once',[])
                 | _ -> t_unrolled_once'
           in
           let fData = Commoncilutil.makeField udata_ci Updateconfig.wDataFieldName t_unrolled_once'' in

           (* and the slop *)

           let fSlop = 
              Commoncilutil.makeField udata_ci Updateconfig.wSlopFieldName (Updateconfig.wMakeSlopType 0)
           in

           (* fData must come first as when initialising unions in C using static initialisers, it
              assumes the initialiser has the type of the first declaration in the union *)
           udata_ci.cfields <- [fData; fPData; fSlop]; 

           TComp(udata_ci,[])
        in



        let ci' = Commoncilutil.makeCompInfo true (Updateconfig.makeWrapperName canonicalTypeName) in

        (* Create the version field,  *)
        let fVersion = Commoncilutil.makeField ci' Updateconfig.wVersionFieldName  Updateconfig.wDefaultVersionFieldType in

        let fData = Commoncilutil.makeField ci' Updateconfig.wDataUnionName data_union_type in

        (* create .data field *)

        (* 
            WARNING !!!! In typeindirection.ml,  we rely on the fact that .version is at offset 0;
            So whenever the wrapper layout below changes, typeindirection.make_empty_tt has to change as well !!!
        *)
        begin
           ci'.ckey <- !Cil.nextCompinfoKey; incr Cil.nextCompinfoKey;
           ci'.cfields <-  [fVersion; fData]; 
           let wt = (TComp (ci',[])) 
           in
           begin
              H.add wrapperTypes sig_t {whd_wrapper_type = wt; whd_data_union = data_union_type};
              H.add isWrapperHash (Updatecilutil.getAbstractTypeName wt) true;
              wt
           end
        end
     end


(* Get the union type of the data field in the wrapper associated with t *)
let getDataUnionType t = 
   if (not (Namedtypes.isAbstractType t)) || (isWrapperType t) then 
     t
   else
      let sig_t = Typeequality.uTypeSigNoAttrs t in
      try 
         (H.find wrapperTypes sig_t).whd_data_union

      with Not_found -> 
         Updatedebug.abort (Updatedebug.bug "getDataUnionType: tried to get data union type for non-wrapper type %a\n"
            d_plaintype t)



(* Visit every type, if it is abstract and updatable (isWrapable) then replace it with a 
   wrapper type *)
class replaceAbstractTypesWithWrapperTypesVisitor (vd : Versiondata.version_data) : cilVisitor = object(self)
  inherit nopCilVisitor

  method vtype t =  
     if (Namedtypes.isWrapable t) 
     then 
        let wt = getWrapperType vd t in
           (ignore (Updatedebug.log [Updatedebug.BTWrappers] "replaceAbstractTypesWithWrapperTypes %s  ->  %s\n" 
                      (Updatedebug.ppType t) 
                      (Updatedebug.ppType wt));
           ChangeTo wt)

     else
        DoChildren
end

(* Visit every expression; if it is a cast to a wrapper (can only be a TComp) 
   then drop the cast as it's illegal; 
   This occurs because CIL inserts a cast to the named type before expressions of a TNamed that unroll to 
   a base type, e.g. 

   typedef int I;
   I i = 1; 

   becomes
   I i = (I) 1;

   See test/update/wrappers/stacked... and test/update/wrappers/static... for examples 
 *)
class elimCastsToTCompsVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  method vexpr e =  
     match e with

        CastE(TComp(ci, _) as t , e') when (isWrapperType t) ->
           ChangeTo e'

       | _ -> 
           DoChildren
end


class fixConAbsBodyVisitor (memcpy_vi : Cil.varinfo) : cilVisitor = object(self)
  inherit nopCilVisitor

  method vfunc fd = 
    let fixConAbsBody (t : Cil.typ) = 
        match fd.svar.vtype with 

          (*------------------  ABS TCOMP --------------------------------------*)    
        | TFun(TVoid _, Some [("abs",(TPtr(wrapperType,[])),[])], false, []) ->
             if isWrapperType wrapperType then 
                let theAbs = List.nth (fd.sformals) 0 in
                let fi_version = Commoncilutil.getFieldinfoByName wrapperType Updateconfig.wVersionFieldName (Some (!Cil.currentLoc)) in
                let version_field = Cil.mkMem ~addr:(Lval(var theAbs)) 
                                              ~off:(Field(fi_version, NoOffset)) in

                let abstractTypeName = Updatecilutil.getAbstractTypeName t in
                let versionVar = 
                   begin
                   try 
                      Typeindirection.getVersionVar abstractTypeName
                   with Not_found ->
                      Updatedebug.abort (Updatedebug.bug "fixConAbsBody (abs): Type %s doesn't have a version var!\n" abstractTypeName)
                   end
                in
                let set_version_instr = [Set(version_field, Lval(var versionVar), Cil.locUnknown)] in
                fd.sbody <- {battrs=[]; bstmts = [Commoncilutil.mkStmt (Instr (set_version_instr))]}

          (*------------------  CON --------------------------------------*)    
	| TFun(ret_t, ((Some [(_, TPtr(wrapperType,_), _)]) as args), _, _) -> 
             if Namedtypes.isWrapable t then
             begin
                let targetFormal = List.hd (fd.sformals) in 

                let fi_union_data = Commoncilutil.getFieldinfoByName wrapperType Updateconfig.wDataUnionName (Some (!Cil.currentLoc)) in
                let fi_data_field = Commoncilutil.getFieldinfoByName 
                                       fi_union_data.ftype Updateconfig.wDataFieldName (Some (!Cil.currentLoc)) in
                let data_field = Cil.mkAddrOf (Cil.mkMem ~addr:(Lval(var targetFormal)) 
                                                         ~off:(Field(fi_union_data,Field(fi_data_field, NoOffset)))) in

                let fi_version = Commoncilutil.getFieldinfoByName wrapperType Updateconfig.wVersionFieldName (Some (!Cil.currentLoc)) in
                let version_field = Cil.mkMem ~addr:(Lval(var targetFormal)) 
                                              ~off:(Field(fi_version, NoOffset)) in

                let abstractTypeName = Updatecilutil.getAbstractTypeName t in
                let typeTransformer = 
                   begin
                   try 
                      Typeindirection.getLatestTypeTransformer abstractTypeName
                   with Not_found ->
                      Updatedebug.abort (Updatedebug.bug "fixConAbsBody (con) : Type %s doesn't have a type transformer!\n" abstractTypeName)
                   end
                in
                let versionVar = 
                   begin
                   try 
                      Typeindirection.getVersionVar abstractTypeName
                   with Not_found ->
                      Updatedebug.abort (Updatedebug.bug "fixConAbsBody (con) : Type %s doesn't have a version var!\n" abstractTypeName)
                   end
                in
                let call_tt_instr = Call(None, Lval(var typeTransformer), [Lval (var targetFormal)], Cil.locUnknown) in
                let set_version_instr = Set(version_field, Lval(var versionVar), Cil.locUnknown) in

                let call_and_set_stmt = Commoncilutil.mkStmt (Instr [call_tt_instr; set_version_instr]) in
                let bt = {battrs=[]; bstmts = [call_and_set_stmt]} in
                let bf = {battrs=[]; bstmts = []} in
                let test_expr = BinOp(Ne,Lval(version_field), Lval(var versionVar), Cil.intType) in
                let test_and_call_tt_if_versions_differ_stmt = Commoncilutil.mkStmt (If(test_expr,bt,bf, Cil.locUnknown)) in                 


                let return_stmt = Commoncilutil.mkStmt (Return (Some(data_field), Cil.locUnknown)) in
                fd.sbody <- {battrs=[]; bstmts = [test_and_call_tt_if_versions_differ_stmt; return_stmt]};
                fd.svar.vtype <- TFun(TPtr(fi_data_field.ftype,[]), args, false, [])
             end


          (*------------------  ABS TYPEDEF --------------------------------------*)    
        | TFun(TVoid _, Some [("con",in_t,[]); ("abs",(TPtr(wrapperType,[])),[])], false, []) ->
             if isWrapperType wrapperType then 
                let theCon = List.nth (fd.sformals) 0 in (* get the first argument, namely the con *) 
                let theAbs = List.nth (fd.sformals) 1 in (* get the second argument, namely the abs *)

                let fi_union_data = Commoncilutil.getFieldinfoByName wrapperType Updateconfig.wDataUnionName (Some (!Cil.currentLoc)) in 
                let fi_data_field = Commoncilutil.getFieldinfoByName 
                                       fi_union_data.ftype Updateconfig.wDataFieldName (Some (!Cil.currentLoc)) in
                let data_field = (Cil.mkMem ~addr:(Lval(var theAbs)) ~off:(Field(fi_union_data,Field(fi_data_field,NoOffset)))) in

                let fi_version = Commoncilutil.getFieldinfoByName wrapperType Updateconfig.wVersionFieldName (Some (!Cil.currentLoc)) in
                let version_field = Cil.mkMem ~addr:(Lval(var theAbs)) 
                                              ~off:(Field(fi_version, NoOffset)) in

                let abstractTypeName = Updatecilutil.getAbstractTypeName t in
                let versionVar = 
                   begin
                   try 
                      Typeindirection.getVersionVar abstractTypeName
                   with Not_found ->
                      Updatedebug.abort (Updatedebug.bug "fixConAbsBody (abs): Type %s doesn't have a version var!\n" abstractTypeName)
                   end
                in
                let set_abs_data_instr = 
                   (* if the repr. type is an array, we can't assign in abs, and have to memcpy instead *)
                  if (Cil.isArrayType in_t)
                  then
		    (* memcpy (abs->__DSU_udata.__DSU_data, con, sizeof(con)) *)
		    [Call(None, 
                          Lval(var memcpy_vi), 
                          [Lval data_field; Lval(var theCon); SizeOfE(Lval(var theCon))], 
                          Cil.locUnknown)]
                  else
		    (* abs->__DSU_udata.__DSU_data = con *)
		    [Set(data_field, 
                         Lval(var theCon), 
                         Cil.locUnknown)] 
                in
                let set_version_instr = [Set(version_field, Lval(var versionVar), Cil.locUnknown)] in
                fd.sbody <- {battrs=[]; bstmts = [Commoncilutil.mkStmt (Instr (set_abs_data_instr @ set_version_instr))]}

        | _ -> ()
    in

    try 
       let t = Namedtypes.getAbstractTypeByConAbsName fd.svar.vname in
       fixConAbsBody t;
       SkipChildren
    with 
    Not_found ->
       SkipChildren

end

let rec createInitWrapper  (w_t:Cil.typ) (i : Cil.init) : (Cil.offset * Cil.init) list =
   let loc = Some !(Cil.currentLoc) in
   let getFieldOffset name = 
      let res =  Field(Commoncilutil.getFieldinfoByName w_t name loc, NoOffset)
      in
(*          ignore (Updatedebug.log [Updatedebug.BTAlways] "getFieldOffset %s -> %a \n" name Commoncilutil.my_d_plainoffset res); *)
         res
   in
   let fi_udata = Commoncilutil.getFieldinfoByName w_t Updateconfig.wDataUnionName (Some !(Cil.currentLoc)) in 
   let fi_data = Commoncilutil.getFieldinfoByName fi_udata.ftype Updateconfig.wDataFieldName (Some !(Cil.currentLoc)) in

   (* .data = original init *)   
   let data_compound_init = CompoundInit (fi_udata.ftype,[(Field(fi_data, NoOffset), i)]) in

   let wrapperInit = 
      [ 
        (* .version  = 0 *)
        ( getFieldOffset Updateconfig.wVersionFieldName, 
          SingleInit Cil.zero );

        (* .udata  = {.data = data_compound_init} *)
        ( getFieldOffset Updateconfig.wDataUnionName, 
          data_compound_init );

      ]
   in
      wrapperInit


let rec foldLeftCompound (ct: Cil.typ) (initl: (Cil.offset * Cil.init) list) : (Cil.offset * Cil.init) list = 
(*    ignore (Updatedebug.log [Updatedebug.BTAlways] "foldLeftCompound %s \n"  *)
(*           (Updatedebug.ppType ct));  *)
  let bail s = Updatedebug.abort (Updatedebug.bug "%s" s)
  in
  match unrollType ct with
    TArray(bt, _, _) -> 
      let ol,initl' = List.split initl in
      let initl'' = List.map (wrapInit bt) initl' in
         List.combine ol initl''

  | TComp (comp, _) as t ->
(*        ignore (Updatedebug.log [Updatedebug.BTAlways] "TComp (comp, _) %s \n"  *)
(*                   (Updatedebug.ppType t));  *)

      let getTypeOffset = function
          Field(f, NoOffset) -> f.ftype
        | Field(f, ofs) -> bail "foldLeftCompound: non-atomic field offset"
        | Index(e,ofs) -> bail "foldLeftCompound: index offset in TComp"
        | _ -> bail "foldLeftCompound: mal-formed initializer"
      in
      let wrapInitOffset ((o, i) : (Cil.offset * Cil.init)) : (Cil.offset * Cil.init) = 
         let t = (getTypeOffset o) in 
         (o, wrapInit t i)
      in
      List.map wrapInitOffset initl

  | _ -> bail "Type of Compound is not array or struct or union"


(* Insert extra initialisers to take account of stacked typedefs *)
and wrapUnroll (wt:Cil.typ) (i:Cil.init) : Cil.init = 
(*    ignore (Updatedebug.log [Updatedebug.BTAlways] "wrapUnroll (%s) %a\n"  *)
(*           (Updatedebug.ppType wt) d_init i);  *)
   if isWrapperType wt then
      let fi_udata = Commoncilutil.getFieldinfoByName wt Updateconfig.wDataUnionName (Some !(Cil.currentLoc)) in 
      let fi_data = Commoncilutil.getFieldinfoByName fi_udata.ftype Updateconfig.wDataFieldName (Some !(Cil.currentLoc)) in
      let oil = createInitWrapper wt (wrapUnroll fi_data.ftype i) in 
         begin
            CompoundInit(wt, oil)
         end
   else
      begin
(*          ignore (Updatedebug.log [Updatedebug.BTAlways] " (NO CHANGE)\n");  *)
         i
      end

and wrapInit (t : Cil.typ) (i : Cil.init) : Cil.init = 
(*    ignore (Updatedebug.log [Updatedebug.BTAlways] "wrapInit %s \n"  *)
(*           (Updatedebug.ppType t));  *)
   let i' = 
      match i with
        SingleInit e -> 
(*            ignore (Updatedebug.log [Updatedebug.BTAlways] "wrapInit : SingleInit %s \n"  *)
(*                   (Updatedebug.ppType t));  *)
           i
      | CompoundInit (t',ol) -> 
(*            ignore (Updatedebug.log [Updatedebug.BTAlways] "wrapInit : CompoundInit %s \n"  *)
(*                   (Updatedebug.ppType t'));  *)
          if isWrapperType t' 
          then
            let fi_udata = Commoncilutil.getFieldinfoByName t' Updateconfig.wDataUnionName (Some !(Cil.currentLoc)) in 
            let fi_data = Commoncilutil.getFieldinfoByName fi_udata.ftype Updateconfig.wDataFieldName (Some !(Cil.currentLoc)) in
              CompoundInit (t',foldLeftCompound fi_data.ftype ol)
            
          else
            CompoundInit (t',foldLeftCompound t' ol)
   in
      wrapUnroll t i' 




let doInit (t : Cil.typ) (ii : Cil.initinfo) : Cil.initinfo =  
(*    ignore (Updatedebug.log [Updatedebug.BTAlways] "doInit %s \n"  *)
(*           (Updatedebug.ppType t));  *)

   match ii.init with 
      Some i -> 
         {init = Some(wrapInit t i)}

    | None   ->  
         {init = None} 


let fixGlobalInit g = 
      match g with 
         GVar (vi,ii,loc) ->
            begin match vi.vtype with 
               TComp(ci,attr) ->
                  GVar(vi, doInit vi.vtype ii, loc)
             | TArray(t,l,attr) ->
                  GVar(vi, doInit vi.vtype ii, loc)
             | TNamed(ti,attr) ->
                  GVar(vi, doInit vi.vtype ii, loc)

             | _ ->
                  g

            end
        | _ -> g



(* Just a helper : split the given global list into two list, the first the globals
   before the con/abs and the second - the rest *)
let splitGlobalsBeforeConAbses (globals : Cil.global list) : (Cil.global list) * (Cil.global list) = 
   let rec splitGlobalsBeforeConAbses' before after =
     match after with
        [] -> (before,after)
      | h::t -> 
        begin 
        match h with
           GFun(fd, _) when Namedtypes.isConAbsFuncName fd.svar.vname -> (before,after)
        |  _ -> splitGlobalsBeforeConAbses' (before @ [h]) t
        end
   in 
   splitGlobalsBeforeConAbses' [] globals



(* Takes a file and wrapperises it *)
let fixWrapperInits (f : Cil.file) : Cil.global list = 
   List.map fixGlobalInit f.globals

(* Takes a file and wrapperises it; 
   returns the memcpy_vi - we'll need it later in ttwrapper *)
let insertWrapperTypes (f : Cil.file) (vd : Versiondata.version_data) : Cil.varinfo =


  (* Translate an abstract type into a global definition of the corresponding wrapper type *)
  let getGlobalWrapper t =
    let wt = getWrapperType vd t in 
    let ut = getDataUnionType t in 
      match wt,ut with
          TComp(w_ci,_), TComp(u_ci,_) -> [GCompTag(u_ci, Cil.locUnknown); GCompTag(w_ci, Cil.locUnknown)]
        | _ ->   Updatedebug.abort (Updatedebug.error 
                                      "getGlobalWrapper : getWrapperType(%s) returned non-TComp\n" 
                                      (Updatedebug.ppType t))
  in

  (* Make a forward declaration of a wrapper type for the struct defined by ci *)
  let makeForwardWrapperDecl t =
    let wt = getWrapperType vd t in 
      match wt with 
          TComp(ci,attr) ->
            GCompTagDecl({ci with cfields=[]}, Cil.locUnknown)

        | _ -> Updatedebug.abort (Updatedebug.error 
                                    "makeForwardWrapperDecl : getWrapperType(%s) returned non-TComp\n" 
                                    (Updatedebug.ppType t))
  in

  (* Generate the global data that we need to support the updating of a 
     named type *)
  let genSupportDataForUpdatableType vd t =
    if Namedtypes.isWrapable t 
    then 
       let res = (getGlobalWrapper t) in
          begin
             match res with
                [GCompTag(u_ci, _); GCompTag(w_ci, _)] ->
                   let fi = Commoncilutil.getFieldinfoByName (TComp(u_ci,[])) Updateconfig.wSlopFieldName None in
                   let slopSize = getSlopSize vd t in   
                   let newType = (Updateconfig.wMakeSlopType slopSize) in
                   fi.ftype <- newType   
              | _ -> Updatedebug.abort (Updatedebug.error 
                                    "genSupportDataForUpdatableType : getGlobalWrapper(%s) returned non-[TComp;TComp]\n" 
                                    (Updatedebug.ppType t))
          end;
          res
    else 
      []
  in


  (* Map from struct type name, S, to either Defined or Undefined(T) where T is a set of 
     typedef names pointing to the undefined struct S *)

  let forwardTypedefHash = H.create 7 in


  (* Given a type t, set its wrapper type to be defined iff t is wrappable *)
  let setWTDefined t = 
    if Namedtypes.isWrapable t then 
      begin
        let wtName = match getWrapperType vd t with 
            TComp(wtci,_) -> wtci.cname 
          | _ -> raise (Failure "globalWrapper_fold:GType") 
        in 
          H.add forwardTypedefHash wtName Defined
      end
  in


  (* given a global, if it is the definition of an abstract type, add the appropriate 
     data to support updating *)
  let globalWrapper_fold accu g = 

    let glist = 
      match g with 
        | GCompTag(ci,_) ->
            let t = TComp(ci,[]) in 
            if Namedtypes.isWrapable t then
              let ciWrapperTypeName = match getWrapperType vd t with 
                  TComp(wtci,_) -> wtci.cname 
                | _ -> raise (Failure "globalWrapper_fold") 
              in

              let forwardTypedefs = 
                begin try 
                  begin match H.find forwardTypedefHash ciWrapperTypeName with
                      Defined     ->
                        Updatedebug.abort (Updatedebug.bug "Reached a defined struct twice. How?\n")

                    | Undefined l -> 
                        H.remove forwardTypedefHash ciWrapperTypeName;
                        l
                  end
                with Not_found ->
                  []
                end
              in

                H.add forwardTypedefHash ciWrapperTypeName Defined;
                g :: (genSupportDataForUpdatableType vd (TComp(ci,[]))) @
                List.flatten (List.map (genSupportDataForUpdatableType vd) forwardTypedefs)
            else
              begin
                (* Type is not wrappable (=> non-updatable or extern), so just mark the struct as defined *)
                H.add forwardTypedefHash ci.cname Defined;
                [g]
              end


        (* typedefs to wrappers / non-updatable structs *)
        | GType( ( {ttype=TComp(ci,attr)} as ti ), _) when Namedtypes.isWrapable (TNamed(ti,[])) ->
            let namedtype = TNamed(ti,[]) in
            let comptype = TComp(ci,attr) in
            begin try
              begin match H.find forwardTypedefHash ci.cname with
                  Defined     ->
                    (* Struct is already defined, no special action needed *)
                    setWTDefined namedtype;
                    g::((genSupportDataForUpdatableType vd) namedtype)

                | Undefined l ->
                    (* The underlying struct is not defined yet and some other
                       typedefs point to it.  Add us to the list  *)
                    H.replace forwardTypedefHash ci.cname ( Undefined ( l @ [namedtype] ) );
                    [makeForwardWrapperDecl comptype; makeForwardWrapperDecl namedtype; g]

              end
            with Not_found -> 
              (* The underlying struct is not defined yet; create an undefined entry for it *)
              H.add forwardTypedefHash ci.cname ( Undefined [namedtype] );
              [makeForwardWrapperDecl comptype; makeForwardWrapperDecl namedtype; g]
            end

        (* Base case: Typdefs to base types  *)
        | GType(ti,_) -> 
            let t = TNamed(ti,[]) in
              setWTDefined t;
              g::(genSupportDataForUpdatableType vd t)

        | _ -> [g] 
    in 
      accu @ glist
  in
    begin
      (* replace types with wrapper types *)
      visitCilFile (new replaceAbstractTypesWithWrapperTypesVisitor vd) f;

      (* eliminate residual casts to abstract types (which are now illegal casts to structs) *)
      visitCilFile (new elimCastsToTCompsVisitor) f;

      (* add infastructure *)
      f.globals <- List.fold_left globalWrapper_fold [] f.globals;

      (* we need memcpy for copying arrays in ABSes *)
      let memcpyFunction_vi : Cil.varinfo = 
         let memcpyName = fst Updateconfig.memcpyDescriptor in

         (* look memcpy up first; if not found, declare it as extern *)   
         let rec memcpyLookup (gl : Cil.global list) : Cil.varinfo = match gl with

               (* couldn't find memcpy among globals; make it up as extern *) 
               [] -> 
                  let memcpyFunction_vi = Cil.makeGlobalVar memcpyName (snd Updateconfig.memcpyDescriptor) in
                   memcpyFunction_vi.vstorage <- Cil.Extern;
                   memcpyFunction_vi

            (* found as fun or extern var *)   
            | GVarDecl(vi,_) :: _ when vi.vname = memcpyName -> vi
            | GFun(fd,_) :: _ when fd.svar.vname = memcpyName -> fd.svar

            (* keep going *)   
            | _ :: t -> memcpyLookup t
         in
         memcpyLookup f.globals
      in
      let memcpyFunction_gvardecl = GVarDecl(memcpyFunction_vi, Cil.locUnknown) in

      (* Add declaration for memcpy just before con/abs *)   
      let splitglobals  = splitGlobalsBeforeConAbses f.globals in
         f.globals <- (fst splitglobals) @ 
                      [memcpyFunction_gvardecl ] @
                      (snd splitglobals);


      (* Now that wrapper types exist we need to write the body of the con functions --
         we need the wrapper types to exist to do this because con needs to return 
         &wrapper.data *)

      visitCilFile (new fixConAbsBodyVisitor memcpyFunction_vi) f;

      memcpyFunction_vi
    end

