
/*
 *
 * Library for garbage collection of functions.
 * We use a thread-safe Function Garbage Collector.
 * The main algorithm is the following:
 * 1. load the function.
 *      - it assigns the number "1" to the number of references to the current version of the function
 * 2. lock the function when in use.
 *      - it increments in a thread-safe way the counter in AllFunctions.
 * 3. unlock the function when not in use anymore.
 *      - it decrements the counter in a thread-safe way.
 * 4. at update points, if an oldVersionsTotalCount is at 0, unload is called.
 * 5. unload removes the function 
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <math.h>

#include "dsu_debug.h"

#define MAX_PATH_LENGTH 300
#define MAX_FCT_LENGTH 100

extern char * __DSU_libname;

/*  Name of the function that installs the patch (should be found in the dynamic patch file).
 *  Note: if you change this, then change Updateconfig.patch_install_function_name too!!!
 */
static char * init_fun_name = "__DSU_install_patch";
typedef void (*init_fun_t)(void *);
int __DSU_TRIGGER_FGC=0;
int __DSU_NUPDATES=0; 
/*
 * Structure for all the elements of the list.
 */
struct __DSU_FVersionElem{
  unsigned int vNumber;
  unsigned int nRef;
  char *libName;
  void *fct;
  void *handle;
  struct __DSU_FVersionElem* next;
  struct __DSU_FVersionElem* prev;
};

/*
 * Creates a __DSU_FVersionElem
 */
struct __DSU_FVersionElem* __DSU_consFVersionElem(char *libName, void *handle, unsigned int vNumber, void *fct){
   struct __DSU_FVersionElem* e = (struct __DSU_FVersionElem*) malloc(sizeof(struct __DSU_FVersionElem));
   e->libName = (char *)malloc(strlen((const char *)libName)+1);
   strcpy(e->libName,libName); 
   e->handle=handle;
   e->nRef = 0;
   e->fct=fct;
   e->vNumber=vNumber;
   e->next=NULL;
   e->prev=NULL;
   return e;
}

/*
 * Structure for the head of the list.
 */ 
struct __DSU_FVersionList{
  char *baseFunName;
  struct __DSU_FVersionElem *first;
  void *activeFunction;
  unsigned int oldVersionsTotalCount;
  unsigned int length;
};

/*
 * Creates an empty __DSU_FVersionList
 */
struct __DSU_FVersionList* __DSU_consFVersionList(char *baseFunName){
   struct __DSU_FVersionList* l = (struct __DSU_FVersionList*) malloc(sizeof(struct __DSU_FVersionList));
   l->baseFunName = (char *)malloc(strlen((const char *)baseFunName)+1);
   strcpy(l->baseFunName,baseFunName); 
   l->first = NULL;
   l->oldVersionsTotalCount = 0;
   l->length = 0;
   return l;
}

/*
 * List for functions to be GCed
 */
struct __DSU_Fct2Clean{
  struct __DSU_FVersionList *l;
  struct __DSU_Fct2Clean *next;

};

/*
 * add to a __DSU_Fct2Clean
 */
void __DSU_addFct2Clean(struct __DSU_Fct2Clean *cleaner,struct __DSU_FVersionList *l){

   struct __DSU_Fct2Clean* cl2;
   cl2= (struct __DSU_Fct2Clean*)malloc(sizeof(struct __DSU_Fct2Clean));
   cl2->l=l;
   cl2->next=cleaner->next;
   cleaner->next=cl2;
   
}

/*
 * Variable used to see if a GC has to be triggered.
 */
struct __DSU_Fct2Clean __DSU_FGARBAGE = {(struct __DSU_FVersionList *)0, (struct __DSU_Fct2Clean *)0};


/*
 * builds patch name
 * returns 0 if incomplete
 *         the pointer to the name
 */
char *__DSU_build_patch_name(char *resultName,char *name,unsigned int f_versionNumber, unsigned int p_versionNumber){
   int max_written;
   max_written=snprintf(resultName,MAX_PATH_LENGTH,"%s/%u/%s.patch.so",__DSU_libname,p_versionNumber,name); 
   if (max_written >= MAX_PATH_LENGTH){
        DSU_ERROR("Incomplete write for %s, please increase MAX_PATH_LENGTH)\n",__DSU_libname);
        return 0;
   }
   return resultName;
}
/*
 * builds function name
 * returns 0 if incomplete
 *         the pointer to the name
 */
char *__DSU_build_function_name(char *resultName,char *name,unsigned int f_versionNumber, unsigned int p_versionNumber){
   int max_written;
   max_written=snprintf(resultName,MAX_FCT_LENGTH,"%s__v%u",name,f_versionNumber);
   if (max_written >= MAX_FCT_LENGTH){
        DSU_ERROR("Incomplete write for %s, please increase MAX_FCT_LENGTH\n",name);
        return 0;
   }
 
   return resultName;

}
/*
 * This loads the function.
 * returns:
 *    0 if it was a failure.
 *    handle_pointer otherwise.
 */
void *__DSU_load(const char *libname){
  void *handle;
  init_fun_t init_fun;
  char *error;
  handle = dlopen (libname, (RTLD_LAZY|RTLD_GLOBAL));
  DSU_DEBUG("LIBFGC loading splinter: %s\n",libname)
  if (!handle) {
    DSU_ERROR( "LIBFGC : %s\n", dlerror());
    return 0;   
  }

  DSU_DEBUG( "LIBFGC : loaded %s\n",libname);
  return handle;
}

/*
 * This unloads the old function.
 * returns:
 *    -1 if it was a failure.
 *    0 if the unloading was effectively made.
 */
int __DSU_unload(struct __DSU_FVersionList *l){
   int i;
   struct __DSU_FVersionElem *e;
   struct __DSU_FVersionElem *e0;
   e=l->first->next;
   if (!(l->oldVersionsTotalCount)){
   
      for(i=1;i<l->length;i++){
        DSU_DEBUG( "FGC : removing version %d of %s length = %d\n",e->vNumber,l->baseFunName,l->length );
        // TODO REACTIVATE THIS WHEN SPLIT
#ifndef NOSPLIT
        dlclose(e->handle);
#endif
        e0=e;
        e=e->next;
        free(e0);
      } 
        l->length = 1;
        DSU_DEBUG( "FGC : done\n");
      return 0;
   }

return -1;
}


/*
 * adds an element at the end of the linkedlist
 * MO: NO NEED TO BE THREAD-SAFE, AT UPDATE-TIME
 * we add it at the beginning for the reason that it is likely that
 * most recent versions will probably be used more often.
 */

#ifdef NOSPLIT
extern void * handle;
#endif

void __DSU_addFVersionList(struct __DSU_FVersionList *l,char *name, unsigned int f_versionNumber, unsigned int p_versionNumber){
   struct __DSU_FVersionElem *e;
#ifndef NOSPLIT
   void *handle;
#endif
   void *fct;
   char functionDLName[MAX_FCT_LENGTH];
   char DLName[MAX_PATH_LENGTH];
   char * error;
   

   
   DSU_DEBUG( "__DSU_addFVersionList called with: %s function version %u patch version %u \n", name, f_versionNumber,p_versionNumber);

   // we load the dynamic library containing the function
   __DSU_build_patch_name(DLName, name, f_versionNumber, p_versionNumber);
   if (DLName==0){
	DSU_ERROR( "LIBFGC ERROR: could not append the name %s\n",DLName); 
        return;
   }
#ifndef NOSPLIT
   handle = __DSU_load(DLName);
   if (handle==0) {
	DSU_ERROR( "LIBFGC ERROR: failed to load the DL %s\n",DLName); 
        return;
   }
#endif
   // we try to get the function from the patch
   __DSU_build_function_name(functionDLName,name, f_versionNumber, p_versionNumber);
   if (functionDLName==0){
     DSU_ERROR( "LIBFGC ERROR: could not append the function name %s\n",DLName); 
     return;
   }
   fct=dlsym(handle, functionDLName);
   if ((error = dlerror()) != NULL)  {
      DSU_ERROR( "LIBFGC ERROR: %s\n", error);
      return;
   }

   // we then build the element to store the patch
   e= __DSU_consFVersionElem(name,handle,f_versionNumber,fct);
   if (__DSU_NUPDATES==0){

       // so we are before any update: we only add the active function
       // Function REMOVALS...
       l->first = e;
       l->activeFunction=e->fct;
   } else {
       // we are in a patch loading
       if (l->first!=NULL){
   
	 // in that case, this is NOT the first version of the function that is added.
         // thus we may triger FGC or not...
         l->oldVersionsTotalCount=(l->oldVersionsTotalCount)+(l->first->nRef);
         if (!(l->oldVersionsTotalCount)) {
             __DSU_TRIGGER_FGC=1;
             __DSU_addFct2Clean(&__DSU_FGARBAGE,l);
         }
         e->next=l->first;
         l->first->prev = e;
       }
       // we assign the function to the active function
       l->first=e;
       l->activeFunction=e->fct;
   }
   l->length++;
}

/*
 * Prints the state of a version list on __DSU_output.
 * MO: SHOULD BE THREAD-SAFE
 */
void __DSU_printFVersionList(struct __DSU_FVersionList *l){
   int i;
   struct __DSU_FVersionElem *e;
   if (l->first == NULL) {
     DSU_DEBUG( "FUNCTION GC : %s oldVersionsLocks: %d actual locks: None\n", l->baseFunName, l->oldVersionsTotalCount);
     return;
   }
   e=l->first;
   DSU_DEBUG( "FUNCTION GC : %s oldVersionsLocks: %d actual locks: %d\n%s\n", l->baseFunName, l->oldVersionsTotalCount,l->first->nRef , e->libName);

   for(i=1;i<l->length;i++){
      e=e->next;
      DSU_DEBUG( " <- %s\n", e->libName);
   }
}


/*
 * this function makes the effective garbage collection
 * 
 * 
 */
void __DSU_FGC(){
  struct  __DSU_Fct2Clean*  tmpGarbage=&__DSU_FGARBAGE;
  struct  __DSU_Fct2Clean*  tmpGarbage2;
 
  DSU_DEBUG( "_DSU_FGC called.\n");
  while(tmpGarbage!=NULL){
     if (tmpGarbage->l!=NULL) __DSU_unload(tmpGarbage->l);
     tmpGarbage2=tmpGarbage->next;
     if (tmpGarbage!=&__DSU_FGARBAGE) free(tmpGarbage);
     else __DSU_FGARBAGE.next=NULL;
     tmpGarbage=tmpGarbage2; 
     
  }
  __DSU_TRIGGER_FGC=0;
  
}


/*
 * This locks the function.
 * MO: SHOULD BE THREAD-SAFE
 * returns:
 *    -1 if it was a failure.
 *    n: number of locks taken on this function (including this one).
 */
int __DSU_lock(struct __DSU_FVersionList *l, unsigned int versionNumber){
   int n=0;
   
   if (l->first!=NULL){
      DSU_FDEBUG( "_DSU_lock called on version %d of %s current version number %d\n",versionNumber,l->baseFunName,l->first->vNumber);
      if (versionNumber==l->first->vNumber) n=l->first->nRef++;
   } else DSU_FDEBUG( "_DSU_lock called on version %d of %s and this is the first version!!!!\n",versionNumber,l->baseFunName);
return n;
}

/*
 * This locks the function.
 * MO: SHOULD BE THREAD-SAFE
 * returns:
 *    -1 if it was a failure.
 *    n: number of locks still taken on this function (after this one is released).
 */
int __DSU_unlock(struct __DSU_FVersionList *l, unsigned int versionNumber){
  int n=0;
   if (l->first!=NULL){
  	DSU_FDEBUG( "_DSU_unlock called on version %d of %s current version number %d old version locks: %d\n",versionNumber,l->baseFunName, l->first->vNumber,l->oldVersionsTotalCount);
  	if (versionNumber==(l->first->vNumber))
		n=(l->first->nRef)--;
  	else {
        	l->oldVersionsTotalCount--;
		n=l->oldVersionsTotalCount;
        	if (n==0) {
           	DSU_DEBUG("***************** TRIGGER FGC ****************\n");
           	__DSU_TRIGGER_FGC=1;
           	__DSU_addFct2Clean(&__DSU_FGARBAGE,l);
        	}
  	}
// we can add so many optimizations here!!!
   } else DSU_FDEBUG( "_DSU_unlock called on version %d of %s and this is the first version!!!!\n",versionNumber,l->baseFunName);
return n;
}


