/* 
 * generic library for dynamic loading/updating
 *
 * Note:  To use this library '__DSU_libname' needs to be initialised at the end of the file
 *        to whatever you want to call your dynamic patch file.
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <signal.h>

#include "dsu_config.h"
#include "dsu_debug.h"
#include "pmeta.h"

#ifdef DSU_USE_THREADS
#include "mt.h"
#endif // DSU_USE_THREADS

/* .so handle */
typedef void * handle_t;

/* file descriptor used for the output */
FILE *__DSU_debug_file = NULL;

/*  Name of dynamic patch file to load; changes from program to program, so
 *  we can't hardcode it; rather, the 'make' process that builds the executable and 
 *  the patch will take care of it.
 */
extern char * __DSU_libname;
extern char * __DSU_progname;

/*
 * flag to decide if we are in the first version or not
 */
extern int __DSU_NUPDATES;

/*  Name/type of the function that installs the patch (should be found in the dynamic patch file).
 *  Note: if you change these, then change Updateconfig.patch_install_function_name/type too!!!
 */
static char * init_fun_name = "__DSU_install_patch";
typedef void (*init_fun_t)(void);


/* Name of the global variable containing the metadata (should be found in the dynamic patch file).
 * Note: if you change this, then change Updateconfig.patch_metadata_var_name/type too!!!
 */
static char * metadata_var_name = "__DSU_metadata";
typedef char * metadata_t;


/*  Specifies the signal number to trigger an update. 
 *  N.B. Make sure this doesn't clash with the signals your program uses !!
 */
static int SIGUPDATE = SIGUSR2;

/*  Indicates whether an update is available
 *  If you force this to be 1, the first __DSU_update() will go ahead
 *  and try to load the patch !
 */
volatile int update_requested = 0;
static unsigned long long update_request_time = 0;


static volatile int patch_pending = 0;


/*  Declaration of sig structure */
static struct sigaction update_action;


/*
 *  Updating handler. It sets 'update_requested' to 1, so the first subsequent __DSU_update()
 *  will load the patch.
 */
static void updating_request_handler(int signum){
  update_requested = 1 ;
  DSU_DEBUG("updating_request_handler called\n"); 
  asm volatile( "rdtsc" : "=A" (update_request_time) ); 

}

#ifdef DSU_USE_THREADS
pthread_mutex_t update_mtx;
#endif // DSU_USE_THREADS


/*  This function should be called at the start of any updatable program
 *  in order to trap SIGUPDATE and route it to our updating_request_handler
 */
void __DSU_init_handler(){
  // code to initialise trace

#if DSU_ENABLE_DEBUG == 1 || DSU_ENABLE_ERROR == 1 || DSU_ENABLE_FDEBUG == 1
  __DSU_debug_file = fopen("DSU_output.txt","w");
  if (!__DSU_debug_file) fprintf(stderr,"Couldn't open DSU_output.txt for writing, disabling tracing\n");
  //debug_file = stdout;
#endif

  update_action.sa_handler = updating_request_handler;
  // MWH: changed from sigemptyset.  This ensures that no application
  // signals will be handled while we're updating; which could cause
  // inconsistent behavior (e.g., if a signal handler was being
  // updated).
  // MWH: do I need to use the SA_RESTART flag?
  sigfillset(&update_action.sa_mask);
  update_action.sa_flags=0;

  // linking SIGUPDATE to updating_request_handler
  sigaction(SIGUPDATE,&update_action,NULL);

  // init multi-threading stuff, if needed
#ifdef DSU_USE_THREADS
  __DSU_mt_init();

  pthread_mutex_init(&update_mtx, NULL);
#endif // DSU_USE_THREADS


}

static int __DSU_CurrentVNumber=0;

/* the last patch number; after the last patch is loaded, 
   statistics are printed */
static int print_stats_threshold = 20;

/* Function garbage collection */
extern int __DSU_TRIGGER_FGC; 
extern void __DSU_FGC();

#define MAX_LIBNAME_LENGTH 300

#ifdef NOSPLIT
void * handle;
#endif
unsigned long long load_times[20];
unsigned long long app_times[20];

/* try to load the patch and return the handle + metadata; if failure, return NULL */
void * load_patch(int version, metadata_t * metadata)
{
#ifndef NOSPLIT
  handle_t handle;
#endif
  char libname0[MAX_LIBNAME_LENGTH];
  unsigned long long time1, time2;
  char *error;

  /* Path name convention for multiple patches
   *
   *          patch_directory/version_number/patch.so
   */
  snprintf(libname0, MAX_LIBNAME_LENGTH, "%s/%d/%s.patch.so", __DSU_libname, version, __DSU_progname);


  DSU_DEBUG("UPDATELIB::load_patch version %d  : Opening %s ...\n", version, libname0);

  asm volatile( "rdtsc" : "=A" (time1) ); 

  /* Open the .patch.so */
  handle = dlopen (libname0, RTLD_LAZY|RTLD_GLOBAL);

  if (!handle) {
    DSU_ERROR("UPDATELIB::load_patch version %d  : %s\n", version, dlerror());
    exit(1);
  }
  DSU_DEBUG("UPDATELIB::load_patch version %d  : Opened.\n", version);

#ifdef USE_PATCH_METADATA
  /* Look up __DSU_metadata */
  DSU_DEBUG("UPDATELIB::load_patch version %d  : Searching for symbol %s ...\n", version, metadata_var_name);

  *metadata = *((metadata_t *)dlsym(handle, metadata_var_name));

  if ((error = dlerror()) != NULL)  {
    DSU_ERROR("UPDATELIB::load_patch version %d  : %s\n", version, error);
    exit(1);
  }
  DSU_DEBUG("UPDATELIB::load_patch version %d  : %s found @ %p [%s ]\n", version, metadata_var_name, *metadata, *metadata);
#endif /* USE_PATCH_METADATA */

  asm volatile( "rdtsc" : "=A" (time2) ); 

  DSU_DEBUG("UPDATELIB::load_patch version %d  : load_patch finished in %lld ticks.\n",  version, (time2 - time1));

  load_times[version]= time2 - time1;

  return handle;
}


int apply_patch(handle_t handle, int version)
{
  init_fun_t init_fun;
  unsigned long long time1, time2;
  char *error;

  asm volatile( "rdtsc" : "=A" (time1) ); 

  /* Look up __DSU_install_patch */
  DSU_DEBUG("UPDATELIB::apply_patch version %d  : Searching for symbol %s ...\n", version, init_fun_name);

  init_fun = (init_fun_t)dlsym(handle, init_fun_name);
  if ((error = dlerror()) != NULL)  {
    DSU_ERROR("UPDATELIB::apply_patch version %d  : %s\n", version, error);
    exit(1);
  }

  DSU_DEBUG("UPDATELIB::apply_patch version %d  : %s found \n", version, init_fun_name);

  /* invoke __DSU_install_patch */
  DSU_DEBUG("UPDATELIB::apply_patch version %d  : Handle %p, Init_fun %p Calling %s \n", version, 
            (void *)handle, 
            (void *)init_fun, 
            init_fun_name);
  init_fun();

  asm volatile( "rdtsc" : "=A" (time2) ); 

  DSU_DEBUG("UPDATELIB::apply_patch version %d  : apply_patch finished in %lld ticks.\n", version,  (time2 - time1));
  app_times[version]= time2 - time1;
  return 1;
}

/*  This is called at every possible update point, so we'd better keep it lite
 *  If there's no pending SIGUPDATE, this function is a no-op.
 */

/* Note: the prototype of __DSU_update is hardcoded in updateconfig.ml, so if you change it here, you must
 *  sync updateconfig.ml too
 */

int __DSU_update(char * forbidden_functions)
{
  handle_t handle;
  metadata_t metadata;  
  int update_performed = 0;
  int i;
  int barrier_val;
  unsigned long long update_apply_time;

#ifdef DSU_USE_THREADS
  pthread_mutex_lock(&update_mtx);
#endif // DSU_USE_THREADS

  DSU_DEBUG("UPDATELIB::__DSU_update %d  : entered.\n",  __DSU_CurrentVNumber);

  /* hook for triggering garbage collection */
  if(__DSU_TRIGGER_FGC) 
    __DSU_FGC();

  if (! update_requested)
    {
#ifdef DSU_USE_THREADS
      pthread_mutex_unlock(&update_mtx);
#endif // DSU_USE_THREADS

      return update_performed;
    }
  //OK, an update has been requested.

  if (!patch_pending)
    { // load patch
      handle = load_patch(__DSU_CurrentVNumber, &metadata);

      patch_pending = 1;

#ifdef USE_PATCH_METADATA
      // inform pmeta about our metadata
      new_metadata(metadata, __DSU_CurrentVNumber);
#endif /* USE_PATCH_METADATA */
    }

  /* At this point we have the patch ready, but not installed yet
     Check permission
   */

#ifdef DSU_USE_THREADS

  // are threads barrier-sync'ed ?
  barrier_val = threads_synchronized_test_set();
  if (barrier_val == THREADS_SYNCED)
    {
#endif // DSU_USE_THREADS

#ifdef USE_PATCH_METADATA
      if (is_allowed_update(forbidden_functions))
        {
#endif /* USE_PATCH_METADATA */
          asm volatile( "rdtsc" : "=A" (update_apply_time) ); 

          DSU_DEBUG("UPDATELIB::__DSU_update %d  : Update allowed. Quiescence reached in %lld ticks (%lld us)\n",  
                    __DSU_CurrentVNumber, 
                    (update_apply_time - update_request_time),
                    (update_apply_time - update_request_time)/2009);

          apply_patch(handle, __DSU_CurrentVNumber);

        patch_pending = 0;
        update_requested = 0;
        update_performed = 1;

        __DSU_CurrentVNumber++;

          /* we're now ready for another update */

#ifdef USE_PATCH_METADATA
        }
      else
        DSU_DEBUG("UPDATELIB::__DSU_update %d  : Update not allowed (capabilities say not safe now).\n",  __DSU_CurrentVNumber);
#endif /* USE_PATCH_METADATA */

#ifdef DSU_USE_THREADS

      // alright, now release the barrier
      threads_synchronized_release();
    }
  else
    DSU_DEBUG("UPDATELIB::__DSU_update %d  : Update not allowed (threads not sync'ed, %d threads missing).\n",  __DSU_CurrentVNumber, barrier_val);

#endif // DSU_USE_THREADS


  if (__DSU_CurrentVNumber > print_stats_threshold)  
    {
      DSU_ERROR("UPDATELIB::PERFORMANCE\n");

      DSU_ERROR("Load: ");
      for (i = 0; i <= print_stats_threshold; i++)
        DSU_ERROR("%.2f ",  load_times[i]/1994.0);
      DSU_ERROR("\n");

      DSU_ERROR("App: ");
      for (i = 0; i <= print_stats_threshold; i++)
        DSU_ERROR("%.2f ",  app_times[i]/1994.0);
      DSU_ERROR("\n");

      DSU_ERROR("Total: ");
      for (i = 0; i <= print_stats_threshold; i++)
        DSU_ERROR("%.2f ",  (load_times[i] + app_times[i])/1994.0);
      DSU_ERROR("\n");


    }

#ifdef DSU_USE_THREADS
      pthread_mutex_unlock(&update_mtx);
#endif // DSU_USE_THREADS

  return update_performed;
}



